@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2021 Canonical Ltd
+ * Copyright (C) 2015-2023 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -20,6 +20,7 @@
 package asserts
 
 import (
+	"errors"
 	"fmt"
 	"regexp"
 	"time"
@@ -32,7 +33,8 @@ var validAccountKeyName = regexp.MustCompile(`^(?:[a-z0-9]+-?)*[a-z](?:-?[a-z0-9
 type AccountKey struct {
 	assertionBase
 	sinceUntil
-	pubKey PublicKey
+	constraintMatchers []attrMatcher
+	pubKey             PublicKey
 }
 
 type sinceUntil struct {
@@ -127,6 +129,37 @@ func (ak *AccountKey) publicKey() PublicKey {
 	return ak.pubKey
 }
 
+// ConstraintsPrecheck checks whether the given type and headers match the signing constraints of the account key.
+func (ak *AccountKey) ConstraintsPrecheck(assertType *AssertionType, headers map[string]interface{}) error {
+	headersWithType := copyHeaders(headers)
+	headersWithType["type"] = assertType.Name
+	if !ak.matchAgainstConstraints(headersWithType) {
+		return fmt.Errorf("headers do not match the account-key constraints")
+	}
+	return nil
+}
+
+func (ak *AccountKey) matchAgainstConstraints(headers map[string]interface{}) bool {
+	matchers := ak.constraintMatchers
+	// no constraints, everything is allowed
+	if len(matchers) == 0 {
+		return true
+	}
+	for _, m := range matchers {
+		if m.match("", headers, &attrMatchingContext{
+			attrWord: "header",
+		}) == nil {
+			return true
+		}
+	}
+	return false
+}
+
+// canSign checks whether the given assertion matches the signing constraints of the account key.
+func (ak *AccountKey) canSign(a Assertion) bool {
+	return ak.matchAgainstConstraints(a.Headers())
+}
+
 func checkPublicKey(ab *assertionBase, keyIDName string) (PublicKey, error) {
 	pubKey, err := DecodePublicKey(ab.Body())
 	if err != nil {
@@ -150,7 +183,7 @@ func (ak *AccountKey) checkConsistency(db RODatabase, acck *AccountKey) error {
 	_, err := db.Find(AccountType, map[string]string{
 		"account-id": ak.AccountID(),
 	})
-	if IsNotFound(err) {
+	if errors.Is(err, &NotFoundError{}) {
 		return fmt.Errorf("account-key assertion for %q does not have a matching account assertion", ak.AccountID())
 	}
 	if err != nil {
@@ -167,7 +200,7 @@ func (ak *AccountKey) checkConsistency(db RODatabase, acck *AccountKey) error {
 			"account-id": ak.AccountID(),
 			"name":       ak.Name(),
 		})
-		if err != nil && !IsNotFound(err) {
+		if err != nil && !errors.Is(err, &NotFoundError{}) {
 			return err
 		}
 		for _, assertion := range assertions {
@@ -215,14 +248,76 @@ func assembleAccountKey(assert assertionBase) (Assertion, error) {
 		return nil, err
 	}
 
+	var matchers []attrMatcher
+	if cs, ok := assert.headers["constraints"]; ok {
+		matchers, err = checkAKConstraints(cs)
+		if err != nil {
+			return nil, err
+		}
+	}
+
 	// ignore extra headers for future compatibility
 	return &AccountKey{
-		assertionBase: assert,
-		sinceUntil:    *sinceUntil,
-		pubKey:        pubk,
+		assertionBase:      assert,
+		sinceUntil:         *sinceUntil,
+		constraintMatchers: matchers,
+		pubKey:             pubk,
 	}, nil
 }
 
+func checkAKConstraints(cs interface{}) ([]attrMatcher, error) {
+	csmaps, ok := cs.([]interface{})
+	if !ok {
+		return nil, fmt.Errorf("assertions constraints must be a list of maps")
+	}
+	if len(csmaps) == 0 {
+		// there is no syntax producing this scenario but be robust
+		return nil, fmt.Errorf("assertions constraints cannot be empty")
+	}
+	matchers := make([]attrMatcher, 0, len(csmaps))
+	for _, csmap := range csmaps {
+		m, ok := csmap.(map[string]interface{})
+		if !ok {
+			return nil, fmt.Errorf("assertions constraints must be a list of maps")
+		}
+		hm, err := checkMapWhat(m, "headers", "constraint")
+		if err != nil {
+			return nil, err
+		}
+		if hm == nil {
+			return nil, fmt.Errorf(`"headers" constraint mandatory in asserions constraints`)
+		}
+		t, ok := hm["type"]
+		if !ok {
+			return nil, fmt.Errorf("type header constraint mandatory in asserions constraints")
+		}
+		tstr, ok := t.(string)
+		if !ok {
+			return nil, fmt.Errorf("type header constraint must be a string")
+		}
+		if tstr != regexp.QuoteMeta(tstr) {
+			return nil, fmt.Errorf("type header constraint must be a precise string and not a regexp")
+		}
+		cc := compileContext{
+			opts: &compileAttrMatcherOptions{},
+		}
+		matcher, err := compileAttrMatcher(cc, hm)
+		if err != nil {
+			return nil, fmt.Errorf("cannot compile headers constraint: %v", err)
+		}
+		matchers = append(matchers, matcher)
+	}
+	return matchers, nil
+}
+
+func accountKeyFormatAnalyze(headers map[string]interface{}, body []byte) (formatnum int, err error) {
+	formatnum = 0
+	if _, ok := headers["constraints"]; ok {
+		formatnum = 1
+	}
+	return formatnum, nil
+}
+
 // AccountKeyRequest holds an account-key-request assertion, which is a self-signed request to prove that the requester holds the private key and wishes to create an account-key assertion for it.
 type AccountKeyRequest struct {
 	assertionBase
@@ -256,8 +351,8 @@ func (akr *AccountKeyRequest) PublicKeyID() string {
 }
 
 // signKey returns the underlying public key of the requested account key.
-func (akr *AccountKeyRequest) signKey() PublicKey {
-	return akr.pubKey
+func (akr *AccountKeyRequest) signKey(db RODatabase) (PublicKey, error) {
+	return akr.pubKey, nil
 }
 
 // Implement further consistency checks.
@@ -265,7 +360,7 @@ func (akr *AccountKeyRequest) checkConsistency(db RODatabase, acck *AccountKey)
 	_, err := db.Find(AccountType, map[string]string{
 		"account-id": akr.AccountID(),
 	})
-	if IsNotFound(err) {
+	if errors.Is(err, &NotFoundError{}) {
 		return fmt.Errorf("account-key-request assertion for %q does not have a matching account assertion", akr.AccountID())
 	}
 	if err != nil {
@@ -308,6 +403,9 @@ func assembleAccountKeyRequest(assert assertionBase) (Assertion, error) {
 		return nil, err
 	}
 
+	// XXX TODO: support constraints also in account-key-request when
+	// implementing more fully automated registration flows
+
 	// ignore extra headers for future compatibility
 	return &AccountKeyRequest{
 		assertionBase: assert,
