@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2022 Canonical Ltd
+ * Copyright (C) 2015-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -23,8 +23,10 @@ import (
 	"bufio"
 	"bytes"
 	"crypto"
+	"encoding/json"
 	"fmt"
 	"io"
+	"reflect"
 	"sort"
 	"strconv"
 	"strings"
@@ -39,6 +41,7 @@ type typeFlags int
 const (
 	noAuthority typeFlags = 1 << iota
 	sequenceForming
+	jsonBody
 )
 
 // MetaHeaders is a list of headers in assertions which are about the assertion
@@ -100,7 +103,7 @@ func (at *AssertionType) MaxSupportedFormat() int {
 	return maxSupportedFormat[at.Name]
 }
 
-// SequencingForming returns true if the assertion type has a positive
+// SequenceForming returns true if the assertion type has a positive
 // integer >= 1 as the last component (preferably called "sequence")
 // of its primary key over which the assertions of the type form
 // sequences, usually without gaps, one sequence per sequence key (the
@@ -123,23 +126,26 @@ func (at *AssertionType) AcceptablePrimaryKey(key []string) bool {
 
 // Understood assertion types.
 var (
-	AccountType         = &AssertionType{"account", []string{"account-id"}, nil, assembleAccount, 0}
-	AccountKeyType      = &AssertionType{"account-key", []string{"public-key-sha3-384"}, nil, assembleAccountKey, 0}
-	RepairType          = &AssertionType{"repair", []string{"brand-id", "repair-id"}, nil, assembleRepair, sequenceForming}
-	ModelType           = &AssertionType{"model", []string{"series", "brand-id", "model"}, nil, assembleModel, 0}
-	SerialType          = &AssertionType{"serial", []string{"brand-id", "model", "serial"}, nil, assembleSerial, 0}
-	BaseDeclarationType = &AssertionType{"base-declaration", []string{"series"}, nil, assembleBaseDeclaration, 0}
-	SnapDeclarationType = &AssertionType{"snap-declaration", []string{"series", "snap-id"}, nil, assembleSnapDeclaration, 0}
-	SnapBuildType       = &AssertionType{"snap-build", []string{"snap-sha3-384"}, nil, assembleSnapBuild, 0}
-	SnapRevisionType    = &AssertionType{"snap-revision", []string{"snap-sha3-384", "provenance"}, map[string]string{"provenance": naming.DefaultProvenance}, assembleSnapRevision, 0}
-	SnapDeveloperType   = &AssertionType{"snap-developer", []string{"snap-id", "publisher-id"}, nil, assembleSnapDeveloper, 0}
-	SystemUserType      = &AssertionType{"system-user", []string{"brand-id", "email"}, nil, assembleSystemUser, 0}
-	ValidationType      = &AssertionType{"validation", []string{"series", "snap-id", "approved-snap-id", "approved-snap-revision"}, nil, assembleValidation, 0}
-	ValidationSetType   = &AssertionType{"validation-set", []string{"series", "account-id", "name", "sequence"}, nil, assembleValidationSet, sequenceForming}
-	StoreType           = &AssertionType{"store", []string{"store"}, nil, assembleStore, 0}
-	PreseedType         = &AssertionType{"preseed", []string{"series", "brand-id", "model", "system-label"}, nil, assemblePreseed, 0}
+	AccountType              = &AssertionType{"account", []string{"account-id"}, nil, assembleAccount, 0}
+	AccountKeyType           = &AssertionType{"account-key", []string{"public-key-sha3-384"}, nil, assembleAccountKey, 0}
+	RepairType               = &AssertionType{"repair", []string{"brand-id", "repair-id"}, nil, assembleRepair, sequenceForming}
+	ModelType                = &AssertionType{"model", []string{"series", "brand-id", "model"}, nil, assembleModel, 0}
+	SerialType               = &AssertionType{"serial", []string{"brand-id", "model", "serial"}, nil, assembleSerial, 0}
+	BaseDeclarationType      = &AssertionType{"base-declaration", []string{"series"}, nil, assembleBaseDeclaration, 0}
+	SnapDeclarationType      = &AssertionType{"snap-declaration", []string{"series", "snap-id"}, nil, assembleSnapDeclaration, 0}
+	SnapBuildType            = &AssertionType{"snap-build", []string{"snap-sha3-384"}, nil, assembleSnapBuild, 0}
+	SnapRevisionType         = &AssertionType{"snap-revision", []string{"snap-sha3-384", "provenance"}, map[string]string{"provenance": naming.DefaultProvenance}, assembleSnapRevision, 0}
+	SnapDeveloperType        = &AssertionType{"snap-developer", []string{"snap-id", "publisher-id"}, nil, assembleSnapDeveloper, 0}
+	SystemUserType           = &AssertionType{"system-user", []string{"brand-id", "email"}, nil, assembleSystemUser, 0}
+	ValidationType           = &AssertionType{"validation", []string{"series", "snap-id", "approved-snap-id", "approved-snap-revision"}, nil, assembleValidation, 0}
+	ValidationSetType        = &AssertionType{"validation-set", []string{"series", "account-id", "name", "sequence"}, nil, assembleValidationSet, sequenceForming}
+	StoreType                = &AssertionType{"store", []string{"store"}, nil, assembleStore, 0}
+	PreseedType              = &AssertionType{"preseed", []string{"series", "brand-id", "model", "system-label"}, nil, assemblePreseed, 0}
+	SnapResourceRevisionType = &AssertionType{"snap-resource-revision", []string{"snap-id", "resource-name", "resource-sha3-384", "provenance"}, map[string]string{"provenance": naming.DefaultProvenance}, assembleSnapResourceRevision, 0}
+	SnapResourcePairType     = &AssertionType{"snap-resource-pair", []string{"snap-id", "resource-name", "resource-revision", "snap-revision", "provenance"}, map[string]string{"provenance": naming.DefaultProvenance}, assembleSnapResourcePair, 0}
+	ConfdbType               = &AssertionType{"confdb", []string{"account-id", "name"}, nil, assembleConfdb, jsonBody}
 
-// ...
+	// ...
 )
 
 // Assertion types without a definite authority set (on the wire and/or self-signed).
@@ -147,28 +153,33 @@ var (
 	DeviceSessionRequestType = &AssertionType{"device-session-request", []string{"brand-id", "model", "serial"}, nil, assembleDeviceSessionRequest, noAuthority}
 	SerialRequestType        = &AssertionType{"serial-request", nil, nil, assembleSerialRequest, noAuthority}
 	AccountKeyRequestType    = &AssertionType{"account-key-request", []string{"public-key-sha3-384"}, nil, assembleAccountKeyRequest, noAuthority}
+	ConfdbControlType        = &AssertionType{"confdb-control", []string{"brand-id", "model", "serial"}, nil, assembleConfdbControl, noAuthority}
 )
 
 var typeRegistry = map[string]*AssertionType{
-	AccountType.Name:         AccountType,
-	AccountKeyType.Name:      AccountKeyType,
-	ModelType.Name:           ModelType,
-	SerialType.Name:          SerialType,
-	BaseDeclarationType.Name: BaseDeclarationType,
-	SnapDeclarationType.Name: SnapDeclarationType,
-	SnapBuildType.Name:       SnapBuildType,
-	SnapRevisionType.Name:    SnapRevisionType,
-	SnapDeveloperType.Name:   SnapDeveloperType,
-	SystemUserType.Name:      SystemUserType,
-	ValidationType.Name:      ValidationType,
-	ValidationSetType.Name:   ValidationSetType,
-	RepairType.Name:          RepairType,
-	StoreType.Name:           StoreType,
+	AccountType.Name:              AccountType,
+	AccountKeyType.Name:           AccountKeyType,
+	ModelType.Name:                ModelType,
+	SerialType.Name:               SerialType,
+	BaseDeclarationType.Name:      BaseDeclarationType,
+	SnapDeclarationType.Name:      SnapDeclarationType,
+	SnapBuildType.Name:            SnapBuildType,
+	SnapRevisionType.Name:         SnapRevisionType,
+	SnapDeveloperType.Name:        SnapDeveloperType,
+	SystemUserType.Name:           SystemUserType,
+	ValidationType.Name:           ValidationType,
+	ValidationSetType.Name:        ValidationSetType,
+	RepairType.Name:               RepairType,
+	StoreType.Name:                StoreType,
+	PreseedType.Name:              PreseedType,
+	SnapResourceRevisionType.Name: SnapResourceRevisionType,
+	SnapResourcePairType.Name:     SnapResourcePairType,
+	ConfdbType.Name:               ConfdbType,
 	// no authority
 	DeviceSessionRequestType.Name: DeviceSessionRequestType,
 	SerialRequestType.Name:        SerialRequestType,
 	AccountKeyRequestType.Name:    AccountKeyRequestType,
-	PreseedType.Name:              PreseedType,
+	ConfdbControlType.Name:        ConfdbControlType,
 }
 
 // Type returns the AssertionType with name or nil
@@ -198,12 +209,16 @@ func init() {
 	// 3: support for on-store/on-brand/on-model device scope constraints
 	// 4: support for plug-names/slot-names constraints
 	// 5: alt attr matcher usage (was unused before, has new behavior now)
-	maxSupportedFormat[SnapDeclarationType.Name] = 5
+	// 6: support for $PLUG_PUBLISHER_ID/$SLOT_PUBLISHER_ID in attr constraints
+	maxSupportedFormat[SnapDeclarationType.Name] = 6
 
 	// 1: support to limit to device serials
 	// 2: support for user-presence constraint
 	maxSupportedFormat[SystemUserType.Name] = 2
 
+	// 1: support for constraints
+	maxSupportedFormat[AccountKeyType.Name] = 1
+
 	for _, at := range typeRegistry {
 		at.validate()
 	}
@@ -235,6 +250,7 @@ func MockOptionalPrimaryKey(assertType *AssertionType, key, defaultValue string)
 }
 
 var formatAnalyzer = map[*AssertionType]func(headers map[string]interface{}, body []byte) (formatnum int, err error){
+	AccountKeyType:      accountKeyFormatAnalyze,
 	SnapDeclarationType: snapDeclarationFormatAnalyze,
 	SystemUserType:      systemUserFormatAnalyze,
 }
@@ -553,7 +569,7 @@ type SequenceMember interface {
 // customSigner represents an assertion with special arrangements for its signing key (e.g. self-signed), rather than the usual case where an assertion is signed by its authority.
 type customSigner interface {
 	// signKey returns the public key material for the key that signed this assertion.  See also SignKeyID.
-	signKey() PublicKey
+	signKey(db RODatabase) (PublicKey, error)
 }
 
 // MediaType is the media type for encoded assertions on the wire.
@@ -984,6 +1000,34 @@ func checkNoAuthority(assertType *AssertionType, headers map[string]interface{})
 	return nil
 }
 
+func checkJSON(assertType *AssertionType, body []byte) (err error) {
+	defer func() {
+		if err != nil {
+			err = fmt.Errorf("assertion %s: %v", assertType.Name, err)
+		}
+	}()
+
+	if body == nil {
+		return fmt.Errorf(`body must contain JSON`)
+	}
+
+	var val interface{}
+	if err := json.Unmarshal(body, &val); err != nil {
+		return fmt.Errorf("invalid JSON in body: %v", err)
+	}
+
+	formatted, err := json.MarshalIndent(val, "", "  ")
+	if err != nil {
+		return fmt.Errorf("invalid JSON in body: %v", err)
+	}
+
+	if !reflect.DeepEqual(body, formatted) {
+		return fmt.Errorf(`JSON in body must be indented with 2 spaces and sort object entries by key`)
+	}
+
+	return nil
+}
+
 // assemble is the internal variant of Assemble, assumes headers are already checked for supported types
 func assemble(headers map[string]interface{}, body, content, signature []byte) (Assertion, error) {
 	length, err := checkIntWithDefault(headers, "body-length", 0)
@@ -1011,6 +1055,12 @@ func assemble(headers map[string]interface{}, body, content, signature []byte) (
 		return nil, fmt.Errorf("unknown assertion type: %q", typ)
 	}
 
+	if assertType.flags&jsonBody != 0 {
+		if err := checkJSON(assertType, body); err != nil {
+			return nil, err
+		}
+	}
+
 	if assertType.flags&noAuthority == 0 {
 		if err := checkAuthority(assertType, headers); err != nil {
 			return nil, fmt.Errorf("assertion: %v", err)
@@ -1071,6 +1121,7 @@ func assembleAndSign(assertType *AssertionType, headers map[string]interface{},
 	}
 
 	withAuthority := assertType.flags&noAuthority == 0
+	withJSONBody := assertType.flags&jsonBody != 0
 
 	err = checkHeaders(headers)
 	if err != nil {
@@ -1083,6 +1134,12 @@ func assembleAndSign(assertType *AssertionType, headers map[string]interface{},
 		return nil, fmt.Errorf("assertion body is not utf8")
 	}
 
+	if withJSONBody {
+		if err := checkJSON(assertType, body); err != nil {
+			return nil, err
+		}
+	}
+
 	finalHeaders := copyHeaders(headers)
 	bodyLength := len(body)
 	finalBody := make([]byte, bodyLength)
