@@ -51,8 +51,9 @@ func (as *assertsSuite) TestTypeNames(c *C) {
 		"account",
 		"account-key",
 		"account-key-request",
-		// XXX "authority-delegation",
 		"base-declaration",
+		"confdb",
+		"confdb-control",
 		"device-session-request",
 		"model",
 		"preseed",
@@ -62,6 +63,8 @@ func (as *assertsSuite) TestTypeNames(c *C) {
 		"snap-build",
 		"snap-declaration",
 		"snap-developer",
+		"snap-resource-pair",
+		"snap-resource-revision",
 		"snap-revision",
 		"store",
 		"system-user",
@@ -78,12 +81,15 @@ func (as *assertsSuite) TestTypeNames(c *C) {
 }
 
 func (as *assertsSuite) TestMaxSupportedFormats(c *C) {
+	accountKeyMaxFormat := asserts.AccountKeyType.MaxSupportedFormat()
 	snapDeclMaxFormat := asserts.SnapDeclarationType.MaxSupportedFormat()
 	systemUserMaxFormat := asserts.SystemUserType.MaxSupportedFormat()
 	// validity
-	c.Check(snapDeclMaxFormat >= 4, Equals, true)
+	c.Check(accountKeyMaxFormat >= 1, Equals, true)
+	c.Check(snapDeclMaxFormat >= 6, Equals, true)
 	c.Check(systemUserMaxFormat >= 2, Equals, true)
 	c.Check(asserts.MaxSupportedFormats(1), DeepEquals, map[string]int{
+		"account-key":      accountKeyMaxFormat,
 		"snap-declaration": snapDeclMaxFormat,
 		"system-user":      systemUserMaxFormat,
 		"test-only":        1,
@@ -1185,25 +1191,28 @@ func (as *assertsSuite) TestWithAuthority(c *C) {
 	withAuthority := []string{
 		"account",
 		"account-key",
-		// XXX "authority-delegation",
 		"base-declaration",
 		"store",
 		"snap-declaration",
 		"snap-build",
 		"snap-revision",
+		"snap-resource-pair",
+		"snap-resource-revision",
 		"snap-developer",
 		"model",
 		"preseed",
+		"confdb",
 		"serial",
 		"system-user",
 		"validation",
 		"validation-set",
 		"repair",
 	}
-	c.Check(withAuthority, HasLen, asserts.NumAssertionType-3) // excluding device-session-request, serial-request, account-key-request
+	// excluding device-session-request, serial-request, account-key-request, confdb-control
+	c.Check(withAuthority, HasLen, asserts.NumAssertionType-4)
 	for _, name := range withAuthority {
 		typ := asserts.Type(name)
-		_, err := asserts.AssembleAndSignInTest(typ, nil, nil, testPrivKey1)
+		_, err := asserts.AssembleAndSignInTest(typ, nil, []byte("{}"), testPrivKey1)
 		c.Check(err, ErrorMatches, `"authority-id" header is mandatory`)
 	}
 }
