@@ -24,6 +24,7 @@ import (
 	"time"
 
 	"github.com/snapcore/snapd/asserts/internal"
+	"github.com/snapcore/snapd/confdb"
 	"github.com/snapcore/snapd/testutil"
 )
 
@@ -260,6 +261,10 @@ func init() {
 	maxSupportedFormat[TestOnlySeqType.Name] = 2
 }
 
+func (ak *AccountKey) CanSign(a Assertion) bool {
+	return ak.canSign(a)
+}
+
 // AccountKeyIsKeyValidAt exposes isKeyValidAt on AccountKey for tests
 func AccountKeyIsKeyValidAt(ak *AccountKey, when time.Time) bool {
 	return ak.isValidAt(when)
@@ -306,11 +311,12 @@ func (gkm *GPGKeypairManager) ParametersForGenerate(passphrase string, name stri
 
 // constraint tests
 
-func CompileAttrMatcher(constraints interface{}, allowedOperations []string) (func(attrs map[string]interface{}, helper AttrMatchContext) error, error) {
+func CompileAttrMatcher(constraints interface{}, allowedOperations, allowedRefs []string) (func(attrs map[string]interface{}, helper AttrMatchContext) error, error) {
 	// XXX adjust
 	cc := compileContext{
 		opts: &compileAttrMatcherOptions{
 			allowedOperations: allowedOperations,
+			allowedRefs:       allowedRefs,
 		},
 	}
 	matcher, err := compileAttrMatcher(cc, constraints)
@@ -363,3 +369,8 @@ func MockAssertionPrereqs(f func(a Assertion) []*Ref) func() {
 	assertionPrereqs = f
 	return r
 }
+
+// ConfdbControl.operators exposed for tests
+func (cc *ConfdbControl) Operators() map[string]*confdb.Operator {
+	return cc.operators
+}
