@@ -21,8 +21,9 @@ package asserts_test
 
 import (
 	"encoding/base64"
-	"io/ioutil"
+	"os"
 	"path/filepath"
+	"regexp"
 	"sort"
 	"strings"
 	"time"
@@ -714,6 +715,26 @@ func (sds *snapDeclSuite) TestSuggestedFormat(c *C) {
 		c.Assert(err, IsNil)
 		c.Check(fmtnum, Equals, 5)
 	}
+
+	for _, cstr := range []string{"$PLUG_PUBLISHER_ID", "$SLOT_PUBLISHER_ID"} {
+		for _, sidePrefix := range []string{"plug", "slot"} {
+			headers = map[string]interface{}{
+				sidePrefix + "s": map[string]interface{}{
+					"interface6": map[string]interface{}{
+						"allow-auto-connection": map[string]interface{}{
+							sidePrefix + "-attributes": map[string]interface{}{
+								"x": cstr,
+							},
+						},
+					},
+				},
+			}
+
+			fmtnum, err = asserts.SuggestFormat(asserts.SnapDeclarationType, headers, nil)
+			c.Assert(err, IsNil)
+			c.Check(fmtnum, Equals, 6)
+		}
+	}
 }
 
 func prereqDevAccount(c *C, storeDB assertstest.SignerDB, db *asserts.Database) {
@@ -786,7 +807,7 @@ func (s *snapFileDigestSuite) TestSnapFileSHA3_384(c *C) {
 
 	tempdir := c.MkDir()
 	snapFn := filepath.Join(tempdir, "ex.snap")
-	err := ioutil.WriteFile(snapFn, exData, 0644)
+	err := os.WriteFile(snapFn, exData, 0644)
 	c.Assert(err, IsNil)
 
 	encDgst, size, err := asserts.SnapFileSHA3_384(snapFn)
@@ -833,6 +854,7 @@ func (sbs *snapBuildSuite) SetUpSuite(c *C) {
 
 const (
 	blobSHA3_384 = "QlqR0uAWEAWF5Nwnzj5kqmmwFslYPu1IL16MKtLKhwhv0kpBv5wKZ_axf_nf_2cL"
+	hexSHA256    = "e2926364a8b1242d92fb1b56081e1ddb86eba35411961252a103a1c083c2be6d"
 )
 
 func (sbs *snapBuildSuite) TestDecodeOK(c *C) {
@@ -1002,6 +1024,33 @@ func (srs *snapRevSuite) makeValidEncoded() string {
 		"AXNpZw=="
 }
 
+func (srs *snapRevSuite) makeValidEncodedWithIntegrity() string {
+	integrityData := "integrity:\n" +
+		"  -\n" +
+		"    type: dm-verity\n" +
+		"    digest: " + hexSHA256 + "\n" +
+		"    version: 1\n" +
+		"    hash-algorithm: sha256\n" +
+		"    data-block-size: 4096\n" +
+		"    hash-block-size: 4096\n" +
+		"    salt: " + hexSHA256 + "\n"
+
+	return "type: snap-revision\n" +
+		"authority-id: store-id1\n" +
+		"snap-sha3-384: " + blobSHA3_384 + "\n" +
+		"snap-id: snap-id-1\n" +
+		"snap-size: 123\n" +
+		"snap-revision: 1\n" +
+		integrityData +
+		"developer-id: dev-id1\n" +
+		"revision: 1\n" +
+		srs.tsLine +
+		"body-length: 0\n" +
+		"sign-key-sha3-384: Jv8_JiHiIzJVcO9M55pPdqSDWUvuhfDIBJUS-3VW7F_idjix7Ffn5qMxB21ZQuij" +
+		"\n\n" +
+		"AXNpZw=="
+}
+
 func makeSnapRevisionHeaders(overrides map[string]interface{}) map[string]interface{} {
 	headers := map[string]interface{}{
 		"authority-id":  "canonical",
@@ -1058,6 +1107,30 @@ func (srs *snapRevSuite) TestDecodeOKWithProvenance(c *C) {
 	c.Check(snapRev.Provenance(), Equals, "foo")
 }
 
+func (srs *snapRevSuite) TestDecodeOKWithIntegrity(c *C) {
+	encoded := srs.makeValidEncodedWithIntegrity()
+	a, err := asserts.Decode([]byte(encoded))
+	c.Assert(err, IsNil)
+	c.Check(a.Type(), Equals, asserts.SnapRevisionType)
+	snapRev := a.(*asserts.SnapRevision)
+	c.Check(snapRev.AuthorityID(), Equals, "store-id1")
+	c.Check(snapRev.Timestamp(), Equals, srs.ts)
+	c.Check(snapRev.SnapID(), Equals, "snap-id-1")
+	c.Check(snapRev.SnapSHA3_384(), Equals, blobSHA3_384)
+	c.Check(snapRev.SnapSize(), Equals, uint64(123))
+	c.Check(snapRev.SnapRevision(), Equals, 1)
+	c.Check(snapRev.DeveloperID(), Equals, "dev-id1")
+	c.Check(snapRev.Revision(), Equals, 1)
+	c.Check(snapRev.Provenance(), Equals, "global-upload")
+	c.Check(snapRev.SnapIntegrityData()[0].Type, Equals, "dm-verity")
+	c.Check(snapRev.SnapIntegrityData()[0].Version, Equals, uint(1))
+	c.Check(snapRev.SnapIntegrityData()[0].HashAlg, Equals, "sha256")
+	c.Check(snapRev.SnapIntegrityData()[0].DataBlockSize, Equals, uint(4096))
+	c.Check(snapRev.SnapIntegrityData()[0].HashBlockSize, Equals, uint(4096))
+	c.Check(snapRev.SnapIntegrityData()[0].Digest, Equals, hexSHA256)
+	c.Check(snapRev.SnapIntegrityData()[0].Salt, Equals, hexSHA256)
+}
+
 const (
 	snapRevErrPrefix = "assertion snap-revision: "
 )
@@ -1098,6 +1171,62 @@ func (srs *snapRevSuite) TestDecodeInvalid(c *C) {
 	}
 }
 
+func (srs *snapRevSuite) TestDecodeInvalidWithIntegrity(c *C) {
+	encoded := srs.makeValidEncodedWithIntegrity()
+
+	integrityHdr := "integrity:\n" +
+		"  -\n" +
+		"    type: dm-verity\n" +
+		"    digest: " + hexSHA256 + "\n" +
+		"    version: 1\n" +
+		"    hash-algorithm: sha256\n" +
+		"    data-block-size: 4096\n" +
+		"    hash-block-size: 4096\n" +
+		"    salt: " + hexSHA256 + "\n"
+
+	integrityTypeHdr := "    type: dm-verity\n"
+	integrityVersionHdr := "    version: 1\n"
+	integrityHashAlgHdr := "    hash-algorithm: sha256\n"
+	integrityDataBlockSizeHdr := "    data-block-size: 4096\n"
+	integrityHashBlockSizeHdr := "    hash-block-size: 4096\n"
+	integrityDigestHdr := "    digest: " + hexSHA256 + "\n"
+	integritySaltHdr := "    salt: " + hexSHA256 + "\n"
+
+	invalidTests := []struct {
+		original,
+		invalid,
+		expectedErr string
+	}{
+		{integrityHdr, "integrity: test\n", `"integrity" header must contain a list of integrity data`},
+		{integrityTypeHdr, "", `"type" of integrity data \[0\] is mandatory`},
+		{integrityTypeHdr, "    type: foo\n", `"type" of integrity data \[0\] must be one of \(dm-verity\)`},
+		{integrityVersionHdr, "", `"version" of integrity data \[0\] of type "dm-verity" is mandatory`},
+		{integrityVersionHdr, "    version: a\n", `"version" of integrity data \[0\] of type "dm-verity" is not an unsigned integer: a`},
+		{integrityVersionHdr, "    version: 2\n", `version of integrity data \[0\] of type "dm-verity" must be one of ` + regexp.QuoteMeta("[1]")},
+		{integrityHashAlgHdr, "", `"hash-algorithm" of integrity data \[0\] of type "dm-verity" is mandatory`},
+		{integrityHashAlgHdr, "    hash-algorithm: 0\n", `hash algorithm of integrity data \[0\] of type "dm-verity" must be one of .*`},
+		{integrityHashAlgHdr, "    hash-algorithm: a\n", `hash algorithm of integrity data \[0\] of type "dm-verity" must be one of .*`},
+		{integrityHashAlgHdr, "    hash-algorithm: sha384\n", `hash algorithm of integrity data \[0\] of type "dm-verity" must be one of .*`},
+		{integrityHashAlgHdr, "    hash-algorithm: sm3\n", `hash algorithm of integrity data \[0\] of type "dm-verity" must be one of .*`},
+		{integrityDataBlockSizeHdr, "", `"data-block-size" of integrity data \[0\] of type "dm-verity" \(sha256\) is mandatory`},
+		{integrityDataBlockSizeHdr, "    data-block-size: a\n", `"data-block-size" of integrity data \[0\] of type "dm-verity" \(sha256\) is not an unsigned integer: a`},
+		{integrityHashBlockSizeHdr, "", `"hash-block-size" of integrity data \[0\] of type "dm-verity" \(sha256\) is mandatory`},
+		{integrityHashBlockSizeHdr, "    hash-block-size: a\n", `"hash-block-size" of integrity data \[0\] of type "dm-verity" \(sha256\) is not an unsigned integer: a`},
+		{integrityDigestHdr, "", `"digest" of integrity data \[0\] of type "dm-verity" \(sha256\) is mandatory`},
+		{integrityDigestHdr, "    digest: a\n", `"digest" of integrity data \[0\] of type "dm-verity" \(sha256\) cannot be decoded: encoding/hex: odd length hex string`},
+		{integrityDigestHdr, "    digest: ab\n", `"digest" of integrity data \[0\] of type "dm-verity" \(sha256\) does not have the expected bit length: 8`},
+		{integritySaltHdr, "", `"salt" of integrity data \[0\] of type "dm-verity" \(sha256\) is mandatory`},
+		{integritySaltHdr, "    salt: a\n", `"salt" of integrity data \[0\] of type "dm-verity" \(sha256\) cannot be decoded: encoding/hex: odd length hex string`},
+		{integritySaltHdr, "    salt: ab\n", `"salt" of integrity data \[0\] of type "dm-verity" \(sha256\) does not have the expected bit length: 8`},
+	}
+
+	for _, test := range invalidTests {
+		invalid := strings.Replace(encoded, test.original, test.invalid, 1)
+		_, err := asserts.Decode([]byte(invalid))
+		c.Check(err, ErrorMatches, snapRevErrPrefix+test.expectedErr)
+	}
+}
+
 func prereqSnapDecl(c *C, storeDB assertstest.SignerDB, db *asserts.Database) {
 	snapDecl, err := storeDB.Sign(asserts.SnapDeclarationType, map[string]interface{}{
 		"series":       "16",
@@ -1456,50 +1585,6 @@ func (srs *snapRevSuite) TestSnapRevisionDelegationRevisionOutOfRange(c *C) {
 	c.Check(err, ErrorMatches, `snap-revision assertion with provenance "prov1" for snap id "snap-id-1" is not signed by an authorized authority: delegated-id`)
 }
 
-func (srs *snapRevSuite) TestSnapRevisionDelegationInconsistentTimestamp(c *C) {
-	c.Skip("authority-delegation disabled")
-
-	storeDB, db := makeStoreAndCheckDB(c)
-
-	prereqDevAccount(c, storeDB, db)
-	prereqSnapDecl(c, storeDB, db)
-
-	otherDB := setup3rdPartySigning(c, "other", storeDB, db)
-
-	headers := srs.makeHeaders(map[string]interface{}{
-		"authority-id": "canonical",
-		"signatory-id": "other",
-		"timestamp":    time.Now().Format(time.RFC3339),
-	})
-	snapRev, err := otherDB.Sign(asserts.SnapRevisionType, headers, nil, "")
-	c.Assert(err, IsNil)
-
-	// move forward in time
-	time.Sleep(100 * time.Millisecond)
-
-	// now add authority-delegation
-	headers = map[string]interface{}{
-		"authority-id": "canonical",
-		"account-id":   "canonical",
-		"delegate-id":  "other",
-		"assertions": []interface{}{
-			map[string]interface{}{
-				"type": "snap-revision",
-				"headers": map[string]interface{}{
-					"snap-id": "snap-id-1",
-				},
-				"since": time.Now().Format(time.RFC3339Nano),
-			},
-		},
-	}
-	ad, err := storeDB.Sign(asserts.AuthorityDelegationType, headers, nil, "")
-	c.Assert(err, IsNil)
-	c.Check(db.Add(ad), IsNil)
-
-	err = db.Check(snapRev)
-	c.Check(err, ErrorMatches, `delegated snap-revision assertion from "canonical" to "other" timestamp ".*" is outside of all supporting delegation constraints validity`)
-}
-
 func (srs *snapRevSuite) TestPrimaryKey(c *C) {
 	storeDB, db := makeStoreAndCheckDB(c)
 
