@@ -24,9 +24,17 @@ import (
 	"fmt"
 
 	"github.com/snapcore/snapd/bootloader"
+	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/snap"
 )
 
+// Unlocker functions are passed from code using boot to indicate that global
+// state should be unlocked during slow operations, e.g sealing/unsealing.
+// Boot code is then expected to call the unlocker around the slow section and
+// relock using the returned function. Unlocker being nil indicates not to do
+// this.
+type Unlocker func() (relock func())
+
 const (
 	// DefaultStatus is the value of a status boot variable when nothing is
 	// being tried
@@ -41,13 +49,13 @@ const (
 )
 
 // RebootInfo contains information about how to perform a reboot if
-// required
+// required.
 type RebootInfo struct {
-	// RebootRequired is true if we need to reboot after an update
+	// RebootRequired is true if we need to reboot after an update.
 	RebootRequired bool
-	// RebootBootloader will not be nil if the bootloader has something to say on
-	// how to perform the reboot
-	RebootBootloader bootloader.RebootBootloader
+	// BootloaderOptions will be used to find the correct bootloader when
+	// checking for any set reboot arguments.
+	BootloaderOptions *bootloader.Options
 }
 
 // NextBootContext carries additional significative information used when
@@ -141,8 +149,25 @@ func Kernel(s snap.PlaceInfo, t snap.Type, dev snap.Device) BootKernel {
 	return trivial{}
 }
 
+// SnapTypeParticipatesInBoot returns whether a snap type participates in the
+// boot for a given device.
+func SnapTypeParticipatesInBoot(t snap.Type, dev snap.Device) bool {
+	if dev.IsClassicBoot() {
+		return false
+	}
+	switch t {
+	case snap.TypeBase, snap.TypeOS:
+		// Bases are not boot participants for classic with modes
+		return !dev.Classic()
+	case snap.TypeKernel, snap.TypeGadget:
+		return true
+	}
+
+	return false
+}
+
 func applicable(s snap.PlaceInfo, t snap.Type, dev snap.Device) bool {
-	if dev.Classic() {
+	if !SnapTypeParticipatesInBoot(t, dev) {
 		return false
 	}
 	// In ephemeral modes we never need to care about updating the boot
@@ -251,18 +276,14 @@ func fixedInUse(inUse bool) InUseFunc {
 // InUse returns a checker for whether a given name/revision is used in the
 // boot environment for snaps of the relevant snap type.
 func InUse(typ snap.Type, dev snap.Device) (InUseFunc, error) {
-	if dev.Classic() {
-		// no boot state on classic
-		return fixedInUse(false), nil
-	}
+	modeenvLock()
+	defer modeenvUnlock()
+
 	if !dev.RunMode() {
 		// ephemeral mode, block manipulations for now
 		return fixedInUse(true), nil
 	}
-	switch typ {
-	case snap.TypeKernel, snap.TypeBase, snap.TypeOS:
-		break
-	default:
+	if !SnapTypeParticipatesInBoot(typ, dev) || typ == snap.TypeGadget {
 		return fixedInUse(false), nil
 	}
 	cands := make([]snap.PlaceInfo, 0, 2)
@@ -299,6 +320,9 @@ var (
 // type of snap, which can be snap.TypeBase (or snap.TypeOS), or snap.TypeKernel.
 // Returns ErrBootNameAndRevisionNotReady if the values are temporarily not established.
 func GetCurrentBoot(t snap.Type, dev snap.Device) (snap.PlaceInfo, error) {
+	modeenvLock()
+	defer modeenvUnlock()
+
 	s, err := bootStateFor(t, dev)
 	if err != nil {
 		return nil, err
@@ -327,26 +351,32 @@ type bootStateUpdate interface {
 // target for rollback.
 //
 // The states that a boot goes through for UC16/18 are the following:
-// - By default snap_mode is "" in which case the bootloader loads
-//   two squashfs'es denoted by variables snap_core and snap_kernel.
-// - On a refresh of core/kernel snapd will set snap_mode=try and
-//   will also set snap_try_{core,kernel} to the core/kernel that
-//   will be tried next.
-// - On reboot the bootloader will inspect the snap_mode and if the
-//   mode is set to "try" it will set "snap_mode=trying" and then
-//   try to boot the snap_try_{core,kernel}".
-// - On a successful boot snapd resets snap_mode to "" and copies
-//   snap_try_{core,kernel} to snap_{core,kernel}. The snap_try_*
-//   values are cleared afterwards.
-// - On a failing boot the bootloader will see snap_mode=trying which
-//   means snapd did not start successfully. In this case the bootloader
-//   will set snap_mode="" and the system will boot with the known good
-//   values from snap_{core,kernel}
+//   - By default snap_mode is "" in which case the bootloader loads
+//     two squashfs'es denoted by variables snap_core and snap_kernel.
+//   - On a refresh of core/kernel snapd will set snap_mode=try and
+//     will also set snap_try_{core,kernel} to the core/kernel that
+//     will be tried next.
+//   - On reboot the bootloader will inspect the snap_mode and if the
+//     mode is set to "try" it will set "snap_mode=trying" and then
+//     try to boot the snap_try_{core,kernel}".
+//   - On a successful boot snapd resets snap_mode to "" and copies
+//     snap_try_{core,kernel} to snap_{core,kernel}. The snap_try_*
+//     values are cleared afterwards.
+//   - On a failing boot the bootloader will see snap_mode=trying which
+//     means snapd did not start successfully. In this case the bootloader
+//     will set snap_mode="" and the system will boot with the known good
+//     values from snap_{core,kernel}
 func MarkBootSuccessful(dev snap.Device) error {
+	modeenvLock()
+	defer modeenvUnlock()
+
 	const errPrefix = "cannot mark boot successful: %s"
 
 	var u bootStateUpdate
 	for _, t := range []snap.Type{snap.TypeBase, snap.TypeKernel} {
+		if !SnapTypeParticipatesInBoot(t, dev) {
+			continue
+		}
 		s, err := bootStateFor(t, dev)
 		if err != nil {
 			return err
@@ -416,10 +446,12 @@ func SetRecoveryBootSystemAndMode(dev snap.Device, systemLabel, mode string) err
 	return bl.SetBootVars(m)
 }
 
-// UpdateManagedBootConfigs updates managed boot config assets if those are
-// present for the ubuntu-boot bootloader. Returns true when an update was
-// carried out.
-func UpdateManagedBootConfigs(dev snap.Device, gadgetSnapOrDir string) (updated bool, err error) {
+// UpdateManagedBootConfigs updates managed boot config assets if
+// those are present for the ubuntu-boot bootloader. To do this it
+// needs information from the model, the gadget we are updating to,
+// and any additional kernel command line arguments coming from system
+// options. Returns true when an update was carried out.
+func UpdateManagedBootConfigs(dev snap.Device, gadgetSnapOrDir, cmdlineAppend string) (updated bool, err error) {
 	if !dev.HasModeenv() {
 		// only UC20 devices use managed boot config
 		return false, nil
@@ -427,10 +459,31 @@ func UpdateManagedBootConfigs(dev snap.Device, gadgetSnapOrDir string) (updated
 	if !dev.RunMode() {
 		return false, fmt.Errorf("internal error: boot config can only be updated in run mode")
 	}
-	return updateManagedBootConfigForBootloader(dev, ModeRun, gadgetSnapOrDir)
+	modeenvLock()
+	defer modeenvUnlock()
+
+	return updateManagedBootConfigForBootloader(dev, ModeRun, gadgetSnapOrDir, cmdlineAppend)
 }
 
-func updateManagedBootConfigForBootloader(dev snap.Device, mode, gadgetSnapOrDir string) (updated bool, err error) {
+func updateCmdlineVars(tbl bootloader.TrustedAssetsBootloader, gadgetSnapOrDir, cmdlineAppend string, candidate bool, dev snap.Device) error {
+	defaultCmdLine, err := tbl.DefaultCommandLine(candidate)
+	if err != nil {
+		return err
+	}
+
+	cmdlineVars, err := bootVarsForTrustedCommandLineFromGadget(gadgetSnapOrDir, cmdlineAppend, defaultCmdLine, dev.Model())
+	if err != nil {
+		return fmt.Errorf("cannot prepare bootloader variables for kernel command line: %v", err)
+	}
+
+	if err := tbl.SetBootVars(cmdlineVars); err != nil {
+		return fmt.Errorf("cannot set run system kernel command line arguments: %v", err)
+	}
+
+	return nil
+}
+
+func updateManagedBootConfigForBootloader(dev snap.Device, mode, gadgetSnapOrDir, cmdlineAppend string) (updated bool, err error) {
 	if mode != ModeRun {
 		return false, fmt.Errorf("internal error: updating boot config of recovery bootloader is not supported yet")
 	}
@@ -447,23 +500,44 @@ func updateManagedBootConfigForBootloader(dev snap.Device, mode, gadgetSnapOrDir
 		}
 		return false, err
 	}
+
 	// boot config update can lead to a change of kernel command line
-	_, err = observeCommandLineUpdate(dev.Model(), commandLineUpdateReasonSnapd, gadgetSnapOrDir)
+	cmdlineChange, err := observeCommandLineUpdate(dev.Model(), commandLineUpdateReasonSnapd, gadgetSnapOrDir, cmdlineAppend)
 	if err != nil {
 		return false, err
 	}
-	return tbl.UpdateBootConfig()
+
+	osutil.MaybeInjectFault("update-config-bootloader")
+
+	if cmdlineChange {
+		candidate := true
+		if err := updateCmdlineVars(tbl, gadgetSnapOrDir, cmdlineAppend, candidate, dev); err != nil {
+			return false, err
+		}
+	}
+
+	assetChange, err := tbl.UpdateBootConfig()
+	if err != nil {
+		return false, err
+	}
+
+	return assetChange || cmdlineChange, nil
 }
 
-// UpdateCommandLineForGadgetComponent handles the update of a gadget that
-// contributes to the kernel command line of the run system. Returns true when a
-// change in command line has been observed and a reboot is needed. The reboot,
-// if needed, should be requested at the the earliest possible occasion.
-func UpdateCommandLineForGadgetComponent(dev snap.Device, gadgetSnapOrDir string) (needsReboot bool, err error) {
+// UpdateCommandLineForGadgetComponent handles the update of a gadget
+// that contributes to the kernel command line of the run system
+// (appending any additional kernel command line arguments coming from
+// system options). Returns true when a change in command line has
+// been observed and a reboot is needed. The reboot, if needed, should
+// be requested at the the earliest possible occasion.
+func UpdateCommandLineForGadgetComponent(dev snap.Device, gadgetSnapOrDir, cmdlineAppend string) (needsReboot bool, err error) {
 	if !dev.HasModeenv() {
 		// only UC20 devices are supported
 		return false, fmt.Errorf("internal error: command line component cannot be updated on pre-UC20 devices")
 	}
+	modeenvLock()
+	defer modeenvUnlock()
+
 	opts := &bootloader.Options{
 		Role: bootloader.RoleRunMode,
 	}
@@ -477,35 +551,22 @@ func UpdateCommandLineForGadgetComponent(dev snap.Device, gadgetSnapOrDir string
 		}
 		return false, err
 	}
+
 	// gadget update can lead to a change of kernel command line
-	cmdlineChange, err := observeCommandLineUpdate(dev.Model(), commandLineUpdateReasonGadget, gadgetSnapOrDir)
+	cmdlineChange, err := observeCommandLineUpdate(dev.Model(), commandLineUpdateReasonGadget, gadgetSnapOrDir, cmdlineAppend)
 	if err != nil {
 		return false, err
 	}
+
 	if !cmdlineChange {
 		return false, nil
 	}
-	// update the bootloader environment, maybe clearing the relevant
-	// variables
-	cmdlineVars, err := bootVarsForTrustedCommandLineFromGadget(gadgetSnapOrDir)
-	if err != nil {
-		return false, fmt.Errorf("cannot prepare bootloader variables for kernel command line: %v", err)
-	}
-	if err := tbl.SetBootVars(cmdlineVars); err != nil {
-		return false, fmt.Errorf("cannot set run system kernel command line arguments: %v", err)
+
+	osutil.MaybeInjectFault("update-command-line-gadget")
+
+	candidate := false
+	if err := updateCmdlineVars(tbl, gadgetSnapOrDir, cmdlineAppend, candidate, dev); err != nil {
+		return false, err
 	}
 	return cmdlineChange, nil
 }
-
-// MarkFactoryResetComplete runs a series of steps in a run system that complete a
-// factory reset process.
-func MarkFactoryResetComplete(encrypted bool) error {
-	if !encrypted {
-		// there is nothing to do on an unencrypted system
-		return nil
-	}
-	if err := postFactoryResetCleanup(); err != nil {
-		return fmt.Errorf("cannot perform post factory reset boot cleanup: %v", err)
-	}
-	return nil
-}
