@@ -22,7 +22,6 @@ package main_test
 import (
 	"bytes"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 	"testing"
@@ -33,6 +32,7 @@ import (
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/osutil"
+	"github.com/snapcore/snapd/sandbox/cgroup"
 	"github.com/snapcore/snapd/testutil"
 )
 
@@ -50,14 +50,20 @@ func (s *mainSuite) SetUpTest(c *C) {
 	s.BaseTest.SetUpTest(c)
 	s.as = &update.Assumptions{}
 	buf, restore := logger.MockLogger()
-	s.BaseTest.AddCleanup(restore)
+	s.AddCleanup(restore)
 	s.log = buf
+	s.AddCleanup(cgroup.MockVersion(cgroup.V1, nil))
 }
 
-func (s *mainSuite) TestComputeAndSaveChanges(c *C) {
+func (s *mainSuite) TestExecuteMountProfileUpdate(c *C) {
 	dirs.SetRootDir(c.MkDir())
 	defer dirs.SetRootDir("/")
 
+	// mount targets look at the the actual local filesystem
+	if !osutil.IsDirectory("/usr/share/fonts") || !osutil.IsDirectory("/usr/local/share/fonts") {
+		c.Skip("missing local directories (/usr/share/fonts or /usr/local/share/fonts)")
+	}
+
 	restore := update.MockChangePerform(func(chg *update.Change, as *update.Assumptions) ([]*update.Change, error) {
 		return nil, nil
 	})
@@ -70,16 +76,17 @@ func (s *mainSuite) TestComputeAndSaveChanges(c *C) {
 	desiredProfilePath := fmt.Sprintf("%s/snap.%s.fstab", dirs.SnapMountPolicyDir, snapName)
 	err := os.MkdirAll(filepath.Dir(desiredProfilePath), 0755)
 	c.Assert(err, IsNil)
-	err = ioutil.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644)
+	err = os.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644)
 	c.Assert(err, IsNil)
 
 	currentProfilePath := fmt.Sprintf("%s/snap.%s.fstab", dirs.SnapRunNsDir, snapName)
 	err = os.MkdirAll(filepath.Dir(currentProfilePath), 0755)
 	c.Assert(err, IsNil)
-	err = ioutil.WriteFile(currentProfilePath, nil, 0644)
+	err = os.WriteFile(currentProfilePath, nil, 0644)
 	c.Assert(err, IsNil)
 
-	err = update.ComputeAndSaveChanges(snapName, s.as)
+	upCtx := update.NewSystemProfileUpdateContext(snapName, false)
+	err = update.ExecuteMountProfileUpdate(upCtx)
 	c.Assert(err, IsNil)
 
 	c.Check(currentProfilePath, testutil.FileEquals, `/var/lib/snapd/hostfs/usr/local/share/fonts /usr/local/share/fonts none bind,ro 0 0
@@ -102,8 +109,8 @@ func (s *mainSuite) TestAddingSyntheticChanges(c *C) {
 
 	c.Assert(os.MkdirAll(filepath.Dir(currentProfilePath), 0755), IsNil)
 	c.Assert(os.MkdirAll(filepath.Dir(desiredProfilePath), 0755), IsNil)
-	c.Assert(ioutil.WriteFile(currentProfilePath, []byte(currentProfileContent), 0644), IsNil)
-	c.Assert(ioutil.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644), IsNil)
+	c.Assert(os.WriteFile(currentProfilePath, []byte(currentProfileContent), 0644), IsNil)
+	c.Assert(os.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644), IsNil)
 
 	// In order to make that work, /usr/share had to be converted to a writable
 	// mimic. Some actions were performed under the hood and now we see a
@@ -146,7 +153,8 @@ func (s *mainSuite) TestAddingSyntheticChanges(c *C) {
 	})
 	defer restore()
 
-	c.Assert(update.ComputeAndSaveChanges(snapName, s.as), IsNil)
+	upCtx := update.NewSystemProfileUpdateContext(snapName, false)
+	c.Assert(update.ExecuteMountProfileUpdate(upCtx), IsNil)
 
 	c.Check(currentProfilePath, testutil.FileEquals,
 		`tmpfs /usr/share tmpfs x-snapd.synthetic,x-snapd.needed-by=/usr/share/mysnap 0 0
@@ -176,8 +184,8 @@ func (s *mainSuite) TestRemovingSyntheticChanges(c *C) {
 
 	c.Assert(os.MkdirAll(filepath.Dir(currentProfilePath), 0755), IsNil)
 	c.Assert(os.MkdirAll(filepath.Dir(desiredProfilePath), 0755), IsNil)
-	c.Assert(ioutil.WriteFile(currentProfilePath, []byte(currentProfileContent), 0644), IsNil)
-	c.Assert(ioutil.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644), IsNil)
+	c.Assert(os.WriteFile(currentProfilePath, []byte(currentProfileContent), 0644), IsNil)
+	c.Assert(os.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644), IsNil)
 
 	n := -1
 	restore := update.MockChangePerform(func(chg *update.Change, as *update.Assumptions) ([]*update.Change, error) {
@@ -189,31 +197,31 @@ func (s *mainSuite) TestRemovingSyntheticChanges(c *C) {
 				Entry: osutil.MountEntry{
 					Name: "/snap/mysnap/42/usr/share/mysnap",
 					Dir:  "/usr/share/mysnap", Type: "none",
-					Options: []string{"bind", "ro"},
+					Options: []string{"bind", "ro", "x-snapd.detach"},
 				},
 			})
 		case 1:
-			c.Assert(chg, DeepEquals, &update.Change{
+			c.Check(chg, DeepEquals, &update.Change{
 				Action: update.Unmount,
 				Entry: osutil.MountEntry{
 					Name: "/usr/share/awk", Dir: "/usr/share/awk", Type: "none",
-					Options: []string{"bind", "ro", "x-snapd.synthetic", "x-snapd.needed-by=/usr/share/mysnap"},
+					Options: []string{"bind", "ro", "x-snapd.synthetic", "x-snapd.needed-by=/usr/share/mysnap", "x-snapd.detach"},
 				},
 			})
 		case 2:
-			c.Assert(chg, DeepEquals, &update.Change{
+			c.Check(chg, DeepEquals, &update.Change{
 				Action: update.Unmount,
 				Entry: osutil.MountEntry{
 					Name: "/usr/share/adduser", Dir: "/usr/share/adduser", Type: "none",
-					Options: []string{"bind", "ro", "x-snapd.synthetic", "x-snapd.needed-by=/usr/share/mysnap"},
+					Options: []string{"bind", "ro", "x-snapd.synthetic", "x-snapd.needed-by=/usr/share/mysnap", "x-snapd.detach"},
 				},
 			})
 		case 3:
-			c.Assert(chg, DeepEquals, &update.Change{
+			c.Check(chg, DeepEquals, &update.Change{
 				Action: update.Unmount,
 				Entry: osutil.MountEntry{
 					Name: "tmpfs", Dir: "/usr/share", Type: "tmpfs",
-					Options: []string{"x-snapd.synthetic", "x-snapd.needed-by=/usr/share/mysnap"},
+					Options: []string{"x-snapd.synthetic", "x-snapd.needed-by=/usr/share/mysnap", "x-snapd.detach"},
 				},
 			})
 		default:
@@ -223,7 +231,8 @@ func (s *mainSuite) TestRemovingSyntheticChanges(c *C) {
 	})
 	defer restore()
 
-	c.Assert(update.ComputeAndSaveChanges(snapName, s.as), IsNil)
+	upCtx := update.NewSystemProfileUpdateContext(snapName, false)
+	c.Assert(update.ExecuteMountProfileUpdate(upCtx), IsNil)
 
 	c.Check(currentProfilePath, testutil.FileEquals, "")
 }
@@ -241,8 +250,8 @@ func (s *mainSuite) TestApplyingLayoutChanges(c *C) {
 
 	c.Assert(os.MkdirAll(filepath.Dir(currentProfilePath), 0755), IsNil)
 	c.Assert(os.MkdirAll(filepath.Dir(desiredProfilePath), 0755), IsNil)
-	c.Assert(ioutil.WriteFile(currentProfilePath, []byte(currentProfileContent), 0644), IsNil)
-	c.Assert(ioutil.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644), IsNil)
+	c.Assert(os.WriteFile(currentProfilePath, []byte(currentProfileContent), 0644), IsNil)
+	c.Assert(os.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644), IsNil)
 
 	n := -1
 	restore := update.MockChangePerform(func(chg *update.Change, as *update.Assumptions) ([]*update.Change, error) {
@@ -265,7 +274,8 @@ func (s *mainSuite) TestApplyingLayoutChanges(c *C) {
 	defer restore()
 
 	// The error was not ignored, we bailed out.
-	c.Assert(update.ComputeAndSaveChanges(snapName, s.as), ErrorMatches, "testing")
+	upCtx := update.NewSystemProfileUpdateContext(snapName, false)
+	c.Assert(update.ExecuteMountProfileUpdate(upCtx), ErrorMatches, "testing")
 
 	c.Check(currentProfilePath, testutil.FileEquals, "")
 }
@@ -283,8 +293,8 @@ func (s *mainSuite) TestApplyingParallelInstanceChanges(c *C) {
 
 	c.Assert(os.MkdirAll(filepath.Dir(currentProfilePath), 0755), IsNil)
 	c.Assert(os.MkdirAll(filepath.Dir(desiredProfilePath), 0755), IsNil)
-	c.Assert(ioutil.WriteFile(currentProfilePath, []byte(currentProfileContent), 0644), IsNil)
-	c.Assert(ioutil.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644), IsNil)
+	c.Assert(os.WriteFile(currentProfilePath, []byte(currentProfileContent), 0644), IsNil)
+	c.Assert(os.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644), IsNil)
 
 	n := -1
 	restore := update.MockChangePerform(func(chg *update.Change, as *update.Assumptions) ([]*update.Change, error) {
@@ -307,7 +317,8 @@ func (s *mainSuite) TestApplyingParallelInstanceChanges(c *C) {
 	defer restore()
 
 	// The error was not ignored, we bailed out.
-	c.Assert(update.ComputeAndSaveChanges(snapName, nil), ErrorMatches, "testing")
+	upCtx := update.NewSystemProfileUpdateContext(snapName, false)
+	c.Assert(update.ExecuteMountProfileUpdate(upCtx), ErrorMatches, "testing")
 
 	c.Check(currentProfilePath, testutil.FileEquals, "")
 }
@@ -325,8 +336,8 @@ func (s *mainSuite) TestApplyIgnoredMissingMount(c *C) {
 
 	c.Assert(os.MkdirAll(filepath.Dir(currentProfilePath), 0755), IsNil)
 	c.Assert(os.MkdirAll(filepath.Dir(desiredProfilePath), 0755), IsNil)
-	c.Assert(ioutil.WriteFile(currentProfilePath, []byte(currentProfileContent), 0644), IsNil)
-	c.Assert(ioutil.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644), IsNil)
+	c.Assert(os.WriteFile(currentProfilePath, []byte(currentProfileContent), 0644), IsNil)
+	c.Assert(os.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644), IsNil)
 
 	n := -1
 	restore := update.MockChangePerform(func(chg *update.Change, as *update.Assumptions) ([]*update.Change, error) {
@@ -350,12 +361,13 @@ func (s *mainSuite) TestApplyIgnoredMissingMount(c *C) {
 	defer restore()
 
 	// The error was ignored, and no mount was recorded in the profile
-	c.Assert(update.ComputeAndSaveChanges(snapName, s.as), IsNil)
+	upCtx := update.NewSystemProfileUpdateContext(snapName, false)
+	c.Assert(update.ExecuteMountProfileUpdate(upCtx), IsNil)
 	c.Check(s.log.String(), Equals, "")
 	c.Check(currentProfilePath, testutil.FileEquals, "")
 }
 
-func (s *mainSuite) TestApplyUserFstab(c *C) {
+func (s *mainSuite) TestApplyUserFstabHomeRequiredAndValid(c *C) {
 	dirs.SetRootDir(c.MkDir())
 	defer dirs.SetRootDir("/")
 
@@ -367,21 +379,58 @@ func (s *mainSuite) TestApplyUserFstab(c *C) {
 	defer restore()
 
 	snapName := "foo"
-	desiredProfileContent := `$XDG_RUNTIME_DIR/doc/by-app/snap.foo $XDG_RUNTIME_DIR/doc none bind,rw 0 0`
-
+	desiredProfileContent := `$XDG_RUNTIME_DIR/doc/by-app/snap.foo $XDG_RUNTIME_DIR/doc none bind,rw 0 0
+none $HOME/.local/share none x-snapd.kind=ensure-dir,x-snapd.must-exist-dir=$HOME 0 0`
 	desiredProfilePath := fmt.Sprintf("%s/snap.%s.user-fstab", dirs.SnapMountPolicyDir, snapName)
 	err := os.MkdirAll(filepath.Dir(desiredProfilePath), 0755)
 	c.Assert(err, IsNil)
-	err = ioutil.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644)
+	err = os.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644)
 	c.Assert(err, IsNil)
 
-	err = update.ApplyUserFstab("foo")
+	tmpHomeDir := c.MkDir()
+	restoreEnv := update.MockSnapConfineUserEnv("/run/user/1000/snap.snapname", tmpHomeDir)
+	defer restoreEnv()
+	upCtx, err := update.NewUserProfileUpdateContext(snapName, true, 1000)
 	c.Assert(err, IsNil)
+	err = update.ExecuteMountProfileUpdate(upCtx)
+	c.Assert(err, IsNil)
+	c.Assert(changes, HasLen, 2)
 
-	xdgRuntimeDir := fmt.Sprintf("%s/%d", dirs.XdgRuntimeDirBase, os.Getuid())
+	c.Assert(changes[0].Entry.Name, Equals, "none")
+	c.Assert(changes[0].Entry.Dir, Equals, tmpHomeDir+"/.local/share")
+	c.Assert(changes[0].Entry.XSnapdMustExistDir(), Equals, tmpHomeDir)
 
-	c.Assert(changes, HasLen, 1)
-	c.Assert(changes[0].Action, Equals, update.Mount)
-	c.Assert(changes[0].Entry.Name, Equals, xdgRuntimeDir+"/doc/by-app/snap.foo")
-	c.Assert(changes[0].Entry.Dir, Matches, xdgRuntimeDir+"/doc")
+	xdgRuntimeDir := fmt.Sprintf("%s/%d", dirs.XdgRuntimeDirBase, 1000)
+	c.Assert(changes[1].Action, Equals, update.Mount)
+	c.Assert(changes[1].Entry.Name, Equals, xdgRuntimeDir+"/doc/by-app/snap.foo")
+	c.Assert(changes[1].Entry.Dir, Matches, xdgRuntimeDir+"/doc")
+}
+
+func (s *mainSuite) TestApplyUserFstabErrorHomeRequiredAndMissing(c *C) {
+	dirs.SetRootDir(c.MkDir())
+	defer dirs.SetRootDir("/")
+	var changes []update.Change
+	restore := update.MockChangePerform(func(chg *update.Change, as *update.Assumptions) ([]*update.Change, error) {
+		changes = append(changes, *chg)
+		return nil, nil
+	})
+	defer restore()
+
+	snapName := "foo"
+	desiredProfileContent := `$XDG_RUNTIME_DIR/doc/by-app/snap.foo $XDG_RUNTIME_DIR/doc none bind,rw 0 0
+none $HOME/.local/share none x-snapd.kind=ensure-dir,x-snapd.must-exist-dir=$HOME 0 0`
+	desiredProfilePath := fmt.Sprintf("%s/snap.%s.user-fstab", dirs.SnapMountPolicyDir, snapName)
+	err := os.MkdirAll(filepath.Dir(desiredProfilePath), 0755)
+	c.Assert(err, IsNil)
+	err = os.WriteFile(desiredProfilePath, []byte(desiredProfileContent), 0644)
+	c.Assert(err, IsNil)
+
+	tmpHomeDir := c.MkDir() + "/does-not-exist"
+	restoreEnv := update.MockSnapConfineUserEnv("/run/user/1000/snap.snapname", tmpHomeDir)
+	defer restoreEnv()
+	upCtx, err := update.NewUserProfileUpdateContext(snapName, true, 1000)
+	c.Assert(err, IsNil)
+	err = update.ExecuteMountProfileUpdate(upCtx)
+	c.Assert(err, ErrorMatches, `cannot expand mount entry \(none \$HOME/.local/share none x-snapd.kind=ensure-dir,x-snapd.must-exist-dir=\$HOME 0 0\): cannot use invalid home directory `+fmt.Sprintf("\"%s\"", tmpHomeDir)+": no such file or directory")
+	c.Assert(changes, HasLen, 0)
 }
