@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2014-2021 Canonical Ltd
+ * Copyright (C) 2014-2023 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -29,7 +29,9 @@ import (
 
 	"github.com/snapcore/snapd/i18n"
 	"github.com/snapcore/snapd/image"
+	"github.com/snapcore/snapd/interfaces/builtin"
 	"github.com/snapcore/snapd/seed/seedwriter"
+	"github.com/snapcore/snapd/snap"
 )
 
 type cmdPrepareImage struct {
@@ -52,10 +54,13 @@ type cmdPrepareImage struct {
 	Customize string `long:"customize" hidden:"yes"`
 
 	// TODO: introduce SnapWithChannel?
-	Snaps              []string `long:"snap" value-name:"<snap>[=<channel>]"`
-	ExtraSnaps         []string `long:"extra-snaps" hidden:"yes"` // DEPRECATED
-	RevisionsFile      string   `long:"revisions"`
-	WriteRevisionsFile string   `long:"write-revisions" optional:"true" optional-value:"./seed.manifest"`
+	Snaps                    []string `long:"snap" value-name:"<snap>[=<channel>]"`
+	Components               []string `long:"comp" value-name:"<snap>+<comp>"`
+	ExtraSnaps               []string `long:"extra-snaps" hidden:"yes"` // DEPRECATED
+	RevisionsFile            string   `long:"revisions"`
+	WriteRevisionsFile       string   `long:"write-revisions" optional:"true" optional-value:"./seed.manifest"`
+	Validation               string   `long:"validation" choice:"ignore" choice:"enforce"`
+	AllowSnapdKernelMismatch bool     `long:"allow-snapd-kernel-mismatch"`
 }
 
 func init() {
@@ -86,6 +91,8 @@ For preparing classic images it supports a --classic mode`),
 			// TRANSLATORS: This should not start with a lowercase letter.
 			"snap": i18n.G("Include the given snap from the store or a local file and/or specify the channel to track for the given snap"),
 			// TRANSLATORS: This should not start with a lowercase letter.
+			"comp": i18n.G("Include the given component from the store or a local file"),
+			// TRANSLATORS: This should not start with a lowercase letter.
 			"extra-snaps": i18n.G("Extra snaps to be installed (DEPRECATED)"),
 			// TRANSLATORS: This should not start with a lowercase letter.
 			"revisions": i18n.G("Specify a seeds.manifest file referencing the exact revisions of the provided snaps which should be installed"),
@@ -95,6 +102,10 @@ For preparing classic images it supports a --classic mode`),
 			"channel": i18n.G("The channel to use"),
 			// TRANSLATORS: This should not start with a lowercase letter.
 			"customize": i18n.G("Image customizations specified as JSON file."),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"validation": i18n.G("Control whether validations should be ignored or enforced. (default: ignore)"),
+			// TRANSLATORS: This should not start with a lowercase letter.
+			"allow-snapd-kernel-mismatch": i18n.G("Whether a mismatch between versions of the snapd snap and snapd in kernel is allowed"),
 		}, []argDesc{
 			{
 				// TRANSLATORS: This needs to begin with < and end with >
@@ -114,12 +125,19 @@ var imagePrepare = image.Prepare
 var seedwriterReadManifest = seedwriter.ReadManifest
 
 func (x *cmdPrepareImage) Execute(args []string) error {
+	// plug/slot sanitization is disabled (no-op) by default at the package
+	// level for "snap" command, for seed/seedwriter used by image however
+	// we want real validation.
+	snap.SanitizePlugsSlots = builtin.SanitizePlugsSlots
+
 	opts := &image.Options{
-		Snaps:            x.ExtraSnaps,
-		ModelFile:        x.Positional.ModelAssertionFn,
-		Channel:          x.Channel,
-		Architecture:     x.Architecture,
-		SeedManifestPath: x.WriteRevisionsFile,
+		Snaps:                    x.ExtraSnaps,
+		Components:               x.Components,
+		ModelFile:                x.Positional.ModelAssertionFn,
+		Channel:                  x.Channel,
+		Architecture:             x.Architecture,
+		SeedManifestPath:         x.WriteRevisionsFile,
+		AllowSnapdKernelMismatch: x.AllowSnapdKernelMismatch,
 	}
 
 	if x.RevisionsFile != "" {
@@ -138,6 +156,11 @@ func (x *cmdPrepareImage) Execute(args []string) error {
 		opts.Customizations = *custo
 	}
 
+	if x.Validation != "" {
+		// validation passed in the command line overrides customizations
+		opts.Customizations.Validation = x.Validation
+	}
+
 	snaps := make([]string, 0, len(x.Snaps)+len(x.ExtraSnaps))
 	snapChannels := make(map[string]string)
 	for _, snapWChannel := range x.Snaps {
