@@ -20,78 +20,57 @@
 package main
 
 import (
+	"errors"
 	"fmt"
-	"strconv"
 	"strings"
 
 	"github.com/jessevdk/go-flags"
 
-	"github.com/snapcore/snapd/asserts"
+	"github.com/snapcore/snapd/asserts/snapasserts"
 	"github.com/snapcore/snapd/client"
 	"github.com/snapcore/snapd/i18n"
+	"github.com/snapcore/snapd/strutil"
 )
 
 type cmdValidate struct {
-	clientMixin
 	Monitor    bool `long:"monitor"`
 	Enforce    bool `long:"enforce"`
 	Forget     bool `long:"forget"`
+	Refresh    bool `long:"refresh"`
 	Positional struct {
 		ValidationSet string `positional-arg-name:"<validation-set>"`
 	} `positional-args:"yes"`
 	colorMixin
+	waitMixin
 }
 
 var shortValidateHelp = i18n.G("List or apply validation sets")
 var longValidateHelp = i18n.G(`
-The validate command lists or applies validations sets
+The validate command lists or applies validation sets that state which snaps
+are required or permitted to be installed together, optionally constrained to
+fixed revisions.
+
+A validation set can either be in monitoring mode, in which case its constraints
+aren't enforced, or in enforcing mode, in which case snapd will not allow
+operations which would result in snaps breaking the validation set's constraints.
 `)
 
 func init() {
-	cmd := addCommand("validate", shortValidateHelp, longValidateHelp, func() flags.Commander { return &cmdValidate{} }, colorDescs.also(map[string]string{
+	addCommand("validate", shortValidateHelp, longValidateHelp, func() flags.Commander { return &cmdValidate{} }, waitDescs.also(colorDescs.also(map[string]string{
 		// TRANSLATORS: This should not start with a lowercase letter.
 		"monitor": i18n.G("Monitor the given validations set"),
 		// TRANSLATORS: This should not start with a lowercase letter.
 		"enforce": i18n.G("Enforce the given validation set"),
 		// TRANSLATORS: This should not start with a lowercase letter.
 		"forget": i18n.G("Forget the given validation set"),
-	}), []argDesc{{
+		// TRANSLATORS: This should not start with a lowercase letter.
+		"refresh": i18n.G("Refresh or install snaps to satisfy enforced validation sets"),
+	})), []argDesc{{
 		// TRANSLATORS: This needs to begin with < and end with >
 		name: i18n.G("<validation-set>"),
 		// TRANSLATORS: This should not start with a lowercase letter.
 		desc: i18n.G("Validation set with an optional pinned sequence point, i.e. account-id/name[=seq]"),
 	}})
-	// XXX: remove once api has landed
-	cmd.hidden = true
-}
-
-func splitValidationSetArg(arg string) (account, name string, seq int, err error) {
-	parts := strings.Split(arg, "=")
-	if len(parts) > 2 {
-		return "", "", 0, fmt.Errorf("cannot parse validation set, expected account/name=seq")
-	}
-	if len(parts) == 2 {
-		seq, err = strconv.Atoi(parts[1])
-		if err != nil {
-			return "", "", 0, err
-		}
-	}
-
-	parts = strings.Split(parts[0], "/")
-	if len(parts) != 2 {
-		return "", "", 0, fmt.Errorf("expected a single account/name")
-	}
-
-	account = parts[0]
-	name = parts[1]
-	if !asserts.IsValidAccountID(account) {
-		return "", "", 0, fmt.Errorf("invalid account ID %q", account)
-	}
-	if !asserts.IsValidValidationSetName(name) {
-		return "", "", 0, fmt.Errorf("invalid validation set name %q", name)
-	}
-
-	return account, name, seq, nil
 }
 
 func fmtValid(res *client.ValidationSetResult) string {
@@ -135,13 +114,46 @@ func (cmd *cmdValidate) Execute(args []string) error {
 	var seq int
 	var err error
 	if cmd.Positional.ValidationSet != "" {
-		accountID, name, seq, err = splitValidationSetArg(cmd.Positional.ValidationSet)
+		accountID, name, seq, err = snapasserts.ParseValidationSet(cmd.Positional.ValidationSet)
 		if err != nil {
-			return fmt.Errorf("cannot parse validation set %q: %v", cmd.Positional.ValidationSet, err)
+			return err
 		}
 	}
 
 	if action != "" {
+		if cmd.Refresh && action != "enforce" {
+			return fmt.Errorf("--refresh can only be used together with --enforce")
+		}
+
+		if cmd.Refresh {
+			changeID, err := cmd.client.RefreshMany(nil, nil, &client.SnapOptions{
+				ValidationSets: []string{cmd.Positional.ValidationSet},
+			})
+			if err != nil {
+				return err
+			}
+			chg, err := cmd.wait(changeID)
+			if err != nil {
+				if err == noWait {
+					return nil
+				}
+				return err
+			}
+
+			var names []string
+			if err := chg.Get("snap-names", &names); err != nil && !errors.Is(err, client.ErrNoData) {
+				return err
+			}
+
+			if len(names) != 0 {
+				fmt.Fprintf(Stdout, i18n.G("Refreshed/installed snaps %s to enforce validation set %q\n"), strutil.Quoted(names), cmd.Positional.ValidationSet)
+			} else {
+				fmt.Fprintf(Stdout, i18n.G("Enforced validation set %q\n"), cmd.Positional.ValidationSet)
+			}
+
+			return nil
+		}
+
 		// forget
 		if cmd.Forget {
 			return cmd.client.ForgetValidationSet(accountID, name, seq)
@@ -198,7 +210,7 @@ func (cmd *cmdValidate) Execute(args []string) error {
 		if err != nil {
 			return err
 		}
-		fmt.Fprintf(Stdout, fmtValid(vset))
+		fmt.Fprint(Stdout, fmtValid(vset))
 		// XXX: exit status 1 if invalid?
 	}
 
