@@ -29,9 +29,12 @@ import (
 	"gopkg.in/check.v1"
 
 	"github.com/snapcore/snapd/client"
+	"github.com/snapcore/snapd/daemon"
 	"github.com/snapcore/snapd/httputil"
+	"github.com/snapcore/snapd/overlord/assertstate"
 	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/snapstate"
+	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/store"
 )
@@ -80,10 +83,18 @@ func (s *findSuite) TestFind(c *check.C) {
 func (s *findSuite) TestFindRefreshes(c *check.C) {
 	s.daemon(c)
 
+	var fetchedValidationSets bool
+	restore := daemon.MockAssertstateFetchAllValidationSets(func(*state.State, int, *assertstate.RefreshAssertionsOptions) error {
+		fetchedValidationSets = true
+		return nil
+	})
+	defer restore()
+
 	s.rsnaps = []*snap.Info{{
 		SideInfo: snap.SideInfo{
 			RealName: "store",
 		},
+		Architectures: []string{"all"},
 		Publisher: snap.StoreAccount{
 			ID:          "foo-id",
 			Username:    "foo",
@@ -103,6 +114,7 @@ func (s *findSuite) TestFindRefreshes(c *check.C) {
 	c.Assert(snaps[0]["name"], check.Equals, "store")
 	c.Check(s.currentSnaps, check.HasLen, 1)
 	c.Check(s.actions, check.HasLen, 1)
+	c.Check(fetchedValidationSets, check.Equals, true)
 }
 
 func (s *findSuite) TestFindRefreshSideloaded(c *check.C) {
@@ -128,10 +140,10 @@ func (s *findSuite) TestFindRefreshSideloaded(c *check.C) {
 	err := snapstate.Get(st, "store", &snapst)
 	st.Unlock()
 	c.Assert(err, check.IsNil)
-	c.Assert(snapst.Sequence, check.HasLen, 1)
+	c.Assert(snapst.Sequence.Revisions, check.HasLen, 1)
 
 	// clear the snapid
-	snapst.Sequence[0].SnapID = ""
+	snapst.Sequence.Revisions[0].Snap.SnapID = ""
 	st.Lock()
 	snapstate.Set(st, "store", &snapst)
 	st.Unlock()
@@ -229,6 +241,22 @@ func (s *findSuite) TestFindSection(c *check.C) {
 	})
 }
 
+func (s *findSuite) TestFindCategory(c *check.C) {
+	s.daemon(c)
+
+	s.rsnaps = []*snap.Info{}
+
+	req, err := http.NewRequest("GET", "/v2/find?q=foo&category=bar", nil)
+	c.Assert(err, check.IsNil)
+
+	_ = s.syncReq(c, req, nil)
+
+	c.Check(s.storeSearch, check.DeepEquals, store.Search{
+		Query:    "foo",
+		Category: "bar",
+	})
+}
+
 func (s *findSuite) TestFindScope(c *check.C) {
 	s.daemon(c)
 
@@ -363,7 +391,12 @@ func (s *findSuite) TestFindOneWithAuth(c *check.C) {
 
 	state := d.Overlord().State()
 	state.Lock()
-	user, err := auth.NewUser(state, "username", "email@test.com", "macaroon", []string{"discharge"})
+	user, err := auth.NewUser(state, auth.NewUserParams{
+		Username:   "username",
+		Email:      "email@test.com",
+		Macaroon:   "macaroon",
+		Discharges: []string{"discharge"},
+	})
 	state.Unlock()
 	c.Check(err, check.IsNil)
 
