@@ -31,24 +31,36 @@ import (
 	"github.com/snapcore/snapd/client"
 	"github.com/snapcore/snapd/i18n"
 	"github.com/snapcore/snapd/overlord/auth"
+	"github.com/snapcore/snapd/overlord/snapshotstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/strutil"
 )
 
 var snapshotCmd = &Command{
 	// TODO: also support /v2/snapshots/<id>
-	Path:     "/v2/snapshots",
-	UserOK:   true,
-	PolkitOK: "io.snapcraft.snapd.manage",
-	GET:      listSnapshots,
-	POST:     changeSnapshots,
+	Path:        "/v2/snapshots",
+	GET:         listSnapshots,
+	POST:        changeSnapshots,
+	ReadAccess:  openAccess{},
+	WriteAccess: authenticatedAccess{Polkit: polkitActionManage},
 }
 
 var snapshotExportCmd = &Command{
-	Path: "/v2/snapshots/{id}/export",
-	GET:  getSnapshotExport,
+	Path:       "/v2/snapshots/{id}/export",
+	GET:        getSnapshotExport,
+	ReadAccess: authenticatedAccess{},
 }
 
+var (
+	snapshotList    = snapshotstate.List
+	snapshotCheck   = snapshotstate.Check
+	snapshotForget  = snapshotstate.Forget
+	snapshotRestore = snapshotstate.Restore
+	snapshotSave    = snapshotstate.Save
+	snapshotExport  = snapshotstate.Export
+	snapshotImport  = snapshotstate.Import
+)
+
 func listSnapshots(c *Command, r *http.Request, user *auth.UserState) Response {
 	query := r.URL.Query()
 	var setID uint64
@@ -63,11 +75,11 @@ func listSnapshots(c *Command, r *http.Request, user *auth.UserState) Response {
 	st := c.d.overlord.State()
 	st.Lock()
 	defer st.Unlock()
-	sets, err := snapshotList(context.TODO(), st, setID, strutil.CommaSeparatedList(r.URL.Query().Get("snaps")))
+	sets, err := snapshotList(r.Context(), st, setID, strutil.CommaSeparatedList(r.URL.Query().Get("snaps")))
 	if err != nil {
 		return InternalError("%v", err)
 	}
-	return SyncResponse(sets, nil)
+	return SyncResponse(sets)
 }
 
 // A snapshotAction is used to request an operation on a snapshot
@@ -150,7 +162,7 @@ func changeSnapshots(c *Command, r *http.Request, user *auth.UserState) Response
 	chg.Set("api-data", map[string]interface{}{"snap-names": affected})
 	ensureStateSoon(st)
 
-	return AsyncResponse(nil, &Meta{Change: chg.ID()})
+	return AsyncResponse(nil, chg.ID())
 }
 
 // getSnapshotExport streams an archive containing an export of existing snapshots.
@@ -169,7 +181,7 @@ func getSnapshotExport(c *Command, r *http.Request, user *auth.UserState) Respon
 		return BadRequest("'id' must be a positive base 10 number; got %q", sid)
 	}
 
-	export, err := snapshotExport(context.TODO(), st, setID)
+	export, err := snapshotExport(r.Context(), st, setID)
 	if err != nil {
 		return BadRequest("cannot export %v: %v", setID, err)
 	}
@@ -196,17 +208,17 @@ func doSnapshotImport(c *Command, r *http.Request, user *auth.UserState) Respons
 
 	// XXX: check that we have enough space to import the compressed snapshots
 	st := c.d.overlord.State()
-	setID, snapNames, err := snapshotImport(context.TODO(), st, limitedBodyReader)
+	setID, snapNames, err := snapshotImport(r.Context(), st, limitedBodyReader)
 	if err != nil {
 		return BadRequest(err.Error())
 	}
 
 	result := map[string]interface{}{"set-id": setID, "snaps": snapNames}
-	return SyncResponse(result, nil)
+	return SyncResponse(result)
 }
 
-func snapshotMany(inst *snapInstruction, st *state.State) (*snapInstructionResult, error) {
-	setID, snapshotted, ts, err := snapshotSave(st, inst.Snaps, inst.Users)
+func snapshotMany(_ context.Context, inst *snapInstruction, st *state.State) (*snapInstructionResult, error) {
+	setID, snapshotted, ts, err := snapshotSave(st, inst.Snaps, inst.Users, inst.SnapshotOptions)
 	if err != nil {
 		return nil, err
 	}
