@@ -44,6 +44,7 @@ import (
 	"github.com/snapcore/snapd/overlord"
 	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/restart"
+	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/standby"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snapdenv"
@@ -52,6 +53,7 @@ import (
 )
 
 var ErrRestartSocket = fmt.Errorf("daemon stop requested to wait for socket activation")
+var ErrNoFailureRecoveryNeeded = fmt.Errorf("no failure recovery needed")
 
 var systemdSdNotify = systemd.SdNotify
 
@@ -76,10 +78,10 @@ type Daemon struct {
 	router          *mux.Router
 	standbyOpinions *standby.StandbyOpinions
 
-	// set to what kind of restart was requested if any
+	// set to what kind of restart was requested (if any)
 	requestedRestart restart.RestartType
 	// reboot info needed to handle reboots
-	rebootInfo boot.RebootInfo
+	rebootInfo *boot.RebootInfo
 	// set to remember that we need to exit the daemon in a way that
 	// prevents systemd from restarting it
 	restartSocket bool
@@ -88,13 +90,14 @@ type Daemon struct {
 
 	expectedRebootDidNotHappen bool
 
-	mu sync.Mutex
+	mu     sync.Mutex
+	cancel func()
 }
 
 // A ResponseFunc handles one of the individual verbs for a method
 type ResponseFunc func(*Command, *http.Request, *auth.UserState) Response
 
-// A Command routes a request to an individual per-verb ResponseFUnc
+// A Command routes a request to an individual per-verb ResponseFunc
 type Command struct {
 	Path       string
 	PathPrefix string
@@ -227,7 +230,7 @@ func (d *Daemon) Init() error {
 		return err
 	}
 
-	// The SnapdSocket is required-- without it, die.
+	// The SnapdSocket is required -- without it, die.
 	if listener, err := netutil.GetListener(dirs.SnapdSocket, listenerMap); err == nil {
 		d.snapdListener = &ucrednetListener{Listener: listener}
 	} else {
@@ -249,15 +252,15 @@ func (d *Daemon) Init() error {
 	return nil
 }
 
-// SetDegradedMode puts the daemon into an degraded mode which will the
-// error given in the "err" argument for commands that are not marked
-// as readonlyOK.
+// SetDegradedMode puts the daemon into a degraded mode. In this mode
+// it will return the error given in the "err" argument for commands
+// that are not pure HTTP GETs.
 //
 // This is useful to report errors to the client when the daemon
 // cannot work because e.g. a snapd squashfs precondition check failed
 // or the system is out of diskspace.
 //
-// When the system is fine again calling "DegradedMode(nil)" is enough
+// When the system is fine again, calling "SetDegradedMode(nil)" is enough
 // to put the daemon into full operation again.
 func (d *Daemon) SetDegradedMode(err error) {
 	d.degradedErr = err
@@ -317,8 +320,9 @@ func (d *Daemon) initStandbyHandling() {
 	d.standbyOpinions.Start()
 }
 
-// Start the Daemon
-func (d *Daemon) Start() error {
+// Start the Daemon. Takes a context which will be used as the base request
+// context in the embedded http.Server.
+func (d *Daemon) Start(ctx context.Context) (err error) {
 	if d.expectedRebootDidNotHappen {
 		// we need to schedule and wait for a system restart
 		d.tomb.Kill(nil)
@@ -330,6 +334,15 @@ func (d *Daemon) Start() error {
 		panic("internal error: no Overlord")
 	}
 
+	ctx, cancel := context.WithCancel(ctx)
+	d.cancel = cancel
+	defer func() {
+		// cancel the context on any errors
+		if err != nil {
+			cancel()
+		}
+	}()
+
 	to, reasoning, err := d.overlord.StartupTimeout()
 	if err != nil {
 		return err
@@ -340,8 +353,12 @@ func (d *Daemon) Start() error {
 		logger.Noticef("adjusting startup timeout by %v (%s)", to, reasoning)
 		systemdSdNotify(fmt.Sprintf("EXTEND_TIMEOUT_USEC=%d", us))
 	}
-	// now perform expensive overlord/manages initiliazation
+	// now perform expensive overlord/manages initialization
 	if err := d.overlord.StartUp(); err != nil {
+		if errors.Is(err, snapstate.ErrUnexpectedRuntimeRestart) {
+			logger.Noticef("detected failure recovery context, but no recovery needed")
+			return ErrNoFailureRecoveryNeeded
+		}
 		return err
 	}
 
@@ -349,6 +366,13 @@ func (d *Daemon) Start() error {
 	d.serve = &http.Server{
 		Handler:   logit(d.router),
 		ConnState: d.connTracker.trackConn,
+		BaseContext: func(net.Listener) context.Context {
+			// requests will use the context provided to Start, as
+			// the caller will likely cancel it when appropriate
+			// thus canceling any outstanding requests to the snapd
+			// API
+			return ctx
+		},
 	}
 
 	// enable standby handling
@@ -397,10 +421,7 @@ func (d *Daemon) HandleRestart(t restart.RestartType, rebootInfo *boot.RebootInf
 			logger.Noticef("%s", err)
 		}
 	}
-	d.rebootInfo = boot.RebootInfo{}
-	if rebootInfo != nil {
-		d.rebootInfo = *rebootInfo
-	}
+	d.rebootInfo = rebootInfo
 
 	// die when asked to restart (systemd should get us back up!) etc
 	switch t {
@@ -437,7 +458,7 @@ var (
 	rebootNoticeWait       = 3 * time.Second
 	rebootWaitTimeout      = 10 * time.Minute
 	rebootRetryWaitTimeout = 5 * time.Minute
-	rebootMaxTentatives    = 3
+	rebootMaxAttempts      = 3
 )
 
 func (d *Daemon) updateMaintenanceFile(rst restart.RestartType) error {
@@ -479,6 +500,10 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 		return fmt.Errorf("internal error: no Overlord")
 	}
 
+	if d.cancel != nil {
+		d.cancel()
+	}
+
 	d.tomb.Kill(nil)
 
 	// check the state associated with a potential restart with the lock to
@@ -487,12 +512,13 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 	// needsFullShutdown is whether the entire system will
 	// shutdown or not as a consequence of this request
 	needsFullShutdown := false
-	switch d.requestedRestart {
+	restartType := d.requestedRestart
+	switch restartType {
 	case restart.RestartSystem, restart.RestartSystemNow, restart.RestartSystemHaltNow, restart.RestartSystemPoweroffNow:
 		needsFullShutdown = true
 	}
 	immediateShutdown := false
-	switch d.requestedRestart {
+	switch restartType {
 	case restart.RestartSystemNow, restart.RestartSystemHaltNow, restart.RestartSystemPoweroffNow:
 		immediateShutdown = true
 	}
@@ -503,13 +529,14 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 	// before not accepting any new client connections we need to write the
 	// maintenance.json file for potential clients to see after the daemon stops
 	// responding so they can read it correctly and handle the maintenance
-	if err := d.updateMaintenanceFile(d.requestedRestart); err != nil {
+	if err := d.updateMaintenanceFile(restartType); err != nil {
 		logger.Noticef("error writing maintenance file: %v", err)
 	}
 
-	d.snapdListener.Close()
-	d.standbyOpinions.Stop()
-
+	// take a timestamp before shutting down the snap listener, and
+	// use the time we may spend on waiting for hooks against the shutdown
+	// delay.
+	ts := time.Now()
 	if d.snapListener != nil {
 		// stop running hooks first
 		// and do it more gracefully if we are restarting
@@ -525,11 +552,18 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 		hookMgr.StopHooks()
 		d.snapListener.Close()
 	}
+	timeSpent := time.Since(ts)
 
-	if needsFullShutdown {
-		// give time to polling clients to notice restart
-		time.Sleep(rebootNoticeWait)
+	// When shutting down the snapd listener wait until the rebootNoticeWait
+	// period has passed before snapdListener is closed to allow polling
+	// clients to access the daemon. For testing we disable this unless SNAPD_SHUTDOWN_DELAY
+	// has been set, to avoid incurring this wait for every daemon restart which happens
+	// quite often in testing.
+	if !snapdenv.Testing() || osutil.GetenvBool("SNAPD_SHUTDOWN_DELAY") {
+		time.Sleep(rebootNoticeWait - timeSpent)
 	}
+	d.snapdListener.Close()
+	d.standbyOpinions.Stop()
 
 	// We're using the background context here because the tomb's
 	// context will likely already have been cancelled when we are
@@ -578,7 +612,7 @@ func (d *Daemon) Stop(sigCh chan<- os.Signal) error {
 	}
 
 	if needsFullShutdown {
-		return d.doReboot(sigCh, d.requestedRestart, &rebootInfo, immediateShutdown, rebootWaitTimeout)
+		return d.doReboot(sigCh, d.requestedRestart, rebootInfo, immediateShutdown, rebootWaitTimeout)
 	}
 
 	if d.restartSocket {
@@ -631,7 +665,7 @@ func (d *Daemon) doReboot(sigCh chan<- os.Signal, rst restart.RestartType, rbi *
 		action = boot.RebootPoweroff
 	}
 	// ask for shutdown and wait for it to happen.
-	// if we exit snapd will be restared by systemd
+	// if we exit snapd will be restarted by systemd
 	if err := reboot(action, rebootDelay, rbi); err != nil {
 		return err
 	}
@@ -671,23 +705,23 @@ var errExpectedReboot = errors.New("expected reboot did not happen")
 
 // RebootDidNotHappen implements part of overlord.RestartBehavior.
 func (d *Daemon) RebootDidNotHappen(st *state.State) error {
-	var nTentative int
-	err := st.Get("daemon-system-restart-tentative", &nTentative)
+	var attempt int
+	err := st.Get("daemon-system-restart-tentative", &attempt)
 	if err != nil && !errors.Is(err, state.ErrNoState) {
 		return err
 	}
-	nTentative++
-	if nTentative > rebootMaxTentatives {
+	attempt++
+	if attempt > rebootMaxAttempts {
 		// giving up, proceed normally, some in-progress refresh
 		// might get rolled back!!
 		restart.ClearReboot(st)
 		clearReboot(st)
-		logger.Noticef("snapd was restarted while a system restart was expected, snapd retried to schedule and waited again for a system restart %d times and is giving up", rebootMaxTentatives)
+		logger.Noticef("snapd was restarted while a system restart was expected, snapd retried to schedule and waited again for a system restart %d times and is giving up", rebootMaxAttempts)
 		return nil
 	}
-	st.Set("daemon-system-restart-tentative", nTentative)
+	st.Set("daemon-system-restart-tentative", attempt)
 	d.state = st
-	logger.Noticef("snapd was restarted while a system restart was expected, snapd will try to schedule and wait for a system restart again (tenative %d/%d)", nTentative, rebootMaxTentatives)
+	logger.Noticef("snapd was restarted while a system restart was expected, snapd will try to schedule and wait for a system restart again (attempt %d/%d)", attempt, rebootMaxAttempts)
 	return errExpectedReboot
 }
 
