@@ -25,7 +25,7 @@ import (
 	_ "crypto/sha1"
 	"fmt"
 	"io"
-	"io/ioutil"
+	"io/fs"
 	"os"
 	"path/filepath"
 	"sort"
@@ -46,8 +46,8 @@ func checkSourceIsDir(src string) error {
 	return nil
 }
 
-func checkContent(content *VolumeContent) error {
-	if content.ResolvedSource() == "" {
+func checkContent(content *ResolvedContent) error {
+	if content.ResolvedSource == "" {
 		return fmt.Errorf("internal error: source cannot be unset")
 	}
 	if content.Target == "" {
@@ -56,7 +56,7 @@ func checkContent(content *VolumeContent) error {
 	return nil
 }
 
-func observe(observer ContentObserver, op ContentOperation, ps *LaidOutStructure, root, dst string, data *ContentChange) (ContentChangeAction, error) {
+func observe(observer ContentObserver, op ContentOperation, partRole, root, dst string, data *ContentChange) (ContentChangeAction, error) {
 	if observer == nil {
 		return ChangeApply, nil
 	}
@@ -69,7 +69,7 @@ func observe(observer ContentObserver, op ContentOperation, ps *LaidOutStructure
 		}
 		relativeTarget = relative
 	}
-	return observer.Observe(op, ps, root, relativeTarget, data)
+	return observer.Observe(op, partRole, root, relativeTarget, data)
 }
 
 // TODO: MountedFilesystemWriter should not be exported
@@ -77,27 +77,24 @@ func observe(observer ContentObserver, op ContentOperation, ps *LaidOutStructure
 // MountedFilesystemWriter assists in writing contents of a structure to a
 // mounted filesystem.
 type MountedFilesystemWriter struct {
-	contentDir string
-	ps         *LaidOutStructure
-	observer   ContentObserver
+	fromPs   *LaidOutStructure
+	ps       *LaidOutStructure
+	observer ContentObserver
 }
 
 // NewMountedFilesystemWriter returns a writer capable of writing provided
 // structure, with content of the structure stored in the given root directory.
-func NewMountedFilesystemWriter(contentDir string, ps *LaidOutStructure, observer ContentObserver) (*MountedFilesystemWriter, error) {
+func NewMountedFilesystemWriter(fromPs, ps *LaidOutStructure, observer ContentObserver) (*MountedFilesystemWriter, error) {
 	if ps == nil {
 		return nil, fmt.Errorf("internal error: *LaidOutStructure is nil")
 	}
 	if !ps.HasFilesystem() {
 		return nil, fmt.Errorf("structure %v has no filesystem", ps)
 	}
-	if contentDir == "" {
-		return nil, fmt.Errorf("internal error: gadget content directory cannot be unset")
-	}
 	fw := &MountedFilesystemWriter{
-		contentDir: contentDir,
-		ps:         ps,
-		observer:   observer,
+		fromPs:   fromPs,
+		ps:       ps,
+		observer: observer,
 	}
 	return fw, nil
 }
@@ -133,7 +130,7 @@ func (m *MountedFilesystemWriter) Write(whereDir string, preserve []string) erro
 		return fmt.Errorf("cannot map preserve entries for destination %q: %v", whereDir, err)
 	}
 
-	for _, c := range m.ps.Content {
+	for _, c := range m.ps.ResolvedContent {
 		if err := m.writeVolumeContent(whereDir, &c, preserveInDst); err != nil {
 			return fmt.Errorf("cannot write filesystem content of %s: %v", c, err)
 		}
@@ -143,8 +140,9 @@ func (m *MountedFilesystemWriter) Write(whereDir string, preserve []string) erro
 
 // writeDirectory copies the source directory, or its contents under target
 // location dst. Follows rsync like semantics, that is:
-//   /foo/ -> /bar - writes contents of foo under /bar
-//   /foo  -> /bar - writes foo and its subtree under /bar
+//
+//	/foo/ -> /bar - writes contents of foo under /bar
+//	/foo  -> /bar - writes foo and its subtree under /bar
 func (m *MountedFilesystemWriter) writeDirectory(volumeRoot, src, dst string, preserveInDst []string) error {
 	hasDirSourceSlash := strings.HasSuffix(src, "/")
 
@@ -157,7 +155,7 @@ func (m *MountedFilesystemWriter) writeDirectory(volumeRoot, src, dst string, pr
 		dst = filepath.Join(dst, filepath.Base(src))
 	}
 
-	fis, err := ioutil.ReadDir(src)
+	fis, err := os.ReadDir(src)
 	if err != nil {
 		return fmt.Errorf("cannot list directory entries: %v", err)
 	}
@@ -195,7 +193,7 @@ func (m *MountedFilesystemWriter) observedWriteFileOrSymlink(volumeRoot, src, ds
 		// with content in this file
 		After: src,
 	}
-	act, err := observe(m.observer, ContentWrite, m.ps, volumeRoot, dst, data)
+	act, err := observe(m.observer, ContentWrite, m.ps.Role(), volumeRoot, dst, data)
 	if err != nil {
 		return fmt.Errorf("cannot observe file write: %v", err)
 	}
@@ -207,8 +205,10 @@ func (m *MountedFilesystemWriter) observedWriteFileOrSymlink(volumeRoot, src, ds
 
 // writeFileOrSymlink writes the source file or a symlink at given location or
 // under given directory. Follows rsync like semantics, that is:
-//   /foo -> /bar/ - writes foo as /bar/foo
-//   /foo  -> /bar - writes foo as /bar
+//
+//	/foo -> /bar/ - writes foo as /bar/foo
+//	/foo  -> /bar - writes foo as /bar
+//
 // The destination location is overwritten.
 func writeFileOrSymlink(src, dst string, preserveInDst []string) error {
 	if strings.HasSuffix(dst, "/") {
@@ -246,28 +246,23 @@ func writeFileOrSymlink(src, dst string, preserveInDst []string) error {
 	return nil
 }
 
-func (m *MountedFilesystemWriter) writeVolumeContent(volumeRoot string, content *VolumeContent, preserveInDst []string) error {
+func (m *MountedFilesystemWriter) writeVolumeContent(volumeRoot string, content *ResolvedContent, preserveInDst []string) error {
 	if err := checkContent(content); err != nil {
 		return err
 	}
-	// TODO: ResolvedSource() will already have resolved m.contentDir
-	realSource := filepath.Join(m.contentDir, content.ResolvedSource())
 	realTarget := filepath.Join(volumeRoot, content.Target)
 
 	// filepath trims the trailing /, restore if needed
 	if strings.HasSuffix(content.Target, "/") {
 		realTarget += "/"
 	}
-	if strings.HasSuffix(content.ResolvedSource(), "/") {
-		realSource += "/"
-	}
 
-	if osutil.IsDirectory(realSource) || strings.HasSuffix(content.ResolvedSource(), "/") {
+	if osutil.IsDirectory(content.ResolvedSource) || strings.HasSuffix(content.ResolvedSource, "/") {
 		// write a directory
-		return m.writeDirectory(volumeRoot, realSource, realTarget, preserveInDst)
+		return m.writeDirectory(volumeRoot, content.ResolvedSource, realTarget, preserveInDst)
 	} else {
 		// write a file
-		return m.observedWriteFileOrSymlink(volumeRoot, realSource, realTarget, preserveInDst)
+		return m.observedWriteFileOrSymlink(volumeRoot, content.ResolvedSource, realTarget, preserveInDst)
 	}
 }
 
@@ -313,9 +308,9 @@ type mountedFilesystemUpdater struct {
 // structure, with structure content coming from provided root directory. The
 // mount is located by calling a mount lookup helper. The backup directory
 // contains backup state information for use during rollback.
-func newMountedFilesystemUpdater(rootDir string, ps *LaidOutStructure, backupDir string, mountLookup mountLookupFunc, observer ContentObserver) (*mountedFilesystemUpdater, error) {
+func newMountedFilesystemUpdater(fromPs, ps *LaidOutStructure, backupDir string, mountLookup mountLookupFunc, observer ContentObserver) (*mountedFilesystemUpdater, error) {
 	// avoid passing observer, writes will not be observed
-	fw, err := NewMountedFilesystemWriter(rootDir, ps, nil)
+	fw, err := NewMountedFilesystemWriter(fromPs, ps, nil)
 	if err != nil {
 		return nil, err
 	}
@@ -340,7 +335,7 @@ func newMountedFilesystemUpdater(rootDir string, ps *LaidOutStructure, backupDir
 }
 
 func fsStructBackupPath(backupDir string, ps *LaidOutStructure) string {
-	return filepath.Join(backupDir, fmt.Sprintf("struct-%v", ps.Index))
+	return filepath.Join(backupDir, fmt.Sprintf("struct-%v", ps.VolumeStructure.YamlIndex))
 }
 
 // entryDestPaths resolves destination and backup paths for given
@@ -361,22 +356,28 @@ func (f *mountedFilesystemUpdater) entryDestPaths(dstRoot, source, target, backu
 	return dstPath, backupPath
 }
 
-// entrySourcePath returns the path of given source entry within the root
-// directory provided during initialization.
-func (f *mountedFilesystemUpdater) entrySourcePath(source string) string {
-	srcPath := filepath.Join(f.contentDir, source)
-
-	if strings.HasSuffix(source, "/") {
-		// restore trailing / if one was there
-		srcPath += "/"
+func getDestinationPath(content VolumeContent) (dst string) {
+	dst = content.Target
+	if strings.HasSuffix(dst, "/") {
+		dst = filepath.Join(dst, filepath.Base(content.UnresolvedSource))
 	}
-	return srcPath
+
+	return dst
+}
+
+func (f *mountedFilesystemUpdater) getKnownContent() (knownContent map[string]bool) {
+	knownContent = make(map[string]bool)
+	for _, c := range f.ps.VolumeStructure.Content {
+		knownContent[getDestinationPath(c)] = true
+	}
+
+	return knownContent
 }
 
 // Update applies an update to a mounted filesystem. The caller must have
 // executed a Backup() before, to prepare a data set for rollback purpose.
 func (f *mountedFilesystemUpdater) Update() error {
-	preserveInDst, err := mapPreserve(f.mountPoint, f.ps.Update.Preserve)
+	preserveInDst, err := mapPreserve(f.mountPoint, f.ps.VolumeStructure.Update.Preserve)
 	if err != nil {
 		return fmt.Errorf("cannot map preserve entries for mount location %q: %v", f.mountPoint, err)
 	}
@@ -384,7 +385,8 @@ func (f *mountedFilesystemUpdater) Update() error {
 	backupRoot := fsStructBackupPath(f.backupDir, f.ps)
 
 	skipped := 0
-	for _, c := range f.ps.Content {
+
+	for _, c := range f.ps.ResolvedContent {
 		if err := f.updateVolumeContent(f.mountPoint, &c, preserveInDst, backupRoot); err != nil {
 			if err == ErrNoUpdate {
 				skipped++
@@ -394,16 +396,48 @@ func (f *mountedFilesystemUpdater) Update() error {
 		}
 	}
 
-	if skipped == len(f.ps.Content) {
+	knownContent := f.getKnownContent()
+	deleted := false
+	if f.fromPs != nil {
+		for _, c := range f.fromPs.VolumeStructure.Content {
+			if knownContent[getDestinationPath(c)] {
+				continue
+			}
+			destPath, backupPath := f.entryDestPaths(f.mountPoint, c.UnresolvedSource, c.Target, backupRoot)
+			preserveStamp := backupPath + ".preserve"
+
+			// We skip directory because we do not know
+			// exactly the content that is supposed to be
+			// in there.
+			// XXX: it might be possible to recursively compare
+			// directories from mounted snaps to detect
+			// what files are removed.
+			if osutil.IsDirectory(destPath) {
+				continue
+			}
+
+			if strutil.SortedListContains(preserveInDst, destPath) || osutil.FileExists(preserveStamp) {
+				continue
+			}
+
+			if err := os.Remove(destPath); err != nil {
+				if os.IsNotExist(err) {
+					continue
+				}
+				return fmt.Errorf("cannot remove content: %v", err)
+			}
+			deleted = true
+		}
+	}
+
+	if !deleted && skipped == len(f.ps.ResolvedContent) {
 		return ErrNoUpdate
 	}
 
 	return nil
 }
 
-func (f *mountedFilesystemUpdater) sourceDirectoryEntries(source string) ([]os.FileInfo, error) {
-	srcPath := f.entrySourcePath(source)
-
+func (f *mountedFilesystemUpdater) sourceDirectoryEntries(srcPath string) ([]fs.DirEntry, error) {
 	if err := checkSourceIsDir(srcPath); err != nil {
 		return nil, err
 	}
@@ -413,13 +447,14 @@ func (f *mountedFilesystemUpdater) sourceDirectoryEntries(source string) ([]os.F
 		return nil, fmt.Errorf("source is a symbolic link")
 	}
 
-	return ioutil.ReadDir(srcPath)
+	return os.ReadDir(srcPath)
 }
 
 // targetInSourceDir resolves the actual target for given source directory name
 // and target specification.
-//  source: /foo/bar/ target: /baz => /bar/    (contents of /foo/bar/ under /baz)
-//  source: /foo/bar  target: /baz => /bar/bar (directory /foo/bar under /baz, contents under /baz/bar)
+//
+//	source: /foo/bar/ target: /baz => /bar/    (contents of /foo/bar/ under /baz)
+//	source: /foo/bar  target: /baz => /bar/bar (directory /foo/bar under /baz, contents under /baz/bar)
 func targetForSourceDir(source, target string) string {
 	if strings.HasSuffix(source, "/") {
 		// contents of source directory land under target
@@ -472,7 +507,6 @@ func (f *mountedFilesystemUpdater) updateDirectory(dstRoot, source, target strin
 }
 
 func (f *mountedFilesystemUpdater) updateOrSkipFile(dstRoot, source, target string, preserveInDst []string, backupDir string) error {
-	srcPath := f.entrySourcePath(source)
 	dstPath, backupPath := f.entryDestPaths(dstRoot, source, target, backupDir)
 	backupName := backupPath + ".backup"
 	sameStamp := backupPath + ".same"
@@ -480,7 +514,7 @@ func (f *mountedFilesystemUpdater) updateOrSkipFile(dstRoot, source, target stri
 	ignoreStamp := backupPath + ".ignore"
 
 	// TODO: enable support for symlinks when needed
-	if osutil.IsSymlink(srcPath) {
+	if osutil.IsSymlink(source) {
 		return fmt.Errorf("cannot update file %s: symbolic links are not supported", source)
 	}
 
@@ -505,23 +539,20 @@ func (f *mountedFilesystemUpdater) updateOrSkipFile(dstRoot, source, target stri
 		}
 	}
 
-	return writeFileOrSymlink(srcPath, dstPath, preserveInDst)
+	return writeFileOrSymlink(source, dstPath, preserveInDst)
 }
 
-func (f *mountedFilesystemUpdater) updateVolumeContent(volumeRoot string, content *VolumeContent, preserveInDst []string, backupDir string) error {
+func (f *mountedFilesystemUpdater) updateVolumeContent(volumeRoot string, content *ResolvedContent, preserveInDst []string, backupDir string) error {
 	if err := checkContent(content); err != nil {
 		return err
 	}
 
-	// TODO: ResolvedSource() will already have resolved f.entrySourcePath
-	srcPath := f.entrySourcePath(content.ResolvedSource())
-
-	if osutil.IsDirectory(srcPath) || strings.HasSuffix(content.ResolvedSource(), "/") {
+	if osutil.IsDirectory(content.ResolvedSource) || strings.HasSuffix(content.ResolvedSource, "/") {
 		// TODO: pass both Unresolved and resolved Source (unresolved for better error reporting)
-		return f.updateDirectory(volumeRoot, content.ResolvedSource(), content.Target, preserveInDst, backupDir)
+		return f.updateDirectory(volumeRoot, content.ResolvedSource, content.Target, preserveInDst, backupDir)
 	} else {
 		// TODO: pass both Unresolved and resolved Source (unresolved for better error reporting)
-		return f.updateOrSkipFile(volumeRoot, content.ResolvedSource(), content.Target, preserveInDst, backupDir)
+		return f.updateOrSkipFile(volumeRoot, content.ResolvedSource, content.Target, preserveInDst, backupDir)
 	}
 }
 
@@ -556,7 +587,6 @@ func (f *mountedFilesystemUpdater) updateVolumeContent(volumeRoot string, conten
 // ├── b.same             <-- stamp indicating ./b is identical to the update data
 // ├── c.ignore           <-- stamp indicating change to ./c was requested to be ignored
 // └── d.preserve         <-- stamp indicating ./d is to be preserved
-//
 func (f *mountedFilesystemUpdater) Backup() error {
 	backupRoot := fsStructBackupPath(f.backupDir, f.ps)
 
@@ -564,17 +594,46 @@ func (f *mountedFilesystemUpdater) Backup() error {
 		return fmt.Errorf("cannot create backup directory: %v", err)
 	}
 
-	preserveInDst, err := mapPreserve(f.mountPoint, f.ps.Update.Preserve)
+	preserveInDst, err := mapPreserve(f.mountPoint, f.ps.VolumeStructure.Update.Preserve)
 	if err != nil {
 		return fmt.Errorf("cannot map preserve entries for mount location %q: %v", f.mountPoint, err)
 	}
 
-	for _, c := range f.ps.Content {
+	for _, c := range f.ps.ResolvedContent {
 		if err := f.backupVolumeContent(f.mountPoint, &c, preserveInDst, backupRoot); err != nil {
 			return fmt.Errorf("cannot backup content: %v", err)
 		}
 	}
 
+	knownContent := f.getKnownContent()
+	if f.fromPs != nil {
+		for _, c := range f.fromPs.VolumeStructure.Content {
+			if knownContent[getDestinationPath(c)] {
+				continue
+			}
+
+			destPath, backupPath := f.entryDestPaths(f.mountPoint, c.UnresolvedSource, c.Target, backupRoot)
+			// We skip directory because we do not know
+			// exactly the content that is supposed to be
+			// in there.
+			// XXX: it might be possible to recursively compare
+			// directories from mounted snaps to detect
+			// what files are removed.
+			if osutil.IsDirectory(destPath) {
+				continue
+			}
+			backupName := backupPath + ".backup"
+
+			if !osutil.FileExists(destPath) {
+				continue
+			}
+
+			if err := writeFileOrSymlink(destPath, backupName, nil); err != nil {
+				return fmt.Errorf("cannot create backup file: %v", err)
+			}
+		}
+	}
+
 	return nil
 }
 
@@ -639,7 +698,7 @@ func (f *mountedFilesystemUpdater) checkpointPrefix(dstRoot, target string, back
 	return nil
 }
 
-func (f *mountedFilesystemUpdater) ignoreChange(change *ContentChange, backupPath string) error {
+func (f *mountedFilesystemUpdater) ignoreChange(backupPath string) error {
 	preserveStamp := backupPath + ".preserve"
 	backupName := backupPath + ".backup"
 	sameStamp := backupPath + ".same"
@@ -662,13 +721,13 @@ func (f *mountedFilesystemUpdater) observedBackupOrCheckpointFile(dstRoot, sourc
 	}
 	if change != nil {
 		dstPath, backupPath := f.entryDestPaths(dstRoot, source, target, backupDir)
-		act, err := observe(f.updateObserver, ContentUpdate, f.ps, f.mountPoint, dstPath, change)
+		act, err := observe(f.updateObserver, ContentUpdate, f.ps.Role(), f.mountPoint, dstPath, change)
 		if err != nil {
 			return fmt.Errorf("cannot observe pending file write %v\n", err)
 		}
 		if act == ChangeIgnore {
 			// observer asked for the change to be ignored
-			if err := f.ignoreChange(change, backupPath); err != nil {
+			if err := f.ignoreChange(backupPath); err != nil {
 				return fmt.Errorf("cannot ignore content change: %v", err)
 			}
 		}
@@ -683,7 +742,6 @@ func (f *mountedFilesystemUpdater) observedBackupOrCheckpointFile(dstRoot, sourc
 // created. Returns a content change if a file will be written by the update
 // pass.
 func (f *mountedFilesystemUpdater) backupOrCheckpointFile(dstRoot, source, target string, preserveInDst []string, backupDir string) (change *ContentChange, err error) {
-	srcPath := f.entrySourcePath(source)
 	dstPath, backupPath := f.entryDestPaths(dstRoot, source, target, backupDir)
 
 	backupName := backupPath + ".backup"
@@ -693,14 +751,14 @@ func (f *mountedFilesystemUpdater) backupOrCheckpointFile(dstRoot, source, targe
 
 	changeWithBackup := &ContentChange{
 		// content of the new data
-		After: srcPath,
+		After: source,
 		// this is the original data that was present before the
 		// update
 		Before: backupName,
 	}
 	changeNewFile := &ContentChange{
 		// content of the new data
-		After: srcPath,
+		After: source,
 	}
 
 	if osutil.FileExists(ignoreStamp) {
@@ -771,7 +829,7 @@ func (f *mountedFilesystemUpdater) backupOrCheckpointFile(dstRoot, source, targe
 	}
 
 	// digest of the update
-	updateDigest, _, err := osutil.FileDigest(srcPath, crypto.SHA1)
+	updateDigest, _, err := osutil.FileDigest(source, crypto.SHA1)
 	if err != nil {
 		backup.Cancel()
 		return nil, fmt.Errorf("cannot checksum update file: %v", err)
@@ -797,24 +855,21 @@ func (f *mountedFilesystemUpdater) backupOrCheckpointFile(dstRoot, source, targe
 	return changeWithBackup, nil
 }
 
-func (f *mountedFilesystemUpdater) backupVolumeContent(volumeRoot string, content *VolumeContent, preserveInDst []string, backupDir string) error {
+func (f *mountedFilesystemUpdater) backupVolumeContent(volumeRoot string, content *ResolvedContent, preserveInDst []string, backupDir string) error {
 	if err := checkContent(content); err != nil {
 		return err
 	}
 
-	// TODO: ResolvedSource() will already have resolved f.entrySourcePath
-	srcPath := f.entrySourcePath(content.ResolvedSource())
-
 	if err := f.checkpointPrefix(volumeRoot, content.Target, backupDir); err != nil {
 		return err
 	}
-	if osutil.IsDirectory(srcPath) || strings.HasSuffix(content.ResolvedSource(), "/") {
+	if osutil.IsDirectory(content.ResolvedSource) || strings.HasSuffix(content.ResolvedSource, "/") {
 		// backup directory contents
 		// TODO: pass both Unresolved and resolved Source (unresolved for better error reporting)
-		return f.backupOrCheckpointDirectory(volumeRoot, content.ResolvedSource(), content.Target, preserveInDst, backupDir)
+		return f.backupOrCheckpointDirectory(volumeRoot, content.ResolvedSource, content.Target, preserveInDst, backupDir)
 	} else {
 		// backup a file
-		return f.observedBackupOrCheckpointFile(volumeRoot, content.ResolvedSource(), content.Target, preserveInDst, backupDir)
+		return f.observedBackupOrCheckpointFile(volumeRoot, content.ResolvedSource, content.Target, preserveInDst, backupDir)
 	}
 }
 
@@ -825,12 +880,39 @@ func (f *mountedFilesystemUpdater) backupVolumeContent(volumeRoot string, conten
 func (f *mountedFilesystemUpdater) Rollback() error {
 	backupRoot := fsStructBackupPath(f.backupDir, f.ps)
 
-	preserveInDst, err := mapPreserve(f.mountPoint, f.ps.Update.Preserve)
+	preserveInDst, err := mapPreserve(f.mountPoint, f.ps.VolumeStructure.Update.Preserve)
 	if err != nil {
 		return fmt.Errorf("cannot map preserve entries for mount location %q: %v", f.mountPoint, err)
 	}
 
-	for _, c := range f.ps.Content {
+	knownContent := f.getKnownContent()
+	if f.fromPs != nil {
+		for _, c := range f.fromPs.VolumeStructure.Content {
+			if knownContent[getDestinationPath(c)] {
+				continue
+			}
+
+			destPath, backupPath := f.entryDestPaths(f.mountPoint, c.UnresolvedSource, c.Target, backupRoot)
+
+			if osutil.IsDirectory(destPath) {
+				continue
+			}
+
+			if err := os.Remove(destPath); err != nil {
+				if !os.IsNotExist(err) {
+					return fmt.Errorf("cannot rollback %s: %v", destPath, err)
+				}
+			}
+
+			backupName := backupPath + ".backup"
+
+			if err := writeFileOrSymlink(backupName, destPath, nil); err != nil {
+				return fmt.Errorf("cannot rollback %s: %v", destPath, err)
+			}
+		}
+	}
+
+	for _, c := range f.ps.ResolvedContent {
 		if err := f.rollbackVolumeContent(f.mountPoint, &c, preserveInDst, backupRoot); err != nil {
 			return fmt.Errorf("cannot rollback content: %v", err)
 		}
@@ -883,7 +965,6 @@ func (f *mountedFilesystemUpdater) rollbackDirectory(dstRoot, source, target str
 }
 
 func (f *mountedFilesystemUpdater) rollbackFile(dstRoot, source, target string, preserveInDst []string, backupDir string) error {
-	srcPath := f.entrySourcePath(source)
 	dstPath, backupPath := f.entryDestPaths(dstRoot, source, target, backupDir)
 
 	backupName := backupPath + ".backup"
@@ -907,7 +988,7 @@ func (f *mountedFilesystemUpdater) rollbackFile(dstRoot, source, target string,
 	}
 
 	data := &ContentChange{
-		After: srcPath,
+		After: source,
 		// original content was in the backup file
 		Before: backupName,
 	}
@@ -926,7 +1007,7 @@ func (f *mountedFilesystemUpdater) rollbackFile(dstRoot, source, target string,
 	}
 	// avoid passing source path during rollback, the file has been restored
 	// to the disk already
-	_, err := observe(f.updateObserver, ContentRollback, f.ps, f.mountPoint, dstPath, data)
+	_, err := observe(f.updateObserver, ContentRollback, f.ps.Role(), f.mountPoint, dstPath, data)
 	if err != nil {
 		return fmt.Errorf("cannot observe pending file rollback %v\n", err)
 	}
@@ -934,21 +1015,18 @@ func (f *mountedFilesystemUpdater) rollbackFile(dstRoot, source, target string,
 	return nil
 }
 
-func (f *mountedFilesystemUpdater) rollbackVolumeContent(volumeRoot string, content *VolumeContent, preserveInDst []string, backupDir string) error {
+func (f *mountedFilesystemUpdater) rollbackVolumeContent(volumeRoot string, content *ResolvedContent, preserveInDst []string, backupDir string) error {
 	if err := checkContent(content); err != nil {
 		return err
 	}
 
-	// TODO: ResolvedSource() will already have resolved f.entrySourcePath
-	srcPath := f.entrySourcePath(content.ResolvedSource())
-
 	var err error
-	if osutil.IsDirectory(srcPath) || strings.HasSuffix(content.ResolvedSource(), "/") {
+	if osutil.IsDirectory(content.ResolvedSource) || strings.HasSuffix(content.ResolvedSource, "/") {
 		// rollback directory
-		err = f.rollbackDirectory(volumeRoot, content.ResolvedSource(), content.Target, preserveInDst, backupDir)
+		err = f.rollbackDirectory(volumeRoot, content.ResolvedSource, content.Target, preserveInDst, backupDir)
 	} else {
 		// rollback file
-		err = f.rollbackFile(volumeRoot, content.ResolvedSource(), content.Target, preserveInDst, backupDir)
+		err = f.rollbackFile(volumeRoot, content.ResolvedSource, content.Target, preserveInDst, backupDir)
 	}
 	if err != nil {
 		return err
