@@ -120,7 +120,9 @@ func (s *GreengrassSupportInterfaceSuite) TestAppArmorSpec(c *C) {
 		s.plug,
 		s.containerModePlug,
 	} {
-		spec := &apparmor.Specification{}
+		appSet, err := interfaces.NewSnapAppSet(plug.Snap(), nil)
+		c.Assert(err, IsNil)
+		spec := apparmor.NewSpecification(appSet)
 		c.Assert(spec.AddConnectedPlug(s.iface, plug, s.slot), IsNil)
 		c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.other.app2"})
 		c.Check(spec.SnippetForTag("snap.other.app2"), testutil.Contains, "mount options=(rw, bind) /var/snap/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/** -> /var/snap/{@{SNAP_NAME},@{SNAP_INSTANCE_NAME}}/** ,\n")
@@ -133,7 +135,9 @@ func (s *GreengrassSupportInterfaceSuite) TestProcessModeAppArmorSpec(c *C) {
 	// no features so should not support userns
 	restore := apparmor_sandbox.MockFeatures(nil, nil, nil, nil)
 	defer restore()
-	spec := &apparmor.Specification{}
+	appSet, err := interfaces.NewSnapAppSet(s.processModePlug.Snap(), nil)
+	c.Assert(err, IsNil)
+	spec := apparmor.NewSpecification(appSet)
 	c.Assert(spec.AddConnectedPlug(s.iface, s.processModePlug, s.slot), IsNil)
 	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.other.app2"})
 	c.Check(spec.SnippetForTag("snap.other.app2"), testutil.Contains, "/ ix,\n")
@@ -147,14 +151,18 @@ func (s *GreengrassSupportInterfaceSuite) TestSecCompSpec(c *C) {
 		s.plug,
 		s.containerModePlug,
 	} {
-		spec := &seccomp.Specification{}
+		appSet, err := interfaces.NewSnapAppSet(plug.Snap(), nil)
+		c.Assert(err, IsNil)
+		spec := seccomp.NewSpecification(appSet)
 		c.Assert(spec.AddConnectedPlug(s.iface, plug, s.slot), IsNil)
 		c.Check(spec.SnippetForTag("snap.other.app2"), testutil.Contains, "# for overlayfs and various bind mounts\nmount\numount2\npivot_root\n")
 	}
 }
 
 func (s *GreengrassSupportInterfaceSuite) TestProcessModeSecCompSpec(c *C) {
-	spec := &seccomp.Specification{}
+	appSet, err := interfaces.NewSnapAppSet(s.processModePlug.Snap(), nil)
+	c.Assert(err, IsNil)
+	spec := seccomp.NewSpecification(appSet)
 	c.Assert(spec.AddConnectedPlug(s.iface, s.processModePlug, s.slot), IsNil)
 	c.Check(spec.SnippetForTag("snap.other.app2"), Not(testutil.Contains), "# for overlayfs and various bind mounts\nmount\numount2\npivot_root\n")
 }
@@ -165,7 +173,9 @@ func (s *GreengrassSupportInterfaceSuite) TestUdevTaggingDisablingRemoveLast(c *
 		s.containerModePlug,
 	} {
 		// make a spec with network-control that has udev tagging
-		spec := &udev.Specification{}
+		appSet, err := interfaces.NewSnapAppSet(s.extraPlug.Snap(), nil)
+		c.Assert(err, IsNil)
+		spec := udev.NewSpecification(appSet)
 		c.Assert(spec.AddConnectedPlug(builtin.MustInterface("network-control"), s.extraPlug, s.extraSlot), IsNil)
 		c.Assert(spec.Snippets(), HasLen, 3)
 
@@ -176,7 +186,9 @@ func (s *GreengrassSupportInterfaceSuite) TestUdevTaggingDisablingRemoveLast(c *
 }
 
 func (s *GreengrassSupportInterfaceSuite) TestProcessModeUdevTaggingWorks(c *C) {
-	spec := &udev.Specification{}
+	appSet, err := interfaces.NewSnapAppSet(s.processModePlug.Snap(), nil)
+	c.Assert(err, IsNil)
+	spec := udev.NewSpecification(appSet)
 	// connect the greengrass-support interface and ensure the spec is nil
 	c.Assert(spec.AddConnectedPlug(s.iface, s.processModePlug, s.slot), IsNil)
 	c.Check(spec.Snippets(), HasLen, 0)
@@ -191,7 +203,9 @@ func (s *GreengrassSupportInterfaceSuite) TestUdevTaggingDisablingRemoveFirst(c
 		s.plug,
 		s.containerModePlug,
 	} {
-		spec := &udev.Specification{}
+		appSet, err := interfaces.NewSnapAppSet(s.plug.Snap(), nil)
+		c.Assert(err, IsNil)
+		spec := udev.NewSpecification(appSet)
 		// connect the greengrass-support interface and ensure the spec is nil
 		c.Assert(spec.AddConnectedPlug(s.iface, s.plug, s.slot), IsNil)
 		c.Check(spec.Snippets(), HasLen, 0)
@@ -214,8 +228,10 @@ func (s *GreengrassSupportInterfaceSuite) TestPermanentSlotAppArmorSessionNative
 		s.plug,
 		s.containerModePlug,
 	} {
-		apparmorSpec := &apparmor.Specification{}
-		err := apparmorSpec.AddConnectedPlug(s.iface, plug, s.slot)
+		appSet, err := interfaces.NewSnapAppSet(plug.Snap(), nil)
+		c.Assert(err, IsNil)
+		apparmorSpec := apparmor.NewSpecification(appSet)
+		err = apparmorSpec.AddConnectedPlug(s.iface, plug, s.slot)
 		c.Assert(err, IsNil)
 		c.Assert(apparmorSpec.SecurityTags(), DeepEquals, []string{"snap.other.app2"})
 
@@ -232,8 +248,10 @@ func (s *GreengrassSupportInterfaceSuite) TestPermanentSlotAppArmorSessionClassi
 		s.plug,
 		s.containerModePlug,
 	} {
-		apparmorSpec := &apparmor.Specification{}
-		err := apparmorSpec.AddConnectedPlug(s.iface, plug, s.slot)
+		appSet, err := interfaces.NewSnapAppSet(plug.Snap(), nil)
+		c.Assert(err, IsNil)
+		apparmorSpec := apparmor.NewSpecification(appSet)
+		err = apparmorSpec.AddConnectedPlug(s.iface, plug, s.slot)
 		c.Assert(err, IsNil)
 		c.Assert(apparmorSpec.SecurityTags(), DeepEquals, []string{"snap.other.app2"})
 
