@@ -42,7 +42,8 @@ capability sys_module,
 @{PROC}/modules r,
 /{,usr/}bin/kmod ixr,
 
-# FIXME: moved to physical-memory-observe (remove this in series 20)
+# FIXME: moved to physical-memory-observe (will be removed when using different
+# policies for different bases)
 /dev/mem r,
 
 # Required to use SYSLOG_ACTION_READ_ALL and SYSLOG_ACTION_SIZE_BUFFER when
@@ -55,7 +56,7 @@ capability syslog,
 # Allow reading information about loaded kernel modules
 /sys/module/{,**} r,
 /etc/modprobe.d/{,**} r,
-/lib/modprobe.d/{,**} r,
+/{,usr/}lib/modprobe.d/{,**} r,
 `
 
 const kernelModuleControlConnectedPlugSecComp = `
@@ -79,6 +80,7 @@ func init() {
 		connectedPlugAppArmor: kernelModuleControlConnectedPlugAppArmor,
 		connectedPlugSecComp:  kernelModuleControlConnectedPlugSecComp,
 		connectedPlugUDev:     kernelModuleControlConnectedPlugUDev,
-		reservedForOS:         true,
+
+		usesSysModuleCapability: true,
 	})
 }
