@@ -20,15 +20,17 @@
 package builtin_test
 
 import (
+	"strings"
+
 	. "gopkg.in/check.v1"
 
 	"github.com/snapcore/snapd/interfaces"
 	"github.com/snapcore/snapd/interfaces/apparmor"
 	"github.com/snapcore/snapd/interfaces/builtin"
+	"github.com/snapcore/snapd/interfaces/ifacetest"
 	"github.com/snapcore/snapd/interfaces/seccomp"
 	apparmor_sandbox "github.com/snapcore/snapd/sandbox/apparmor"
 	"github.com/snapcore/snapd/snap"
-	"github.com/snapcore/snapd/snap/snaptest"
 	"github.com/snapcore/snapd/testutil"
 )
 
@@ -351,313 +353,388 @@ var _ = Suite(&PosixMQInterfaceSuite{
 func (s *PosixMQInterfaceSuite) SetUpTest(c *C) {
 	s.BaseTest.SetUpTest(c)
 
-	slotSnap := snaptest.MockInfo(c, slotSnapInfoYaml, nil)
+	slotSnapAppSet := ifacetest.MockInfoAndAppSet(c, slotSnapInfoYaml, nil, nil)
+	slotSnap := slotSnapAppSet.Info()
 
 	s.testReadWriteSlotInfo = slotSnap.Slots["test-rw"]
-	s.testReadWriteSlot = interfaces.NewConnectedSlot(s.testReadWriteSlotInfo, nil, nil)
+	s.testReadWriteSlot = interfaces.NewConnectedSlot(s.testReadWriteSlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testDefaultPermsSlotInfo = slotSnap.Slots["test-default"]
-	s.testDefaultPermsSlot = interfaces.NewConnectedSlot(s.testDefaultPermsSlotInfo, nil, nil)
+	s.testDefaultPermsSlot = interfaces.NewConnectedSlot(s.testDefaultPermsSlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testReadOnlySlotInfo = slotSnap.Slots["test-ro"]
-	s.testReadOnlySlot = interfaces.NewConnectedSlot(s.testReadOnlySlotInfo, nil, nil)
+	s.testReadOnlySlot = interfaces.NewConnectedSlot(s.testReadOnlySlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testAllPermsSlotInfo = slotSnap.Slots["test-all-perms"]
-	s.testAllPermsSlot = interfaces.NewConnectedSlot(s.testAllPermsSlotInfo, nil, nil)
+	s.testAllPermsSlot = interfaces.NewConnectedSlot(s.testAllPermsSlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testPathArraySlotInfo = slotSnap.Slots["test-path-array"]
-	s.testPathArraySlot = interfaces.NewConnectedSlot(s.testPathArraySlotInfo, nil, nil)
+	s.testPathArraySlot = interfaces.NewConnectedSlot(s.testPathArraySlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testEmptyPathArraySlotInfo = slotSnap.Slots["test-empty-path-array"]
-	s.testEmptyPathArraySlot = interfaces.NewConnectedSlot(s.testEmptyPathArraySlotInfo, nil, nil)
+	s.testEmptyPathArraySlot = interfaces.NewConnectedSlot(s.testEmptyPathArraySlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testOneEmptyPathArraySlotInfo = slotSnap.Slots["test-one-empty-path-array"]
-	s.testOneEmptyPathArraySlot = interfaces.NewConnectedSlot(s.testOneEmptyPathArraySlotInfo, nil, nil)
+	s.testOneEmptyPathArraySlot = interfaces.NewConnectedSlot(s.testOneEmptyPathArraySlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testEmptyPathSlotInfo = slotSnap.Slots["test-empty-path"]
-	s.testEmptyPathSlot = interfaces.NewConnectedSlot(s.testEmptyPathSlotInfo, nil, nil)
+	s.testEmptyPathSlot = interfaces.NewConnectedSlot(s.testEmptyPathSlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testInvalidPath1SlotInfo = slotSnap.Slots["test-invalid-path-1"]
-	s.testInvalidPath1Slot = interfaces.NewConnectedSlot(s.testInvalidPath1SlotInfo, nil, nil)
+	s.testInvalidPath1Slot = interfaces.NewConnectedSlot(s.testInvalidPath1SlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testInvalidPath2SlotInfo = slotSnap.Slots["test-invalid-path-2"]
-	s.testInvalidPath2Slot = interfaces.NewConnectedSlot(s.testInvalidPath2SlotInfo, nil, nil)
+	s.testInvalidPath2Slot = interfaces.NewConnectedSlot(s.testInvalidPath2SlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testInvalidPath3SlotInfo = slotSnap.Slots["test-invalid-path-3"]
-	s.testInvalidPath3Slot = interfaces.NewConnectedSlot(s.testInvalidPath3SlotInfo, nil, nil)
+	s.testInvalidPath3Slot = interfaces.NewConnectedSlot(s.testInvalidPath3SlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testInvalidPath4SlotInfo = slotSnap.Slots["test-invalid-path-4"]
-	s.testInvalidPath4Slot = interfaces.NewConnectedSlot(s.testInvalidPath4SlotInfo, nil, nil)
+	s.testInvalidPath4Slot = interfaces.NewConnectedSlot(s.testInvalidPath4SlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testInvalidPath5SlotInfo = slotSnap.Slots["test-invalid-path-5"]
-	s.testInvalidPath5Slot = interfaces.NewConnectedSlot(s.testInvalidPath5SlotInfo, nil, nil)
+	s.testInvalidPath5Slot = interfaces.NewConnectedSlot(s.testInvalidPath5SlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testInvalidPerms1SlotInfo = slotSnap.Slots["test-invalid-perms-1"]
-	s.testInvalidPerms1Slot = interfaces.NewConnectedSlot(s.testInvalidPerms1SlotInfo, nil, nil)
+	s.testInvalidPerms1Slot = interfaces.NewConnectedSlot(s.testInvalidPerms1SlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testInvalidPerms2SlotInfo = slotSnap.Slots["test-invalid-perms-2"]
-	s.testInvalidPerms2Slot = interfaces.NewConnectedSlot(s.testInvalidPerms2SlotInfo, nil, nil)
+	s.testInvalidPerms2Slot = interfaces.NewConnectedSlot(s.testInvalidPerms2SlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testInvalidPerms3SlotInfo = slotSnap.Slots["test-invalid-perms-3"]
-	s.testInvalidPerms3Slot = interfaces.NewConnectedSlot(s.testInvalidPerms3SlotInfo, nil, nil)
+	s.testInvalidPerms3Slot = interfaces.NewConnectedSlot(s.testInvalidPerms3SlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testLabelSlotInfo = slotSnap.Slots["test-label"]
-	s.testLabelSlot = interfaces.NewConnectedSlot(s.testLabelSlotInfo, nil, nil)
+	s.testLabelSlot = interfaces.NewConnectedSlot(s.testLabelSlotInfo, slotSnapAppSet, nil, nil)
 
 	s.testInvalidLabelSlotInfo = slotSnap.Slots["test-broken-label"]
-	s.testInvalidLabelSlot = interfaces.NewConnectedSlot(s.testInvalidLabelSlotInfo, nil, nil)
+	s.testInvalidLabelSlot = interfaces.NewConnectedSlot(s.testInvalidLabelSlotInfo, slotSnapAppSet, nil, nil)
 
-	plugSnap0 := snaptest.MockInfo(c, rwPlugSnapInfoYaml, nil)
+	plugSnap0AppSet := ifacetest.MockInfoAndAppSet(c, rwPlugSnapInfoYaml, nil, nil)
+	plugSnap0 := plugSnap0AppSet.Info()
 	s.testReadWritePlugInfo = plugSnap0.Plugs["test-rw"]
-	s.testReadWritePlug = interfaces.NewConnectedPlug(s.testReadWritePlugInfo, nil, nil)
+	s.testReadWritePlug = interfaces.NewConnectedPlug(s.testReadWritePlugInfo, plugSnap0AppSet, nil, nil)
 
-	plugSnap1 := snaptest.MockInfo(c, defaultRWPlugSnapInfoYaml, nil)
+	plugSnap1AppSet := ifacetest.MockInfoAndAppSet(c, defaultRWPlugSnapInfoYaml, nil, nil)
+	plugSnap1 := plugSnap1AppSet.Info()
 	s.testDefaultPermsPlugInfo = plugSnap1.Plugs["test-default"]
-	s.testDefaultPermsPlug = interfaces.NewConnectedPlug(s.testDefaultPermsPlugInfo, nil, nil)
+	s.testDefaultPermsPlug = interfaces.NewConnectedPlug(s.testDefaultPermsPlugInfo, plugSnap1AppSet, nil, nil)
 
-	plugSnap2 := snaptest.MockInfo(c, roPlugSnapInfoYaml, nil)
+	plugSnap2AppSet := ifacetest.MockInfoAndAppSet(c, roPlugSnapInfoYaml, nil, nil)
+	plugSnap2 := plugSnap2AppSet.Info()
 	s.testReadOnlyPlugInfo = plugSnap2.Plugs["test-ro"]
-	s.testReadOnlyPlug = interfaces.NewConnectedPlug(s.testReadOnlyPlugInfo, nil, nil)
+	s.testReadOnlyPlug = interfaces.NewConnectedPlug(s.testReadOnlyPlugInfo, plugSnap2AppSet, nil, nil)
 
-	plugSnap3 := snaptest.MockInfo(c, allPermsPlugSnapInfoYaml, nil)
+	plugSnap3AppSet := ifacetest.MockInfoAndAppSet(c, allPermsPlugSnapInfoYaml, nil, nil)
+	plugSnap3 := plugSnap3AppSet.Info()
 	s.testAllPermsPlugInfo = plugSnap3.Plugs["test-all-perms"]
-	s.testAllPermsPlug = interfaces.NewConnectedPlug(s.testAllPermsPlugInfo, nil, nil)
+	s.testAllPermsPlug = interfaces.NewConnectedPlug(s.testAllPermsPlugInfo, plugSnap3AppSet, nil, nil)
 
-	plugSnap4 := snaptest.MockInfo(c, invalidPerms1PlugSnapInfoYaml, nil)
+	plugSnap4AppSet := ifacetest.MockInfoAndAppSet(c, invalidPerms1PlugSnapInfoYaml, nil, nil)
+	plugSnap4 := plugSnap4AppSet.Info()
 	s.testInvalidPerms1PlugInfo = plugSnap4.Plugs["test-invalid-perms-1"]
-	s.testInvalidPerms1Plug = interfaces.NewConnectedPlug(s.testInvalidPerms1PlugInfo, nil, nil)
+	s.testInvalidPerms1Plug = interfaces.NewConnectedPlug(s.testInvalidPerms1PlugInfo, plugSnap4AppSet, nil, nil)
 
-	plugSnap5 := snaptest.MockInfo(c, testLabelPlugSnapInfoYaml, nil)
+	plugSnap5AppSet := ifacetest.MockInfoAndAppSet(c, testLabelPlugSnapInfoYaml, nil, nil)
+	plugSnap5 := plugSnap5AppSet.Info()
 	s.testLabelPlugInfo = plugSnap5.Plugs["test-label"]
-	s.testLabelPlug = interfaces.NewConnectedPlug(s.testLabelPlugInfo, nil, nil)
+	s.testLabelPlug = interfaces.NewConnectedPlug(s.testLabelPlugInfo, plugSnap5AppSet, nil, nil)
 
-	plugSnap6 := snaptest.MockInfo(c, testInvalidLabelPlugSnapInfoYaml, nil)
+	plugSnap6AppSet := ifacetest.MockInfoAndAppSet(c, testInvalidLabelPlugSnapInfoYaml, nil, nil)
+	plugSnap6 := plugSnap6AppSet.Info()
 	s.testInvalidLabelPlugInfo = plugSnap6.Plugs["test-invalid-label"]
-	s.testInvalidLabelPlug = interfaces.NewConnectedPlug(s.testInvalidLabelPlugInfo, nil, nil)
+	s.testInvalidLabelPlug = interfaces.NewConnectedPlug(s.testInvalidLabelPlugInfo, plugSnap6AppSet, nil, nil)
 
-	plugSnap7 := snaptest.MockInfo(c, invalidPerms3PlugSnapInfoYaml, nil)
+	plugSnap7AppSet := ifacetest.MockInfoAndAppSet(c, invalidPerms3PlugSnapInfoYaml, nil, nil)
+	plugSnap7 := plugSnap7AppSet.Info()
 	s.testInvalidPerms3PlugInfo = plugSnap7.Plugs["test-invalid-perms-3"]
-	s.testInvalidPerms3Plug = interfaces.NewConnectedPlug(s.testInvalidPerms3PlugInfo, nil, nil)
+	s.testInvalidPerms3Plug = interfaces.NewConnectedPlug(s.testInvalidPerms3PlugInfo, plugSnap7AppSet, nil, nil)
 
-	plugSnap8 := snaptest.MockInfo(c, pathArrayPlugSnapInfoYaml, nil)
+	plugSnap8AppSet := ifacetest.MockInfoAndAppSet(c, pathArrayPlugSnapInfoYaml, nil, nil)
+	plugSnap8 := plugSnap8AppSet.Info()
 	s.testPathArrayPlugInfo = plugSnap8.Plugs["test-path-array"]
-	s.testPathArrayPlug = interfaces.NewConnectedPlug(s.testPathArrayPlugInfo, nil, nil)
+	s.testPathArrayPlug = interfaces.NewConnectedPlug(s.testPathArrayPlugInfo, plugSnap8AppSet, nil, nil)
+}
+
+// splitSnippet converts the trimmed string snippet to a string slice
+func splitSnippet(snippet string) []string {
+	return strings.Split(strings.TrimSpace(snippet), "\n")
 }
 
 func (s *PosixMQInterfaceSuite) checkSlotSeccompSnippet(c *C, spec *seccomp.Specification) {
 	slotSnippet := spec.SnippetForTag("snap.producer.app")
+
+	c.Check(splitSnippet(slotSnippet), HasLen, 8)
 	c.Check(slotSnippet, testutil.Contains, "mq_open")
 	c.Check(slotSnippet, testutil.Contains, "mq_unlink")
 	c.Check(slotSnippet, testutil.Contains, "mq_getsetattr")
 	c.Check(slotSnippet, testutil.Contains, "mq_notify")
 	c.Check(slotSnippet, testutil.Contains, "mq_timedreceive")
+	c.Check(slotSnippet, testutil.Contains, "mq_timedreceive_time64")
 	c.Check(slotSnippet, testutil.Contains, "mq_timedsend")
+	c.Check(slotSnippet, testutil.Contains, "mq_timedsend_time64")
 }
 
 func (s *PosixMQInterfaceSuite) TestReadWriteMQAppArmor(c *C) {
-	spec := &apparmor.Specification{}
+	spec := apparmor.NewSpecification(s.testReadWriteSlot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testReadWriteSlotInfo)
 	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testReadWritePlug, s.testReadWriteSlot)
-	c.Assert(err, IsNil)
-	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app", "snap.producer.app"})
+
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.producer.app"})
 
 	slotSnippet := spec.SnippetForTag("snap.producer.app")
 	c.Check(slotSnippet, testutil.Contains, `# POSIX Message Queue slot: test-rw`)
-	c.Check(slotSnippet, testutil.Contains, `mqueue (open read write create delete) "/test-rw",`)
+	c.Check(slotSnippet, testutil.Contains, `mqueue (open read write create delete getattr setattr) type=posix "/test-rw",`)
+
+	spec = apparmor.NewSpecification(s.testReadOnlyPlug.AppSet())
+	err = spec.AddConnectedPlug(s.iface, s.testReadWritePlug, s.testReadWriteSlot)
+	c.Assert(err, IsNil)
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
 
 	plugSnippet := spec.SnippetForTag("snap.consumer.app")
 	c.Check(plugSnippet, testutil.Contains, `# POSIX Message Queue plug: test-rw`)
-	c.Check(plugSnippet, testutil.Contains, `mqueue (read write open) "/test-rw",`)
+	c.Check(plugSnippet, testutil.Contains, `mqueue (read write open getattr setattr) type=posix "/test-rw",`)
 }
 
 func (s *PosixMQInterfaceSuite) TestReadWriteMQSeccomp(c *C) {
-	spec := &seccomp.Specification{}
+	spec := seccomp.NewSpecification(s.testReadWriteSlot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testReadWriteSlotInfo)
 	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testReadWritePlug, s.testReadWriteSlot)
-	c.Assert(err, IsNil)
-	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app", "snap.producer.app"})
+
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.producer.app"})
 
 	s.checkSlotSeccompSnippet(c, spec)
+
+	spec = seccomp.NewSpecification(s.testReadWritePlug.AppSet())
+	err = spec.AddConnectedPlug(s.iface, s.testReadWritePlug, s.testReadWriteSlot)
+	c.Assert(err, IsNil)
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
+
 	plugSnippet := spec.SnippetForTag("snap.consumer.app")
+	c.Check(splitSnippet(plugSnippet), HasLen, 7)
 	c.Check(plugSnippet, testutil.Contains, "mq_open")
 	c.Check(plugSnippet, testutil.Contains, "mq_notify")
 	c.Check(plugSnippet, testutil.Contains, "mq_timedreceive")
+	c.Check(plugSnippet, testutil.Contains, "mq_timedreceive_time64")
 	c.Check(plugSnippet, testutil.Contains, "mq_timedsend")
+	c.Check(plugSnippet, testutil.Contains, "mq_timedsend_time64")
 	c.Check(plugSnippet, testutil.Contains, "mq_getsetattr")
 	c.Check(plugSnippet, Not(testutil.Contains), "mq_unlink")
 }
 
 func (s *PosixMQInterfaceSuite) TestDefaultReadWriteMQAppArmor(c *C) {
-	spec := &apparmor.Specification{}
+	spec := apparmor.NewSpecification(s.testDefaultPermsSlot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testDefaultPermsSlotInfo)
 	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testDefaultPermsPlug, s.testDefaultPermsSlot)
-	c.Assert(err, IsNil)
-	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app", "snap.producer.app"})
+
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.producer.app"})
 
 	slotSnippet := spec.SnippetForTag("snap.producer.app")
 	c.Check(slotSnippet, testutil.Contains, `# POSIX Message Queue slot: test-default`)
-	c.Check(slotSnippet, testutil.Contains, `mqueue (open read write create delete) "/test-default",`)
+	c.Check(slotSnippet, testutil.Contains, `mqueue (open read write create delete getattr setattr) type=posix "/test-default",`)
+
+	spec = apparmor.NewSpecification(s.testDefaultPermsPlug.AppSet())
+	err = spec.AddConnectedPlug(s.iface, s.testDefaultPermsPlug, s.testDefaultPermsSlot)
+	c.Assert(err, IsNil)
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
 
 	plugSnippet := spec.SnippetForTag("snap.consumer.app")
 	c.Check(plugSnippet, testutil.Contains, `# POSIX Message Queue plug: test-default`)
-	c.Check(plugSnippet, testutil.Contains, `mqueue (read write open) "/test-default",`)
+	c.Check(plugSnippet, testutil.Contains, `mqueue (read write open getattr setattr) type=posix "/test-default",`)
 }
 
 func (s *PosixMQInterfaceSuite) TestDefaultReadWriteMQSeccomp(c *C) {
-	spec := &seccomp.Specification{}
+	spec := seccomp.NewSpecification(s.testDefaultPermsSlot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testDefaultPermsSlotInfo)
 	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testDefaultPermsPlug, s.testDefaultPermsSlot)
-	c.Assert(err, IsNil)
-	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app", "snap.producer.app"})
 
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.producer.app"})
 	s.checkSlotSeccompSnippet(c, spec)
 
+	spec = seccomp.NewSpecification(s.testDefaultPermsPlug.AppSet())
+	err = spec.AddConnectedPlug(s.iface, s.testDefaultPermsPlug, s.testDefaultPermsSlot)
+	c.Assert(err, IsNil)
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
+
 	plugSnippet := spec.SnippetForTag("snap.consumer.app")
+	c.Check(splitSnippet(plugSnippet), HasLen, 7)
 	c.Check(plugSnippet, testutil.Contains, "mq_open")
 	c.Check(plugSnippet, testutil.Contains, "mq_notify")
 	c.Check(plugSnippet, testutil.Contains, "mq_timedreceive")
+	c.Check(plugSnippet, testutil.Contains, "mq_timedreceive_time64")
 	c.Check(plugSnippet, testutil.Contains, "mq_timedsend")
+	c.Check(plugSnippet, testutil.Contains, "mq_timedsend_time64")
 	c.Check(plugSnippet, testutil.Contains, "mq_getsetattr")
 	c.Check(plugSnippet, Not(testutil.Contains), "mq_unlink")
 }
 
 func (s *PosixMQInterfaceSuite) TestReadOnlyMQAppArmor(c *C) {
-	spec := &apparmor.Specification{}
+	spec := apparmor.NewSpecification(s.testReadOnlySlot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testReadOnlySlotInfo)
 	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testReadOnlyPlug, s.testReadOnlySlot)
-	c.Assert(err, IsNil)
-	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app", "snap.producer.app"})
+
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.producer.app"})
 
 	slotSnippet := spec.SnippetForTag("snap.producer.app")
-	c.Check(slotSnippet, testutil.Contains, `mqueue (open read write create delete) "/test-ro",`)
+	c.Check(slotSnippet, testutil.Contains, `mqueue (open read write create delete getattr setattr) type=posix "/test-ro",`)
+
+	spec = apparmor.NewSpecification(s.testReadOnlyPlug.AppSet())
+	err = spec.AddConnectedPlug(s.iface, s.testReadOnlyPlug, s.testReadOnlySlot)
+	c.Assert(err, IsNil)
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
 
 	plugSnippet := spec.SnippetForTag("snap.consumer.app")
-	c.Check(plugSnippet, testutil.Contains, `mqueue (read open) "/test-ro",`)
+	c.Check(plugSnippet, testutil.Contains, `mqueue (read open getattr) type=posix "/test-ro",`)
 }
 
 func (s *PosixMQInterfaceSuite) TestReadOnlyMQSeccomp(c *C) {
-	spec := &seccomp.Specification{}
+	spec := seccomp.NewSpecification(s.testReadOnlySlot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testReadOnlySlotInfo)
 	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testReadOnlyPlug, s.testReadOnlySlot)
-	c.Assert(err, IsNil)
-	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app", "snap.producer.app"})
+
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.producer.app"})
 
 	s.checkSlotSeccompSnippet(c, spec)
 
+	spec = seccomp.NewSpecification(s.testReadOnlyPlug.AppSet())
+	err = spec.AddConnectedPlug(s.iface, s.testReadOnlyPlug, s.testReadOnlySlot)
+	c.Assert(err, IsNil)
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
+
 	plugSnippet := spec.SnippetForTag("snap.consumer.app")
+	c.Check(splitSnippet(plugSnippet), HasLen, 5)
 	c.Check(plugSnippet, testutil.Contains, "mq_open")
 	c.Check(plugSnippet, testutil.Contains, "mq_notify")
 	c.Check(plugSnippet, testutil.Contains, "mq_timedreceive")
+	c.Check(plugSnippet, testutil.Contains, "mq_timedreceive_time64")
 	c.Check(plugSnippet, testutil.Contains, "mq_getsetattr")
 	c.Check(plugSnippet, Not(testutil.Contains), "mq_timedsend")
+	c.Check(plugSnippet, Not(testutil.Contains), "mq_timedsend_time64")
 	c.Check(plugSnippet, Not(testutil.Contains), "mq_unlink")
 }
 
 func (s *PosixMQInterfaceSuite) TestPathArrayMQAppArmor(c *C) {
-	spec := &apparmor.Specification{}
+	spec := apparmor.NewSpecification(s.testPathArraySlot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testPathArraySlotInfo)
 	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testPathArrayPlug, s.testPathArraySlot)
-	c.Assert(err, IsNil)
-	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app", "snap.producer.app"})
+
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.producer.app"})
 
 	slotSnippet := spec.SnippetForTag("snap.producer.app")
-	c.Check(slotSnippet, testutil.Contains, `  mqueue (open read write create delete) "/test-array-1",
-  mqueue (open read write create delete) "/test-array-2",
-  mqueue (open read write create delete) "/test-array-3",
+	c.Check(slotSnippet, testutil.Contains, `  mqueue (open read write create delete getattr setattr) type=posix "/test-array-1",
+  mqueue (open read write create delete getattr setattr) type=posix "/test-array-2",
+  mqueue (open read write create delete getattr setattr) type=posix "/test-array-3",
 `)
 
+	spec = apparmor.NewSpecification(s.testPathArrayPlug.AppSet())
+	err = spec.AddConnectedPlug(s.iface, s.testPathArrayPlug, s.testPathArraySlot)
+	c.Assert(err, IsNil)
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
+
 	plugSnippet := spec.SnippetForTag("snap.consumer.app")
-	c.Check(plugSnippet, testutil.Contains, `  mqueue (read write open) "/test-array-1",
-  mqueue (read write open) "/test-array-2",
-  mqueue (read write open) "/test-array-3",
+	c.Check(plugSnippet, testutil.Contains, `  mqueue (read write open getattr setattr) type=posix "/test-array-1",
+  mqueue (read write open getattr setattr) type=posix "/test-array-2",
+  mqueue (read write open getattr setattr) type=posix "/test-array-3",
 `)
 }
 
 func (s *PosixMQInterfaceSuite) TestPathArrayMQSeccomp(c *C) {
-	spec := &seccomp.Specification{}
+	spec := seccomp.NewSpecification(s.testPathArraySlot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testPathArraySlotInfo)
 	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testPathArrayPlug, s.testPathArraySlot)
-	c.Assert(err, IsNil)
-	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app", "snap.producer.app"})
+
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.producer.app"})
 
 	s.checkSlotSeccompSnippet(c, spec)
 
+	spec = seccomp.NewSpecification(s.testPathArrayPlug.AppSet())
+	err = spec.AddConnectedPlug(s.iface, s.testPathArrayPlug, s.testPathArraySlot)
+	c.Assert(err, IsNil)
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
+
 	plugSnippet := spec.SnippetForTag("snap.consumer.app")
+	c.Check(splitSnippet(plugSnippet), HasLen, 7)
 	c.Check(plugSnippet, testutil.Contains, "mq_open")
 	c.Check(plugSnippet, testutil.Contains, "mq_notify")
 	c.Check(plugSnippet, testutil.Contains, "mq_timedreceive")
+	c.Check(plugSnippet, testutil.Contains, "mq_timedreceive_time64")
 	c.Check(plugSnippet, testutil.Contains, "mq_timedsend")
+	c.Check(plugSnippet, testutil.Contains, "mq_timedsend_time64")
 	c.Check(plugSnippet, testutil.Contains, "mq_getsetattr")
 	c.Check(plugSnippet, Not(testutil.Contains), "mq_unlink")
 }
 
 func (s *PosixMQInterfaceSuite) TestAllPermsMQAppArmor(c *C) {
-	spec := &apparmor.Specification{}
+	spec := apparmor.NewSpecification(s.testAllPermsSlot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testAllPermsSlotInfo)
 	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testAllPermsPlug, s.testAllPermsSlot)
-	c.Assert(err, IsNil)
-	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app", "snap.producer.app"})
+
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.producer.app"})
 
 	slotSnippet := spec.SnippetForTag("snap.producer.app")
-	c.Check(slotSnippet, testutil.Contains, `mqueue (open read write create delete) "/test-all-perms",`)
+	c.Check(slotSnippet, testutil.Contains, `mqueue (open read write create delete getattr setattr) type=posix "/test-all-perms",`)
+
+	spec = apparmor.NewSpecification(s.testAllPermsPlug.AppSet())
+	err = spec.AddConnectedPlug(s.iface, s.testAllPermsPlug, s.testAllPermsSlot)
+	c.Assert(err, IsNil)
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
 
 	plugSnippet := spec.SnippetForTag("snap.consumer.app")
-	c.Check(plugSnippet, testutil.Contains, `mqueue (create delete read write open) "/test-all-perms",`)
+	c.Check(plugSnippet, testutil.Contains, `mqueue (create delete read write open getattr setattr) type=posix "/test-all-perms",`)
 }
 
 func (s *PosixMQInterfaceSuite) TestAllPermsMQSeccomp(c *C) {
-	spec := &seccomp.Specification{}
+	spec := seccomp.NewSpecification(s.testAllPermsSlot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testAllPermsSlotInfo)
 	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testAllPermsPlug, s.testAllPermsSlot)
-	c.Assert(err, IsNil)
-	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app", "snap.producer.app"})
 
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.producer.app"})
 	s.checkSlotSeccompSnippet(c, spec)
 
+	spec = seccomp.NewSpecification(s.testAllPermsPlug.AppSet())
+	err = spec.AddConnectedPlug(s.iface, s.testAllPermsPlug, s.testAllPermsSlot)
+	c.Assert(err, IsNil)
+	c.Assert(spec.SecurityTags(), DeepEquals, []string{"snap.consumer.app"})
+
 	plugSnippet := spec.SnippetForTag("snap.consumer.app")
+	c.Check(splitSnippet(plugSnippet), HasLen, 8)
 	c.Check(plugSnippet, testutil.Contains, "mq_open")
 	c.Check(plugSnippet, testutil.Contains, "mq_unlink")
 	c.Check(plugSnippet, testutil.Contains, "mq_getsetattr")
 	c.Check(plugSnippet, testutil.Contains, "mq_notify")
 	c.Check(plugSnippet, testutil.Contains, "mq_timedreceive")
+	c.Check(plugSnippet, testutil.Contains, "mq_timedreceive_time64")
 	c.Check(plugSnippet, testutil.Contains, "mq_timedsend")
+	c.Check(plugSnippet, testutil.Contains, "mq_timedsend_time64")
 }
 
 func (s *PosixMQInterfaceSuite) TestPathValidationPosixMQ(c *C) {
-	spec := &apparmor.Specification{}
+	spec := apparmor.NewSpecification(s.testInvalidPath1Slot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testInvalidPath1SlotInfo)
 	c.Check(err, ErrorMatches,
 		`posix-mq "path" attribute must conform to the POSIX message queue name specifications \(see "man mq_overview"\): /../../test-invalid`)
 }
 
 func (s *PosixMQInterfaceSuite) TestPathValidationAppArmorRegex(c *C) {
-	spec := &apparmor.Specification{}
+	spec := apparmor.NewSpecification(s.testInvalidPath2Slot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testInvalidPath2SlotInfo)
 	c.Check(err, ErrorMatches, `posix-mq "path" attribute is invalid: /test-invalid-2"\["`)
 }
 
 func (s *PosixMQInterfaceSuite) TestPathStringValidation(c *C) {
-	spec := &apparmor.Specification{}
+	spec := apparmor.NewSpecification(s.testInvalidPath3Slot.AppSet())
 	err := spec.AddPermanentSlot(s.iface, s.testInvalidPath3SlotInfo)
 	c.Check(err, ErrorMatches, `snap "producer" has interface "posix-mq" with invalid value type map\[string\]interface {} for "path" attribute: \*\[\]string`)
 }
 
 func (s *PosixMQInterfaceSuite) TestInvalidPerms1(c *C) {
-	spec := &apparmor.Specification{}
+	spec := apparmor.NewSpecification(s.testInvalidPerms1Slot.AppSet())
 	// The slot should function correctly here as it receives the full list
 	// of built-in permissions, not what's listed in the configuration
 	err := spec.AddPermanentSlot(s.iface, s.testInvalidPerms1SlotInfo)
 	c.Assert(err, IsNil)
+
+	spec = apparmor.NewSpecification(s.testInvalidPerms1Plug.AppSet())
 	// The plug should fail to connect as it receives the given list of
 	// invalid permissions
 	err = spec.AddConnectedPlug(s.iface, s.testInvalidPerms1Plug, s.testInvalidPerms1Slot)
@@ -666,10 +743,8 @@ func (s *PosixMQInterfaceSuite) TestInvalidPerms1(c *C) {
 }
 
 func (s *PosixMQInterfaceSuite) TestInvalidPerms3(c *C) {
-	spec := &apparmor.Specification{}
-	err := spec.AddPermanentSlot(s.iface, s.testInvalidPerms3SlotInfo)
-	c.Assert(err, IsNil)
-	err = spec.AddConnectedPlug(s.iface, s.testInvalidPerms3Plug, s.testInvalidPerms3Slot)
+	spec := apparmor.NewSpecification(s.testInvalidPerms3Plug.AppSet())
+	err := spec.AddConnectedPlug(s.iface, s.testInvalidPerms3Plug, s.testInvalidPerms3Slot)
 	c.Check(err, ErrorMatches,
 		`snap "producer" has interface "posix-mq" with invalid value type \[\]interface {} for "permissions" attribute: \*\[\]string`)
 }
@@ -705,7 +780,7 @@ func (s *PosixMQInterfaceSuite) checkPlugPosixMQAttr(c *C, plug *snap.PlugInfo)
 
 func (s *PosixMQInterfaceSuite) TestSanitizeSlot(c *C) {
 	// Ensure that the mqueue feature is detected
-	restore := apparmor_sandbox.MockFeatures([]string{}, nil, []string{"mqueue"}, nil)
+	restore := apparmor_sandbox.MockFeatures([]string{}, nil, []string{"mqueue-posix"}, nil)
 	defer restore()
 
 	c.Assert(interfaces.BeforePrepareSlot(s.iface, s.testReadWriteSlotInfo), IsNil)
@@ -748,7 +823,7 @@ func (s *PosixMQInterfaceSuite) TestSanitizeSlot(c *C) {
 
 func (s *PosixMQInterfaceSuite) TestSanitizePlug(c *C) {
 	// Ensure that the mqueue feature is detected
-	restore := apparmor_sandbox.MockFeatures([]string{}, nil, []string{"mqueue"}, nil)
+	restore := apparmor_sandbox.MockFeatures([]string{}, nil, []string{"mqueue-posix"}, nil)
 	defer restore()
 
 	c.Assert(interfaces.BeforePreparePlug(s.iface, s.testReadWritePlugInfo), IsNil)
