@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016 Canonical Ltd
+ * Copyright (C) 2016-2017 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -30,11 +30,18 @@ const tpmBaseDeclarationSlots = `
 `
 
 const tpmConnectedPlugAppArmor = `
-# Description: for those who need to talk to the system TPM chip over /dev/tpm0
+# Description: for those who need to talk to the system TPM chip over
+# /dev/tpm[0-9]* and kernel TPM resource manager /dev/tpmrm[0-0]* (4.12+)
 
-/dev/tpm0 rw,
+/dev/tpm[0-9]* rw,
+/dev/tpmrm[0-9]* rw,
 `
 
+var tpmConnectedPlugUDev = []string{
+	`KERNEL=="tpm[0-9]*"`,
+	`KERNEL=="tpmrm[0-9]*"`,
+}
+
 func init() {
 	registerIface(&commonInterface{
 		name:                  "tpm",
@@ -43,6 +50,6 @@ func init() {
 		implicitOnClassic:     true,
 		baseDeclarationSlots:  tpmBaseDeclarationSlots,
 		connectedPlugAppArmor: tpmConnectedPlugAppArmor,
-		reservedForOS:         true,
+		connectedPlugUDev:     tpmConnectedPlugUDev,
 	})
 }
