@@ -112,6 +112,11 @@ var u2fDevices = []u2fDevice{
 		VendorIDPattern:  "20a0",
 		ProductIDPattern: "42b2",
 	},
+	{
+		Name:             "Nitrokey Passkey",
+		VendorIDPattern:  "20a0",
+		ProductIDPattern: "42f3",
+	},
 	{
 		Name:             "Google Titan U2F",
 		VendorIDPattern:  "18d1",
@@ -195,7 +200,7 @@ var u2fDevices = []u2fDevice{
 	{
 		Name:             "OneSpan DIGIPASS FX Series",
 		VendorIDPattern:  "1a44",
-		ProductIDPattern: "1501|1502|1503|1506|1507",
+		ProductIDPattern: "1501|1502|1503|1506|1507|1508|1509|150A",
 	},
 }
 
