@@ -23,7 +23,6 @@ package requestprompts
 
 import (
 	"encoding/json"
-	"errors"
 	"fmt"
 	"os"
 	"path/filepath"
@@ -32,16 +31,27 @@ import (
 
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/interfaces/prompting"
+	prompting_errors "github.com/snapcore/snapd/interfaces/prompting/errors"
 	"github.com/snapcore/snapd/interfaces/prompting/internal/maxidmmap"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/sandbox/apparmor/notify/listener"
 	"github.com/snapcore/snapd/strutil"
+	"github.com/snapcore/snapd/timeutil"
 )
 
-var (
-	ErrClosed         = errors.New("prompt DB has already been closed")
-	ErrNotFound       = errors.New("cannot find prompt with the given ID for the given user")
-	ErrTooManyPrompts = errors.New("cannot add new prompt, too many outstanding")
+const (
+	// initialTimeout is the duration before which prompts for a given user
+	// will expire if there has been no retrieval of prompt details for that
+	// user since the previous timeout, or if the user prompt DB was just
+	// created.
+	initialTimeout = 10 * time.Second
+	// activityTimeout is the duration before which prompts for a given user
+	// will expire after the most recent retrieval of prompt details for that
+	// user.
+	activityTimeout = 10 * time.Minute
+	// maxOutstandingPromptsPerUser is an arbitrary limit.
+	// TODO: review this limit after some usage.
+	maxOutstandingPromptsPerUser int = 1000
 )
 
 // Prompt contains information about a request for which a user should be
@@ -76,7 +86,7 @@ type jsonPromptConstraints struct {
 func (p *Prompt) MarshalJSON() ([]byte, error) {
 	constraints := &jsonPromptConstraints{
 		Path:                 p.Constraints.path,
-		RequestedPermissions: p.Constraints.remainingPermissions,
+		RequestedPermissions: p.Constraints.outstandingPermissions,
 		AvailablePermissions: p.Constraints.availablePermissions,
 	}
 	toMarshal := &jsonPrompt{
@@ -89,9 +99,43 @@ func (p *Prompt) MarshalJSON() ([]byte, error) {
 	return json.Marshal(toMarshal)
 }
 
+func (p *Prompt) sendReply(outcome prompting.OutcomeType) error {
+	allow, err := outcome.AsBool()
+	if err != nil {
+		// This should not occur
+		return err
+	}
+	// Reply with any permissions which were previously allowed
+	// If outcome is allow, then reply by allowing all originally-requested
+	// permissions. If outcome is deny, only allow permissions which were
+	// originally requested but have since been allowed by rules, and deny any
+	// outstanding permissions.
+	var deniedPermissions []string
+	if !allow {
+		deniedPermissions = p.Constraints.outstandingPermissions
+	}
+	allowedPermission := p.Constraints.buildResponse(p.Interface, deniedPermissions)
+	return p.sendReplyWithPermission(allowedPermission)
+}
+
+func (p *Prompt) sendReplyWithPermission(allowedPermission any) error {
+	for _, listenerReq := range p.listenerReqs {
+		if err := sendReply(listenerReq, allowedPermission); err != nil {
+			// Error should only occur if reply is malformed, and since these
+			// listener requests should be identical, if a reply is malformed
+			// for one, it should be malformed for all. Malformed replies should
+			// leave the listener request unchanged. Thus, return early.
+			return err
+		}
+	}
+	return nil
+}
+
+var sendReply = (*listener.Request).Reply
+
 // promptConstraints store the path which was requested, along with three
 // lists of permissions: the original permissions associated with the request,
-// the remaining unsatisfied permissions (as rules may satisfy some of the
+// the outstanding unsatisfied permissions (as rules may satisfy some of the
 // permissions from a prompt before the prompt is fully resolved), and the
 // available permissions for the interface associated with the prompt, so that
 // the client may reply with a broader set of permissions than was originally
@@ -99,9 +143,9 @@ func (p *Prompt) MarshalJSON() ([]byte, error) {
 type promptConstraints struct {
 	// path is the path to which the application is requesting access.
 	path string
-	// remainingPermissions are the remaining unsatisfied permissions for which
-	// the application is requesting access.
-	remainingPermissions []string
+	// outstandingPermissions are the outstanding unsatisfied permissions for
+	// which the application is requesting access.
+	outstandingPermissions []string
 	// availablePermissions are the permissions which are supported by the
 	// interface associated with the prompt to which the constraints apply.
 	availablePermissions []string
@@ -132,20 +176,91 @@ func (pc *promptConstraints) equals(other *promptConstraints) bool {
 	return true
 }
 
-// subtractPermissions removes all of the given permissions from the list of
-// permissions in the constraints.
-func (pc *promptConstraints) subtractPermissions(permissions []string) (modified bool) {
-	newPermissions := make([]string, 0, len(pc.remainingPermissions))
-	for _, perm := range pc.remainingPermissions {
-		if !strutil.ListContains(permissions, perm) {
-			newPermissions = append(newPermissions, perm)
+// applyRuleConstraints modifies the prompt constraints, removing any outstanding
+// permissions which are matched by the given rule constraints.
+//
+// Returns whether the prompt constraints were affected by the rule constraints,
+// whether the prompt requires a response (either because all permissions were
+// allowed or at least one permission was denied), and the list of any
+// permissions which were denied. If an error occurs, it is returned, and the
+// other return values can be ignored.
+//
+// If the path pattern does not match the prompt path, or the permissions in
+// the rule constraints do not include any of the outstanding prompt permissions,
+// then affectedByRule is false, and no changes are made to the prompt
+// constraints.
+func (pc *promptConstraints) applyRuleConstraints(constraints *prompting.RuleConstraints) (affectedByRule, respond bool, deniedPermissions []string, err error) {
+	pathMatched, err := constraints.Match(pc.path)
+	if err != nil {
+		// Should not occur, only error is if path pattern is malformed,
+		// which would have thrown an error while parsing, not now.
+		return false, false, nil, err
+	}
+	if !pathMatched {
+		return false, false, nil, nil
+	}
+
+	// Path pattern matched, now check if any permissions match
+
+	newOutstandingPermissions := make([]string, 0, len(pc.outstandingPermissions))
+	for _, perm := range pc.outstandingPermissions {
+		entry, exists := constraints.Permissions[perm]
+		if !exists {
+			// Permission not covered by rule constraints, so permission
+			// should continue to be in outstandingPermissions.
+			newOutstandingPermissions = append(newOutstandingPermissions, perm)
+			continue
+		}
+		affectedByRule = true
+		allow, err := entry.Outcome.AsBool()
+		if err != nil {
+			// This should not occur, as rule constraints are built internally
+			return false, false, nil, err
+		}
+		if !allow {
+			deniedPermissions = append(deniedPermissions, perm)
 		}
 	}
-	if len(newPermissions) != len(pc.remainingPermissions) {
-		pc.remainingPermissions = newPermissions
-		return true
+	if !affectedByRule {
+		// No permissions matched, so nothing changes, no need to record a
+		// notice or send a response.
+		return false, false, nil, nil
 	}
-	return false
+
+	pc.outstandingPermissions = newOutstandingPermissions
+
+	if len(pc.outstandingPermissions) == 0 || len(deniedPermissions) > 0 {
+		// All permissions allowed or at least one permission denied, so tell
+		// the caller to send a response back to the kernel.
+		respond = true
+	}
+
+	return affectedByRule, respond, deniedPermissions, nil
+}
+
+// buildResponse creates a listener response to the receiving prompt constraints
+// based on the given interface and list of denied permissions.
+//
+// The response is the AppArmor permission which should be allowed. This
+// corresponds to the originally requested permissions from the prompt
+// constraints, except with all denied permissions removed.
+func (pc *promptConstraints) buildResponse(iface string, deniedPermissions []string) any {
+	allowedPerms := pc.originalPermissions
+	if len(deniedPermissions) > 0 {
+		allowedPerms = make([]string, 0, len(pc.originalPermissions)-len(deniedPermissions))
+		for _, perm := range pc.originalPermissions {
+			if !strutil.ListContains(deniedPermissions, perm) {
+				allowedPerms = append(allowedPerms, perm)
+			}
+		}
+	}
+	allowedPermission, err := prompting.AbstractPermissionsToAppArmorPermissions(iface, allowedPerms)
+	if err != nil {
+		// This should not occur, but if so, permission should be set to the
+		// empty value for its corresponding permission type.
+		logger.Noticef("internal error: cannot convert abstract permissions to AppArmor permissions: %v", err)
+	}
+	return allowedPermission
 }
 
 // Path returns the path associated with the request to which the receiving
@@ -154,10 +269,10 @@ func (pc *promptConstraints) Path() string {
 	return pc.path
 }
 
-// Permissions returns the remaining unsatisfied permissions associated with
-// the prompt.
-func (pc *promptConstraints) RemainingPermissions() []string {
-	return pc.remainingPermissions
+// OutstandingPermissions returns the outstanding unsatisfied permissions
+// associated with the prompt.
+func (pc *promptConstraints) OutstandingPermissions() []string {
+	return pc.outstandingPermissions
 }
 
 // userPromptDB maps prompt IDs to prompts for a single user.
@@ -166,13 +281,15 @@ type userPromptDB struct {
 	ids map[prompting.IDType]int
 	// prompts is the list of prompts which apply to the given user.
 	prompts []*Prompt
+	// expirationTimer clears the prompts for the given user when it expires.
+	expirationTimer timeutil.Timer
 }
 
 // get returns the prompt with the given ID from the user prompt DB.
 func (udb *userPromptDB) get(id prompting.IDType) (*Prompt, error) {
 	index, ok := udb.ids[id]
 	if !ok {
-		return nil, ErrNotFound
+		return nil, prompting_errors.ErrPromptNotFound
 	}
 	return udb.prompts[index], nil
 }
@@ -195,7 +312,7 @@ func (udb *userPromptDB) add(prompt *Prompt) {
 func (udb *userPromptDB) remove(id prompting.IDType) (*Prompt, error) {
 	index, ok := udb.ids[id]
 	if !ok {
-		return nil, ErrNotFound
+		return nil, prompting_errors.ErrPromptNotFound
 	}
 	prompt := udb.prompts[index]
 	// Remove the prompt with the given ID by copying the final prompt in
@@ -212,6 +329,57 @@ func (udb *userPromptDB) remove(id prompting.IDType) (*Prompt, error) {
 	return prompt, nil
 }
 
+// timeoutCallback is the function which should be called when the expiration
+// timer for the receiving user prompt DB expires. This method should never be
+// called directly outside of a `time.AfterFunc` call which initializes the
+// timer for the user prompt DB when it is first created.
+func (udb *userPromptDB) timeoutCallback(pdb *PromptDB, user uint32) {
+	pdb.mutex.Lock()
+	// We don't defer Unlock() since we may need to manually unlock later in
+	// the function in order to record a notice and send a reply without
+	// holding the DB lock.
+
+	// If the DB has been closed, do nothing. Thus, there's no need to stop the
+	// expiration timer when the DB has been closed, since the timer will fire
+	// and do nothing.
+	if pdb.isClosed() {
+		pdb.mutex.Unlock()
+		return
+	}
+	// Restart expiration timer while holding the lock, so we don't
+	// overwrite a newly-set activity timeout with an initial timeout.
+	// With the lock held, no activity can occur, so no activity timeout
+	// can be set.
+	if udb.expirationTimer.Reset(initialTimeout) {
+		// Timer was active again, suggesting that some activity caused
+		// the timer to be reset at some point between the timer firing
+		// and the lock being released and subsequently acquired by this
+		// function. So reset the timer to activityTimeout, and do not
+		// purge prompts.
+		udb.activityResetExpiration()
+		pdb.mutex.Unlock()
+		return
+	}
+	expiredPrompts := udb.prompts
+	// Clear all outstanding prompts for the user
+	udb.prompts = nil
+	udb.ids = make(map[prompting.IDType]int) // TODO: clear() once we're on Go 1.21+
+	// Unlock now so we can record notices without holding the prompt DB lock
+	pdb.mutex.Unlock()
+	data := map[string]string{"resolved": "expired"}
+	for _, p := range expiredPrompts {
+		pdb.notifyPrompt(user, p.ID, data)
+		p.sendReply(prompting.OutcomeDeny) // ignore any error, should not occur
+	}
+}
+
+// activityResetExpiration resets the expiration timer for prompts for the
+// receiving user prompt DB. Returns true if the timer had been active, false
+// if the timer had expired or been stopped.
+func (udb *userPromptDB) activityResetExpiration() bool {
+	return udb.expirationTimer.Reset(activityTimeout)
+}
+
 // PromptDB stores outstanding prompts in memory and ensures that new prompts
 // are created with a unique ID.
 type PromptDB struct {
@@ -228,12 +396,6 @@ type PromptDB struct {
 	notifyPrompt func(userID uint32, promptID prompting.IDType, data map[string]string) error
 }
 
-const (
-	// maxOutstandingPromptsPerUser is an arbitrary limit.
-	// TODO: review this limit after some usage.
-	maxOutstandingPromptsPerUser int = 1000
-)
-
 // New creates and returns a new prompt database.
 //
 // The given notifyPrompt closure will be called when a prompt is added,
@@ -257,6 +419,10 @@ func New(notifyPrompt func(userID uint32, promptID prompting.IDType, data map[st
 	return &pdb, nil
 }
 
+var timeAfterFunc = func(d time.Duration, f func()) timeutil.Timer {
+	return timeutil.AfterFunc(d, f)
+}
+
 // AddOrMerge checks if the given prompt contents are identical to an existing
 // prompt and, if so, merges with it by adding the given listenerReq to it.
 // Otherwise, adds a new prompt with the given contents to the prompt DB.
@@ -269,7 +435,7 @@ func New(notifyPrompt func(userID uint32, promptID prompting.IDType, data map[st
 //
 // The caller must ensure that the given permissions are in the order in which
 // they appear in the available permissions list for the given interface.
-func (pdb *PromptDB) AddOrMerge(metadata *prompting.Metadata, path string, requestedPermissions []string, remainingPermissions []string, listenerReq *listener.Request) (*Prompt, bool, error) {
+func (pdb *PromptDB) AddOrMerge(metadata *prompting.Metadata, path string, requestedPermissions []string, outstandingPermissions []string, listenerReq *listener.Request) (*Prompt, bool, error) {
 	availablePermissions, err := prompting.AvailablePermissions(metadata.Interface)
 	if err != nil {
 		// Error should be impossible, since caller has already validated that
@@ -281,23 +447,27 @@ func (pdb *PromptDB) AddOrMerge(metadata *prompting.Metadata, path string, reque
 	pdb.mutex.Lock()
 	defer pdb.mutex.Unlock()
 
-	if pdb.maxIDMmap.IsClosed() {
-		return nil, false, ErrClosed
+	if pdb.isClosed() {
+		return nil, false, prompting_errors.ErrPromptsClosed
 	}
 
 	userEntry, ok := pdb.perUser[metadata.User]
 	if !ok {
-		pdb.perUser[metadata.User] = &userPromptDB{
+		// New user entry, so create it and set up the expiration timer
+		userEntry = &userPromptDB{
 			ids: make(map[prompting.IDType]int),
 		}
-		userEntry = pdb.perUser[metadata.User]
+		userEntry.expirationTimer = timeAfterFunc(initialTimeout, func() {
+			userEntry.timeoutCallback(pdb, metadata.User)
+		})
+		pdb.perUser[metadata.User] = userEntry
 	}
 
 	constraints := &promptConstraints{
-		path:                 path,
-		remainingPermissions: remainingPermissions,
-		availablePermissions: availablePermissions,
-		originalPermissions:  requestedPermissions,
+		path:                   path,
+		outstandingPermissions: outstandingPermissions,
+		availablePermissions:   availablePermissions,
+		originalPermissions:    requestedPermissions,
 	}
 
 	// Search for an identical existing prompt, merge if found
@@ -316,9 +486,10 @@ func (pdb *PromptDB) AddOrMerge(metadata *prompting.Metadata, path string, reque
 
 	if len(userEntry.prompts) >= maxOutstandingPromptsPerUser {
 		logger.Noticef("WARNING: too many outstanding prompts for user %d; auto-denying new one", metadata.User)
-		allowedPermission := responseForInterfaceConstraintsOutcome(metadata.Interface, constraints, prompting.OutcomeDeny)
+		// Deny all permissions which are not already allowed by existing rules
+		allowedPermission := constraints.buildResponse(metadata.Interface, constraints.outstandingPermissions)
 		sendReply(listenerReq, allowedPermission)
-		return nil, false, ErrTooManyPrompts
+		return nil, false, prompting_errors.ErrTooManyPrompts
 	}
 
 	id, _ := pdb.maxIDMmap.NextID() // err must be nil because maxIDMmap is not nil and lock is held
@@ -336,70 +507,57 @@ func (pdb *PromptDB) AddOrMerge(metadata *prompting.Metadata, path string, reque
 	return prompt, false, nil
 }
 
-func responseForInterfaceConstraintsOutcome(iface string, constraints *promptConstraints, outcome prompting.OutcomeType) any {
-	allow, err := outcome.AsBool()
-	if err != nil {
-		// This should not occur, but if so, default to deny
-		allow = false
-		logger.Debugf("%v", err)
-	}
-	allowedPerms := constraints.originalPermissions
-	if !allow {
-		// Remaining permissions were denied, so allow permissions which were
-		// previously allowed by prompt rules
-		allowedPerms = make([]string, 0, len(constraints.originalPermissions)-len(constraints.remainingPermissions))
-		for _, perm := range constraints.originalPermissions {
-			// Exclude any permissions which were remaining at time of denial
-			if !strutil.ListContains(constraints.remainingPermissions, perm) {
-				allowedPerms = append(allowedPerms, perm)
-			}
-		}
-	}
-	allowedPermission, err := prompting.AbstractPermissionsToAppArmorPermissions(iface, allowedPerms)
-	if err != nil {
-		// This should not occur, but if so, permission should be set to the
-		// empty value for its corresponding permission type.
-		logger.Debugf("internal error: cannot convert abstract permissions to AppArmor permissions: %v", err)
-	}
-	return allowedPermission
-}
-
 // Prompts returns a slice of all outstanding prompts for the given user.
-func (pdb *PromptDB) Prompts(user uint32) ([]*Prompt, error) {
+//
+// If clientActivity is true, reset the expiration timeout for prompts for
+// the given user.
+func (pdb *PromptDB) Prompts(user uint32, clientActivity bool) ([]*Prompt, error) {
 	pdb.mutex.RLock()
 	defer pdb.mutex.RUnlock()
-	if pdb.maxIDMmap.IsClosed() {
-		return nil, ErrClosed
+	if pdb.isClosed() {
+		return nil, prompting_errors.ErrPromptsClosed
 	}
 	userEntry, ok := pdb.perUser[user]
 	if !ok || len(userEntry.prompts) == 0 {
 		// No prompts for user, but no error
 		return nil, nil
 	}
+	if clientActivity {
+		userEntry.activityResetExpiration()
+	}
 	promptsCopy := make([]*Prompt, len(userEntry.prompts))
 	copy(promptsCopy, userEntry.prompts)
 	return promptsCopy, nil
 }
 
 // PromptWithID returns the prompt with the given ID for the given user.
-func (pdb *PromptDB) PromptWithID(user uint32, id prompting.IDType) (*Prompt, error) {
+//
+// If clientActivity is true, reset the expiration timeout for prompts for
+// the given user.
+func (pdb *PromptDB) PromptWithID(user uint32, id prompting.IDType, clientActivity bool) (*Prompt, error) {
 	pdb.mutex.RLock()
 	defer pdb.mutex.RUnlock()
-	_, prompt, err := pdb.promptWithID(user, id)
+	_, prompt, err := pdb.promptWithID(user, id, clientActivity)
 	return prompt, err
 }
 
 // promptWithID returns the user entry for the given user and the prompt with
 // the given ID for the that user.
 //
-// The caller should hold a read lock on the prompt DB mutex.
-func (pdb *PromptDB) promptWithID(user uint32, id prompting.IDType) (*userPromptDB, *Prompt, error) {
-	if pdb.maxIDMmap.IsClosed() {
-		return nil, nil, ErrClosed
+// If clientActivity is true, reset the expiration timeout for prompts for
+// the given user.
+//
+// The caller should hold a read (or write) lock on the prompt DB mutex.
+func (pdb *PromptDB) promptWithID(user uint32, id prompting.IDType, clientActivity bool) (*userPromptDB, *Prompt, error) {
+	if pdb.isClosed() {
+		return nil, nil, prompting_errors.ErrPromptsClosed
 	}
 	userEntry, ok := pdb.perUser[user]
 	if !ok {
-		return nil, nil, ErrNotFound
+		return nil, nil, prompting_errors.ErrPromptNotFound
+	}
+	if clientActivity {
+		userEntry.activityResetExpiration()
 	}
 	prompt, err := userEntry.get(id)
 	if err != nil {
@@ -413,22 +571,18 @@ func (pdb *PromptDB) promptWithID(user uint32, id prompting.IDType) (*userPrompt
 // prompt from the prompt DB.
 //
 // Records a notice for the prompt, and returns the prompt's former contents.
-func (pdb *PromptDB) Reply(user uint32, id prompting.IDType, outcome prompting.OutcomeType) (*Prompt, error) {
+//
+// If clientActivity is true, reset the expiration timeout for prompts for
+// the given user.
+func (pdb *PromptDB) Reply(user uint32, id prompting.IDType, outcome prompting.OutcomeType, clientActivity bool) (*Prompt, error) {
 	pdb.mutex.Lock()
 	defer pdb.mutex.Unlock()
-	userEntry, prompt, err := pdb.promptWithID(user, id)
+	userEntry, prompt, err := pdb.promptWithID(user, id, clientActivity)
 	if err != nil {
 		return nil, err
 	}
-	allowedPermission := responseForInterfaceConstraintsOutcome(prompt.Interface, prompt.Constraints, outcome)
-	for _, listenerReq := range prompt.listenerReqs {
-		if err := sendReply(listenerReq, allowedPermission); err != nil {
-			// Error should only occur if reply is malformed, and since these
-			// listener requests should be identical, if a reply is malformed
-			// for one, it should be malformed for all. Malformed replies should
-			// leave the listener request unchanged. Thus, return early.
-			return nil, err
-		}
+	if err = prompt.sendReply(outcome); err != nil {
+		return nil, err
 	}
 	userEntry.remove(id)
 	data := map[string]string{"resolved": "replied"}
@@ -436,37 +590,32 @@ func (pdb *PromptDB) Reply(user uint32, id prompting.IDType, outcome prompting.O
 	return prompt, nil
 }
 
-var sendReply = func(listenerReq *listener.Request, allowedPermission any) error {
-	return listenerReq.Reply(allowedPermission)
-}
-
 // HandleNewRule checks if any existing prompts are satisfied by the given rule
 // contents and, if so, sends back a decision to their listener requests.
 //
 // A prompt is satisfied by the given rule contents if the user, snap,
-// interface, and path of the prompt match those of the rule, and if either the
-// outcome is "allow" and all of the prompt's permissions are matched by those
-// of the rule contents, or if the outcome is "deny" and any of the permissions
-// match.
+// interface, and path of the prompt match those of the rule, and all
+// outstanding permissions are covered by permissions in the rule constraints
+// or at least one of the outstanding permissions is covered by a rule
+// permission which has an outcome of "deny".
 //
 // Records a notice for any prompt which was satisfied, or which had some of
 // its permissions satisfied by the rule contents. In the future, only the
-// remaining unsatisfied permissions of a partially-satisfied prompt must be
+// outstanding unsatisfied permissions of a partially-satisfied prompt must be
 // satisfied for the prompt as a whole to be satisfied.
 //
 // Returns the IDs of any prompts which were fully satisfied by the given rule
 // contents.
-func (pdb *PromptDB) HandleNewRule(metadata *prompting.Metadata, constraints *prompting.Constraints, outcome prompting.OutcomeType) ([]prompting.IDType, error) {
-	// Validate outcome before locking
-	allow, err := outcome.AsBool()
-	if err != nil {
-		return nil, err
-	}
+//
+// Since rule is new, we don't check the expiration timestamps for any
+// permissions, since any permissions with lifespan timespan were validated to
+// have a non-zero duration, and we handle this rule as it was at its creation.
+func (pdb *PromptDB) HandleNewRule(metadata *prompting.Metadata, constraints *prompting.RuleConstraints) ([]prompting.IDType, error) {
 	pdb.mutex.Lock()
 	defer pdb.mutex.Unlock()
 
-	if pdb.maxIDMmap.IsClosed() {
-		return nil, ErrClosed
+	if pdb.isClosed() {
+		return nil, prompting_errors.ErrPromptsClosed
 	}
 
 	userEntry, ok := pdb.perUser[metadata.User]
@@ -478,39 +627,51 @@ func (pdb *PromptDB) HandleNewRule(metadata *prompting.Metadata, constraints *pr
 		if !(prompt.Snap == metadata.Snap && prompt.Interface == metadata.Interface) {
 			continue
 		}
-		matched, err := constraints.Match(prompt.Constraints.path)
+
+		affectedByRule, respond, deniedPermissions, err := prompt.Constraints.applyRuleConstraints(constraints)
 		if err != nil {
+			// Should not occur, only error is if path pattern is malformed,
+			// which would have thrown an error while parsing, not now.
 			return nil, err
 		}
-		if !matched {
+		if !affectedByRule {
+			continue
+		}
+		if !respond {
+			// No response necessary, though the prompt constraints were
+			// modified, so just record a notice for the prompt.
+			pdb.notifyPrompt(metadata.User, prompt.ID, nil)
 			continue
 		}
 
-		// Record all allowed permissions at the time of match, in case a
-		// permission is denied and we need to send back a response.
-		allowedPermission := responseForInterfaceConstraintsOutcome(metadata.Interface, prompt.Constraints, outcome)
-
-		// See if the permission matches any of the prompt's remaining permissions
-		modified := prompt.Constraints.subtractPermissions(constraints.Permissions)
-		if !modified {
-			// No permission was matched
-			continue
+		// A response is necessary, so either all permissions were allowed or
+		// at least one permission was denied. Construct a response and send it
+		// back to the kernel, and record a notice that the prompt was satisfied.
+		if len(deniedPermissions) > 0 {
+			// At least one permission was denied by new rule, and we want to
+			// send a response immediately, so include any outstanding
+			// permissions as denied as well.
+			//
+			// This could be done as part of applyRuleConstraints instead, but
+			// it seems semantically clearer to only return the permissions
+			// which were explicitly denied by the rule, rather than all
+			// outstanding permissions because at least one was denied. It's
+			// the prorogative of the caller (this function) to treat the
+			// outstanding permissions as denied since we want to send a
+			// response without waiting for future rules to satisfy the
+			// outstanding permissions.
+			deniedPermissions = append(deniedPermissions, prompt.Constraints.outstandingPermissions...)
 		}
-		id := prompt.ID
-		if len(prompt.Constraints.remainingPermissions) > 0 && allow == true {
-			pdb.notifyPrompt(metadata.User, id, nil)
-			continue
-		}
-		// All permissions of prompt satisfied, or any permission denied
-		for _, listenerReq := range prompt.listenerReqs {
-			// Reply with any permissions which were allowed, either by this
-			// new rule or by previous rules.
-			sendReply(listenerReq, allowedPermission)
-		}
-		userEntry.remove(id)
-		satisfiedPromptIDs = append(satisfiedPromptIDs, id)
+		// Build and send a response with any permissions which were allowed,
+		// either by this new rule or by previous rules.
+		allowedPermission := prompt.Constraints.buildResponse(metadata.Interface, deniedPermissions)
+		prompt.sendReplyWithPermission(allowedPermission)
+		// Now that a response has been sent, remove the rule from the rule DB
+		// and record a notice indicating that it has been satisfied.
+		userEntry.remove(prompt.ID)
+		satisfiedPromptIDs = append(satisfiedPromptIDs, prompt.ID)
 		data := map[string]string{"resolved": "satisfied"}
-		pdb.notifyPrompt(metadata.User, id, data)
+		pdb.notifyPrompt(metadata.User, prompt.ID, data)
 	}
 	return satisfiedPromptIDs, nil
 }
@@ -523,8 +684,8 @@ func (pdb *PromptDB) Close() error {
 	pdb.mutex.Lock()
 	defer pdb.mutex.Unlock()
 
-	if pdb.maxIDMmap.IsClosed() {
-		return ErrClosed
+	if pdb.isClosed() {
+		return prompting_errors.ErrPromptsClosed
 	}
 
 	if err := pdb.maxIDMmap.Close(); err != nil {
@@ -535,9 +696,19 @@ func (pdb *PromptDB) Close() error {
 	// not want to send {"resolved": "cancelled"} in the notice data.
 	data := map[string]string{"resolved": "cancelled"}
 	for user, userEntry := range pdb.perUser {
+		// No need to explicitly stop the timer, since this is racy with the
+		// timer expiring, and since we've already closed the maxIDMmap with
+		// the lock held, the callback will acquire the lock, see that the DB
+		// has been closed, and immediately return.
 		for id := range userEntry.ids {
 			pdb.notifyPrompt(user, id, data)
 		}
+		// There's no need to explicitly send responses back to the kernel
+		// here, as the listener will send deny responses for all outstanding
+		// listener requests when it is closed, and even if it didn't, the
+		// kernel will automatically reclaim unanswered notifications when
+		// the notify socket FD is closed, and either roll them over to a new
+		// FD or auto-deny them after some time.
 	}
 
 	// Clear all outstanding prompts
@@ -545,6 +716,13 @@ func (pdb *PromptDB) Close() error {
 	return nil
 }
 
+// isClosed returns true if the prompt DB is already closed.
+//
+// The caller must ensure that the DB lock is held.
+func (pdb *PromptDB) isClosed() bool {
+	return pdb.maxIDMmap.IsClosed()
+}
+
 // MockSendReply mocks the function to send a reply back to the listener so
 // tests, both for this package and for consumers of this package, can mock
 // the listener.
