@@ -26,9 +26,10 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
+	"os"
 	"os/exec"
 	"path/filepath"
+	"regexp"
 	"sort"
 	"strconv"
 	"strings"
@@ -55,17 +56,17 @@ func MockUdevPropertiesForDevice(new func(string, string) (map[string]string, er
 	old := udevadmProperties
 	// for better testing we mock the udevadm command output so that we still
 	// test the parsing
-	udevadmProperties = func(typeOpt, dev string) ([]byte, error) {
+	udevadmProperties = func(typeOpt, dev string) ([]byte, []byte, error) {
 		props, err := new(typeOpt, dev)
 		if err != nil {
-			return []byte(err.Error()), err
+			return []byte(err.Error()), []byte{}, err
 		}
 		// put it into udevadm format output, i.e. "KEY=VALUE\n"
 		output := ""
 		for k, v := range props {
 			output += fmt.Sprintf("%s=%s\n", k, v)
 		}
-		return []byte(output), nil
+		return []byte(output), []byte{}, nil
 	}
 	return func() {
 		udevadmProperties = old
@@ -89,17 +90,16 @@ func parseDeviceMajorMinor(s string) (int, int, error) {
 	return maj, min, nil
 }
 
-var udevadmProperties = func(typeOpt, device string) ([]byte, error) {
+var udevadmProperties = func(typeOpt, device string) ([]byte, []byte, error) {
 	// TODO: maybe combine with gadget interfaces hotplug code where the udev
 	// db is parsed?
-	cmd := exec.Command("udevadm", "info", "--query", "property", typeOpt, device)
-	return cmd.CombinedOutput()
+	return osutil.RunSplitOutput("udevadm", "info", "--query", "property", typeOpt, device)
 }
 
 func udevProperties(typeOpt, device string) (map[string]string, error) {
-	out, err := udevadmProperties(typeOpt, device)
+	out, stderr, err := udevadmProperties(typeOpt, device)
 	if err != nil {
-		return nil, osutil.OutputErr(out, err)
+		return nil, osutil.OutputErrCombine(out, stderr, err)
 	}
 	r := bytes.NewBuffer(out)
 
@@ -496,13 +496,13 @@ func parentPartitionPropsForOptions(props map[string]string) (map[string]string,
 	majmin := props["MAJOR"] + ":" + props["MINOR"]
 
 	dmDir := filepath.Join(dirs.SysfsDir, "dev", "block", majmin, "dm")
-	dmUUID, err := ioutil.ReadFile(filepath.Join(dmDir, "uuid"))
+	dmUUID, err := os.ReadFile(filepath.Join(dmDir, "uuid"))
 	if err != nil {
 		return nil, fmt.Errorf(errFmt, err)
 	}
 	dmUUID = bytes.TrimSpace(dmUUID)
 
-	dmName, err := ioutil.ReadFile(filepath.Join(dmDir, "name"))
+	dmName, err := os.ReadFile(filepath.Join(dmDir, "name"))
 	if err != nil {
 		return nil, fmt.Errorf(errFmt, err)
 	}
@@ -610,7 +610,7 @@ func partitionPropsFromMountPoint(mountpoint string) (source string, props map[s
 		}
 	}
 	if source == "" {
-		return "", nil, fmt.Errorf("cannot find mountpoint %q", mountpoint)
+		return "", nil, errMountPointNotFoundImpl{path: mountpoint}
 	}
 
 	// now we have the partition for this mountpoint, we need to tie that back
@@ -688,7 +688,7 @@ func (d *disk) populatePartitions() error {
 			// /dev/mmcblk0boot0 disk device on the dragonboard which exists
 			// under the /dev/mmcblk0 disk, but is not a partition and is
 			// instead a proper disk
-			_, err := ioutil.ReadFile(filepath.Join(path, "partition"))
+			_, err := os.ReadFile(filepath.Join(path, "partition"))
 			if err != nil {
 				continue
 			}
@@ -935,7 +935,7 @@ func AllPhysicalDisks() ([]Disk, error) {
 	// get disks for every block device in /sys/block/
 	blockDir := filepath.Join(dirs.SysfsDir, "block")
 
-	files, err := ioutil.ReadDir(blockDir)
+	files, err := os.ReadDir(blockDir)
 	if err != nil {
 		return nil, err
 	}
@@ -964,41 +964,60 @@ func AllPhysicalDisks() ([]Disk, error) {
 	return disks, nil
 }
 
-// PartitionUUIDFromMountPoint returns the UUID of the partition which is a
-// source of a given mount point.
-func PartitionUUIDFromMountPoint(mountpoint string, opts *Options) (string, error) {
+var dmUUIDRe = regexp.MustCompile(`^CRYPT-(?P<type>.*)-(?P<uuid1>[0-9a-f]{8})(?P<uuid2>[0-9a-f]{4})(?P<uuid3>[0-9a-f]{4})(?P<uuid4>[0-9a-f]{4})(?P<uuid5>[0-9a-f]{12})-(?P<name>.*)$`)
+
+// DMCryptUUIDFromMountPoint finds the UUID of a device mapper device
+// mounted at mountpoint.
+func DMCryptUUIDFromMountPoint(mountpoint string) (string, error) {
 	_, props, err := partitionPropsFromMountPoint(mountpoint)
 	if err != nil {
 		return "", err
 	}
 
-	if opts != nil && opts.IsDecryptedDevice {
-		props, err = parentPartitionPropsForOptions(props)
-		if err != nil {
-			return "", err
+	dmUUID, hasDmUUID := props["DM_UUID"]
+	if !hasDmUUID {
+		// There is a bug in old UC20 kernels where udev does
+		// not retrieve data from cold plugged disks
+		// initialized in initrd.
+		// https://github.com/canonical/core-initrd/pull/58
+		// https://github.com/canonical/core-initrd/pull/203
+		devPath, hasDevPath := props["DEVPATH"]
+		if !hasDevPath {
+			return "", ErrNoDmUUID
 		}
+		devUUIDPath := filepath.Join(dirs.SysfsDir, devPath, "dm", "uuid")
+		data, err := os.ReadFile(devUUIDPath)
+		if err != nil {
+			return "", ErrNoDmUUID
+		}
+		dmUUID = strings.TrimSuffix(string(data), "\n")
+
 	}
-	partUUID := props["ID_PART_ENTRY_UUID"]
-	if partUUID == "" {
-		partDev := filepath.Join("/dev", props["DEVNAME"])
-		return "", fmt.Errorf("cannot get required partition UUID udev property for device %s", partDev)
+
+	match := dmUUIDRe.FindStringSubmatchIndex(dmUUID)
+	if match == nil {
+		return "", fmt.Errorf("value of DM_UUID is not recognized")
 	}
-	return partUUID, nil
+
+	result := []byte{}
+	result = dmUUIDRe.ExpandString(result, "${uuid1}-${uuid2}-${uuid3}-${uuid4}-${uuid5}", dmUUID, match)
+	return string(result), nil
 }
 
-// PartitionUUID returns the UUID of a given partition
-func PartitionUUID(node string) (string, error) {
+// FilesystemUUID retrieves the UUID of the filesystem for a give node
+// path
+func FilesystemUUID(node string) (string, error) {
 	props, err := udevPropertiesForName(node)
 	if err != nil && props == nil {
 		// only fail here if props is nil, if it's available we validate it
 		// below
 		return "", fmt.Errorf("cannot process udev properties: %v", err)
 	}
-	partUUID := props["ID_PART_ENTRY_UUID"]
-	if partUUID == "" {
-		return "", fmt.Errorf("cannot get required udev partition UUID property")
+	uuid := props["ID_FS_UUID"]
+	if uuid == "" {
+		return "", fmt.Errorf("cannot get required udev ID_FS_UUID property")
 	}
-	return partUUID, nil
+	return uuid, nil
 }
 
 func SectorSize(devname string) (uint64, error) {
@@ -1016,3 +1035,16 @@ func filesystemTypeForPartition(devname string) (string, error) {
 
 	return props["ID_FS_TYPE"], nil
 }
+
+// Devlinks returns all the /dev symlinks for a node
+func Devlinks(node string) ([]string, error) {
+	props, err := udevPropertiesForName(node)
+	if err != nil && props == nil {
+		return []string{}, fmt.Errorf("cannot process udev properties: %v", err)
+	}
+	devlinks := props["DEVLINKS"]
+	if devlinks == "" {
+		return []string{}, fmt.Errorf("cannot get required udev DEVLINKS property")
+	}
+	return strings.Split(devlinks, " "), nil
+}
