@@ -22,7 +22,6 @@ package disks_test
 import (
 	"errors"
 	"fmt"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 
@@ -122,7 +121,7 @@ func createVirtioDevicesInSysfs(c *C, path string, devsToPartition map[string]bo
 		err := os.MkdirAll(filepath.Join(diskDir, dev), 0755)
 		c.Assert(err, IsNil)
 		if isPartition {
-			err = ioutil.WriteFile(filepath.Join(diskDir, dev, "partition"), []byte("1"), 0644)
+			err = os.WriteFile(filepath.Join(diskDir, dev, "partition"), []byte("1"), 0644)
 			c.Assert(err, IsNil)
 		}
 	}
@@ -744,11 +743,11 @@ func (s *diskSuite) TestDiskFromMountPointIsDecryptedLUKSDeviceVolumeHappy(c *C)
 	c.Assert(err, IsNil)
 
 	b := []byte("something")
-	err = ioutil.WriteFile(filepath.Join(dmDir, "name"), b, 0644)
+	err = os.WriteFile(filepath.Join(dmDir, "name"), b, 0644)
 	c.Assert(err, IsNil)
 
 	b = []byte("CRYPT-LUKS2-5a522809c87e4dfa81a88dc5667d1304-something")
-	err = ioutil.WriteFile(filepath.Join(dmDir, "uuid"), b, 0644)
+	err = os.WriteFile(filepath.Join(dmDir, "uuid"), b, 0644)
 	c.Assert(err, IsNil)
 
 	opts := &disks.Options{IsDecryptedDevice: true}
@@ -1201,11 +1200,11 @@ func (s *diskSuite) TestDiskFromMountPointDecryptedDevicePartitionsHappy(c *C) {
 	c.Assert(err, IsNil)
 
 	b := []byte("ubuntu-data-3776bab4-8bcc-46b7-9da2-6a84ce7f93b4")
-	err = ioutil.WriteFile(filepath.Join(dmDir, "name"), b, 0644)
+	err = os.WriteFile(filepath.Join(dmDir, "name"), b, 0644)
 	c.Assert(err, IsNil)
 
 	b = []byte("CRYPT-LUKS2-5a522809c87e4dfa81a88dc5667d1304-ubuntu-data-3776bab4-8bcc-46b7-9da2-6a84ce7f93b4")
-	err = ioutil.WriteFile(filepath.Join(dmDir, "uuid"), b, 0644)
+	err = os.WriteFile(filepath.Join(dmDir, "uuid"), b, 0644)
 	c.Assert(err, IsNil)
 
 	// mock the dev nodes in sysfs for the partitions
@@ -1781,7 +1780,7 @@ func (s *diskSuite) TestAllPhysicalDisks(c *C) {
 	c.Assert(err, IsNil)
 	devsToCreate := []string{"sda", "loop1", "loop2", "sdb", "nvme0n1", "mmcblk0"}
 	for _, dev := range devsToCreate {
-		err := ioutil.WriteFile(filepath.Join(blockDir, dev), nil, 0644)
+		err := os.WriteFile(filepath.Join(blockDir, dev), nil, 0644)
 		c.Assert(err, IsNil)
 	}
 
@@ -1850,12 +1849,13 @@ func (s *diskSuite) TestAllPhysicalDisks(c *C) {
 	c.Assert(d[3].KernelDeviceNode(), Equals, "/dev/sdb")
 }
 
-func (s *diskSuite) TestPartitionUUIDFromMopuntPointErrs(c *C) {
+func (s *diskSuite) TestDMCryptUUIDFromMountPointErrs(c *C) {
 	restore := osutil.MockMountInfo(``)
 	defer restore()
 
-	_, err := disks.PartitionUUIDFromMountPoint("/run/mnt/blah", nil)
-	c.Assert(err, ErrorMatches, "cannot find mountpoint \"/run/mnt/blah\"")
+	_, err := disks.DMCryptUUIDFromMountPoint("/run/mnt/blah")
+	c.Assert(err, ErrorMatches, `cannot find mountpoint "/run/mnt/blah"`)
+	c.Assert(errors.Is(err, disks.ErrMountPointNotFound), Equals, true)
 
 	restore = osutil.MockMountInfo(`130 30 42:1 / /run/mnt/point rw,relatime shared:54 - ext4 /dev/vda4 rw
 `)
@@ -1871,30 +1871,26 @@ func (s *diskSuite) TestPartitionUUIDFromMopuntPointErrs(c *C) {
 	})
 	defer restore()
 
-	_, err = disks.PartitionUUIDFromMountPoint("/run/mnt/point", nil)
-	c.Assert(err, ErrorMatches, "cannot get required partition UUID udev property for device /dev/vda4")
-}
+	_, err = disks.DMCryptUUIDFromMountPoint("/run/mnt/point")
+	c.Assert(err, ErrorMatches, "device has no DM_UUID")
+	c.Assert(err, Equals, disks.ErrNoDmUUID)
 
-func (s *diskSuite) TestPartitionUUIDFromMountPointPlain(c *C) {
-	restore := osutil.MockMountInfo(`130 30 42:1 / /run/mnt/point rw,relatime shared:54 - ext4 /dev/vda4 rw
-`)
-	defer restore()
 	restore = disks.MockUdevPropertiesForDevice(func(typeOpt, dev string) (map[string]string, error) {
 		c.Assert(typeOpt, Equals, "--name")
 		c.Assert(dev, Equals, "/dev/vda4")
 		return map[string]string{
-			"DEVTYPE":            "disk",
-			"ID_PART_ENTRY_UUID": "foo-uuid",
+			"DEVNAME": "vda4",
+			"prop":    "hello",
+			"DM_UUID": "garbage",
 		}, nil
 	})
 	defer restore()
 
-	uuid, err := disks.PartitionUUIDFromMountPoint("/run/mnt/point", nil)
-	c.Assert(err, IsNil)
-	c.Assert(uuid, Equals, "foo-uuid")
+	_, err = disks.DMCryptUUIDFromMountPoint("/run/mnt/point")
+	c.Assert(err, ErrorMatches, "value of DM_UUID is not recognized")
 }
 
-func (s *diskSuite) TestPartitionUUIDFromMopuntPointDecrypted(c *C) {
+func (s *diskSuite) TestDMCryptUUIDFromMountPoint(c *C) {
 	restore := osutil.MockMountInfo(`130 30 42:1 / /run/mnt/point rw,relatime shared:54 - ext4 /dev/mapper/something rw
 `)
 	defer restore()
@@ -1903,13 +1899,7 @@ func (s *diskSuite) TestPartitionUUIDFromMopuntPointDecrypted(c *C) {
 		switch dev {
 		case "/dev/mapper/something":
 			return map[string]string{
-				"DEVTYPE": "disk",
-				"MAJOR":   "242",
-				"MINOR":   "1",
-			}, nil
-		case "/dev/disk/by-uuid/5a522809-c87e-4dfa-81a8-8dc5667d1304":
-			return map[string]string{
-				"ID_PART_ENTRY_UUID": "foo-uuid",
+				"DM_UUID": "CRYPT-LUKS2-5a522809c87e4dfa81a88dc5667d1304-something",
 			}, nil
 		default:
 			c.Errorf("unexpected udev device properties requested: %s", dev)
@@ -1918,40 +1908,22 @@ func (s *diskSuite) TestPartitionUUIDFromMopuntPointDecrypted(c *C) {
 	})
 	defer restore()
 
-	// mock the sysfs dm uuid and name files
-	dmDir := filepath.Join(filepath.Join(dirs.SysfsDir, "dev", "block"), "242:1", "dm")
-	err := os.MkdirAll(dmDir, 0755)
+	uuid, err := disks.DMCryptUUIDFromMountPoint("/run/mnt/point")
 	c.Assert(err, IsNil)
-
-	b := []byte("something")
-	err = ioutil.WriteFile(filepath.Join(dmDir, "name"), b, 0644)
-	c.Assert(err, IsNil)
-
-	b = []byte("CRYPT-LUKS2-5a522809c87e4dfa81a88dc5667d1304-something")
-	err = ioutil.WriteFile(filepath.Join(dmDir, "uuid"), b, 0644)
-	c.Assert(err, IsNil)
-
-	uuid, err := disks.PartitionUUIDFromMountPoint("/run/mnt/point", &disks.Options{
-		IsDecryptedDevice: true,
-	})
-	c.Assert(err, IsNil)
-	c.Assert(uuid, Equals, "foo-uuid")
+	c.Assert(uuid, Equals, "5a522809-c87e-4dfa-81a8-8dc5667d1304")
 }
 
-func (s *diskSuite) TestPartitionUUID(c *C) {
-	restore := disks.MockUdevPropertiesForDevice(func(typeOpt, dev string) (map[string]string, error) {
+func (s *diskSuite) TestDMCryptUUIDFromMountPointFallback(c *C) {
+	restore := osutil.MockMountInfo(`130 30 42:1 / /run/mnt/point rw,relatime shared:54 - ext4 /dev/mapper/something rw
+`)
+	defer restore()
+	restore = disks.MockUdevPropertiesForDevice(func(typeOpt, dev string) (map[string]string, error) {
 		c.Assert(typeOpt, Equals, "--name")
 		switch dev {
-		case "/dev/vda4":
+		case "/dev/mapper/something":
 			return map[string]string{
-				"ID_PART_ENTRY_UUID": "foo-uuid",
+				"DEVPATH": "/devices/virtual/mydevice",
 			}, nil
-		case "/dev/no-uuid":
-			return map[string]string{
-				"no-uuid": "no-uuid",
-			}, nil
-		case "/dev/mock-failure":
-			return nil, fmt.Errorf("mock failure")
 		default:
 			c.Errorf("unexpected udev device properties requested: %s", dev)
 			return nil, fmt.Errorf("unexpected udev device: %s", dev)
@@ -1959,17 +1931,15 @@ func (s *diskSuite) TestPartitionUUID(c *C) {
 	})
 	defer restore()
 
-	uuid, err := disks.PartitionUUID("/dev/vda4")
+	err := os.MkdirAll(filepath.Join(dirs.SysfsDir, "devices/virtual/mydevice", "dm"), 0755)
+	c.Assert(err, IsNil)
+	b := []byte("CRYPT-LUKS2-5a522809c87e4dfa81a88dc5667d1304-something\n")
+	err = os.WriteFile(filepath.Join(dirs.SysfsDir, "devices/virtual/mydevice", "dm/uuid"), b, 0644)
 	c.Assert(err, IsNil)
-	c.Assert(uuid, Equals, "foo-uuid")
 
-	uuid, err = disks.PartitionUUID("/dev/no-uuid")
-	c.Assert(err, ErrorMatches, "cannot get required udev partition UUID property")
-	c.Check(uuid, Equals, "")
-
-	uuid, err = disks.PartitionUUID("/dev/mock-failure")
-	c.Assert(err, ErrorMatches, "cannot process udev properties: mock failure")
-	c.Check(uuid, Equals, "")
+	uuid, err := disks.DMCryptUUIDFromMountPoint("/run/mnt/point")
+	c.Assert(err, IsNil)
+	c.Assert(uuid, Equals, "5a522809-c87e-4dfa-81a8-8dc5667d1304")
 }
 
 func (s *diskSuite) TestFilesystemTypeForPartition(c *C) {
@@ -2035,3 +2005,77 @@ func (s *diskSuite) TestFindMatchingPartitionWithFsLabel(c *C) {
 		}
 	}
 }
+
+func (s *diskSuite) TestMockDisksChecking(c *C) {
+	f := func() {
+		disks.MockDeviceNameToDiskMapping(map[string]*disks.MockDiskMapping{
+			"/dev/vda": {
+				Structure: []disks.Partition{
+					{KernelDeviceNode: "/dev/vda1"},
+					{KernelDeviceNode: "/dev/vda1"},
+				},
+			},
+		})
+	}
+	c.Check(f, Panics, "mock error: duplicated kernel device nodes for partitions in disk mapping")
+}
+
+func (s *diskSuite) TestFilesystemUUID(c *C) {
+	restore := disks.MockUdevPropertiesForDevice(func(typeOpt, dev string) (map[string]string, error) {
+		c.Assert(typeOpt, Equals, "--name")
+		switch dev {
+		case "/dev/vda4":
+			return map[string]string{
+				"ID_FS_UUID": "the-expected-uuid",
+			}, nil
+		case "/dev/vda5":
+			return map[string]string{}, nil
+		case "/dev/vda6":
+			return nil, fmt.Errorf("some error")
+		default:
+			c.Errorf("unexpected udev device properties requested: %s", dev)
+			return nil, fmt.Errorf("unexpected udev device: %s", dev)
+		}
+	})
+	defer restore()
+
+	uuid, err := disks.FilesystemUUID("/dev/vda4")
+	c.Assert(err, IsNil)
+	c.Check(uuid, Equals, "the-expected-uuid")
+
+	_, err = disks.FilesystemUUID("/dev/vda5")
+	c.Check(err, ErrorMatches, `cannot get required udev ID_FS_UUID property`)
+
+	_, err = disks.FilesystemUUID("/dev/vda6")
+	c.Check(err, ErrorMatches, `cannot process udev properties: some error`)
+}
+
+func (s *diskSuite) TestDevlinks(c *C) {
+	restore := disks.MockUdevPropertiesForDevice(func(typeOpt, dev string) (map[string]string, error) {
+		c.Assert(typeOpt, Equals, "--name")
+		switch dev {
+		case "/dev/some/main/link":
+			return map[string]string{
+				"DEVLINKS": "/dev/some/other/link /dev/yet/another/link",
+			}, nil
+		case "/dev/no/links":
+			return map[string]string{}, nil
+		case "/dev/some/error":
+			return nil, fmt.Errorf("some error")
+		default:
+			c.Errorf("unexpected udev device properties requested: %s", dev)
+			return nil, fmt.Errorf("unexpected udev device: %s", dev)
+		}
+	})
+	defer restore()
+
+	links, err := disks.Devlinks("/dev/some/main/link")
+	c.Assert(err, IsNil)
+	c.Check(links, DeepEquals, []string{"/dev/some/other/link", "/dev/yet/another/link"})
+
+	_, err = disks.Devlinks("/dev/no/links")
+	c.Check(err, ErrorMatches, `cannot get required udev DEVLINKS property`)
+
+	_, err = disks.Devlinks("/dev/some/error")
+	c.Check(err, ErrorMatches, `cannot process udev properties: some error`)
+}
