@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2020 Canonical Ltd
+ * Copyright (C) 2020-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -21,12 +21,14 @@ package assertstate
 
 import (
 	"context"
+	"errors"
 	"fmt"
 	"sort"
 	"strings"
 
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/asserts/snapasserts"
+	"github.com/snapcore/snapd/confdb"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/release"
@@ -105,7 +107,7 @@ func bulkRefreshSnapDeclarations(s *state.State, snapStates map[string]*snapstat
 			PrimaryKey: []string{modelAs.Store()},
 		}
 		if err := pool.AddToUpdate(&storeRef, storeGroup); err != nil {
-			if !asserts.IsNotFound(err) {
+			if !errors.Is(err, &asserts.NotFoundError{}) {
 				return fmt.Errorf("cannot prepare store assertion refresh: %v", err)
 			}
 			// assertion is not present in the db yet,
@@ -126,7 +128,7 @@ func bulkRefreshSnapDeclarations(s *state.State, snapStates map[string]*snapstat
 	}
 
 	if mergedRPErr != nil {
-		if e := mergedRPErr.errors[storeGroup]; asserts.IsNotFound(e) || e == asserts.ErrUnresolved {
+		if e := mergedRPErr.errors[storeGroup]; errors.Is(e, &asserts.NotFoundError{}) || e == asserts.ErrUnresolved {
 			// ignore
 			delete(mergedRPErr.errors, storeGroup)
 		}
@@ -140,6 +142,26 @@ func bulkRefreshSnapDeclarations(s *state.State, snapStates map[string]*snapstat
 	return nil
 }
 
+func bulkRefreshConfdbs(s *state.State, confdbs []*confdb.Confdb, userID int, deviceCtx snapstate.DeviceContext, opts *RefreshAssertionsOptions) error {
+	db := cachedDB(s)
+
+	// all assertion refs will be in the same group
+	pool := asserts.NewPool(db, maxGroups)
+	for _, confdb := range confdbs {
+		account, name := confdb.Account, confdb.Name
+		ref := &asserts.Ref{
+			Type:       asserts.ConfdbType,
+			PrimaryKey: []string{account, name},
+		}
+
+		if err := pool.AddToUpdate(ref, storeGroup); err != nil {
+			return fmt.Errorf("cannot prepare confdb assertion %s/%s for refresh: %v", account, name, err)
+		}
+	}
+
+	return resolvePool(s, pool, nil, userID, deviceCtx, opts)
+}
+
 func bulkRefreshValidationSetAsserts(s *state.State, vsets map[string]*ValidationSetTracking, beforeCommitChecker func(*asserts.Database, asserts.Backstore) error, userID int, deviceCtx snapstate.DeviceContext, opts *RefreshAssertionsOptions) error {
 	db := cachedDB(s)
 	pool := asserts.NewPool(db, maxGroups)
@@ -191,7 +213,7 @@ func bulkRefreshValidationSetAsserts(s *state.State, vsets map[string]*Validatio
 		// ignore resolving errors for validation sets that are local only (no
 		// assertion in the store).
 		for group := range ignoreNotFound {
-			if e := rerr.errors[group]; asserts.IsNotFound(e) || e == asserts.ErrUnresolved {
+			if e := rerr.errors[group]; errors.Is(e, &asserts.NotFoundError{}) || e == asserts.ErrUnresolved {
 				delete(rerr.errors, group)
 			}
 		}
@@ -255,7 +277,7 @@ func resolvePool(s *state.State, pool *asserts.Pool, checkBeforeCommit func(*ass
 	unsupported := handleUnsupported(db)
 
 	for {
-		storeOpts := &store.RefreshOptions{IsAutoRefresh: opts.IsAutoRefresh}
+		storeOpts := &store.RefreshOptions{Scheduled: opts.IsAutoRefresh}
 		s.Unlock()
 		_, aresults, err := sto.SnapAction(context.TODO(), nil, nil, pool, user, storeOpts)
 		s.Lock()
