@@ -22,11 +22,12 @@ package config
 import (
 	"bytes"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"regexp"
+	"sort"
 	"strings"
 
-	"github.com/snapcore/snapd/features"
 	"github.com/snapcore/snapd/jsonutil"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
@@ -47,8 +48,42 @@ func ParseKey(key string) (subkeys []string, err error) {
 	return subkeys, nil
 }
 
-func PatchConfig(snapName string, subkeys []string, pos int, config interface{}, value *json.RawMessage) (interface{}, error) {
+func purgeNulls(config interface{}) interface{} {
+	switch config := config.(type) {
+	// map of json raw messages is the starting point for purgeNulls, this is the configuration we receive
+	case map[string]*json.RawMessage:
+		for k, v := range config {
+			if cfg := purgeNulls(v); cfg != nil {
+				config[k] = cfg.(*json.RawMessage)
+			} else {
+				delete(config, k)
+			}
+		}
+	case map[string]interface{}:
+		for k, v := range config {
+			if cfg := purgeNulls(v); cfg != nil {
+				config[k] = cfg
+			} else {
+				delete(config, k)
+			}
+		}
+	case *json.RawMessage:
+		if config == nil {
+			return nil
+		}
+		var configm interface{}
+		if err := jsonutil.DecodeWithNumber(bytes.NewReader(*config), &configm); err != nil {
+			panic(fmt.Errorf("internal error: cannot unmarshal configuration: %v", err))
+		}
+		if cfg := purgeNulls(configm); cfg != nil {
+			return jsonRaw(cfg)
+		}
+		return nil
+	}
+	return config
+}
 
+func PatchConfig(snapName string, subkeys []string, pos int, config interface{}, value *json.RawMessage) (interface{}, error) {
 	switch config := config.(type) {
 	case nil:
 		// Missing update map. Create and nest final value under it.
@@ -66,11 +101,25 @@ func PatchConfig(snapName string, subkeys []string, pos int, config interface{},
 		if err := jsonutil.DecodeWithNumber(bytes.NewReader(*config), &configm); err != nil {
 			return nil, fmt.Errorf("snap %q option %q is not a map", snapName, strings.Join(subkeys[:pos], "."))
 		}
-		_, err := PatchConfig(snapName, subkeys, pos, configm, value)
+		// preserve the invariant that if PatchConfig is
+		// passed a map[string]interface{} it is not nil.
+		// If the value was set to null in the same
+		// transaction use (interface{})(nil) which is handled
+		// by the first case here.
+		// (see LP: #1920773)
+		var cfg interface{}
+		if configm != nil {
+			cfg = configm
+		}
+		result, err := PatchConfig(snapName, subkeys, pos, cfg, value)
 		if err != nil {
 			return nil, err
 		}
-		return jsonRaw(configm), nil
+
+		// PatchConfig may have recreated higher level element that was previously set to null in same
+		// transaction; returning the result for PatchConfig rather than just configm ensures we do
+		// support cases where a previously unset path is set back.
+		return jsonRaw(result), nil
 
 	case map[string]interface{}:
 		// Update map to apply against pristine on commit.
@@ -86,61 +135,14 @@ func PatchConfig(snapName string, subkeys []string, pos int, config interface{},
 			return config, nil
 		}
 	}
-	panic(fmt.Errorf("internal error: unexpected configuration type %T", config))
-}
-
-// Get unmarshals into result the value of the provided snap's configuration key.
-// If the key does not exist, an error of type *NoOptionError is returned.
-// The provided key may be formed as a dotted key path through nested maps.
-// For example, the "a.b.c" key describes the {a: {b: {c: value}}} map.
-func GetFromChange(snapName string, subkeys []string, pos int, config map[string]interface{}, result interface{}) error {
-	// special case - get root document
-	if len(subkeys) == 0 {
-		if config == nil {
-			return &NoOptionError{SnapName: snapName, Key: ""}
-		}
-		raw := jsonRaw(config)
-
-		if err := jsonutil.DecodeWithNumber(bytes.NewReader(*raw), &result); err != nil {
-			return fmt.Errorf("internal error: cannot unmarshal snap %q root document: %s", snapName, err)
-		}
-		return nil
-	}
-	value, ok := config[subkeys[pos]]
-	if !ok {
-		return &NoOptionError{SnapName: snapName, Key: strings.Join(subkeys[:pos+1], ".")}
-	}
-
-	if pos+1 == len(subkeys) {
-		raw, ok := value.(*json.RawMessage)
-		if !ok {
-			raw = jsonRaw(value)
-		}
-		if err := jsonutil.DecodeWithNumber(bytes.NewReader(*raw), &result); err != nil {
-			key := strings.Join(subkeys, ".")
-			return fmt.Errorf("internal error: cannot unmarshal snap %q option %q into %T: %s, json: %s", snapName, key, result, err, *raw)
-		}
-		return nil
-	}
-
-	configm, ok := value.(map[string]interface{})
-	if !ok {
-		raw, ok := value.(*json.RawMessage)
-		if !ok {
-			raw = jsonRaw(value)
-		}
-		if err := jsonutil.DecodeWithNumber(bytes.NewReader(*raw), &configm); err != nil {
-			return fmt.Errorf("snap %q option %q is not a map", snapName, strings.Join(subkeys[:pos+1], "."))
-		}
-	}
-	return GetFromChange(snapName, subkeys, pos+1, configm, result)
+	return nil, fmt.Errorf("internal error: unexpected configuration type %T", config)
 }
 
 // GetSnapConfig retrieves the raw configuration of a given snap.
 func GetSnapConfig(st *state.State, snapName string) (*json.RawMessage, error) {
 	var config map[string]*json.RawMessage
 	err := st.Get("config", &config)
-	if err == state.ErrNoState {
+	if errors.Is(err, state.ErrNoState) {
 		return nil, nil
 	}
 	if err != nil {
@@ -157,8 +159,9 @@ func GetSnapConfig(st *state.State, snapName string) (*json.RawMessage, error) {
 func SetSnapConfig(st *state.State, snapName string, snapcfg *json.RawMessage) error {
 	var config map[string]*json.RawMessage
 	err := st.Get("config", &config)
-	isNil := snapcfg == nil || len(*snapcfg) == 0
-	if err == state.ErrNoState {
+	// empty nil snapcfg should be an empty message, but deal with "null" as well.
+	isNil := snapcfg == nil || len(*snapcfg) == 0 || bytes.Compare(*snapcfg, []byte("null")) == 0
+	if errors.Is(err, state.ErrNoState) {
 		if isNil {
 			// bail out early
 			return nil
@@ -185,7 +188,7 @@ func SaveRevisionConfig(st *state.State, snapName string, rev snap.Revision) err
 
 	// Get current configuration of the snap from state
 	err := st.Get("config", &config)
-	if err == state.ErrNoState {
+	if errors.Is(err, state.ErrNoState) {
 		return nil
 	} else if err != nil {
 		return fmt.Errorf("internal error: cannot unmarshal configuration: %v", err)
@@ -196,7 +199,7 @@ func SaveRevisionConfig(st *state.State, snapName string, rev snap.Revision) err
 	}
 
 	err = st.Get("revision-config", &revisionConfig)
-	if err == state.ErrNoState {
+	if errors.Is(err, state.ErrNoState) {
 		revisionConfig = make(map[string]map[string]*json.RawMessage)
 	} else if err != nil {
 		return err
@@ -219,14 +222,14 @@ func RestoreRevisionConfig(st *state.State, snapName string, rev snap.Revision)
 	var revisionConfig map[string]map[string]*json.RawMessage // snap => revision => configuration
 
 	err := st.Get("revision-config", &revisionConfig)
-	if err == state.ErrNoState {
+	if errors.Is(err, state.ErrNoState) {
 		return nil
 	} else if err != nil {
 		return fmt.Errorf("internal error: cannot unmarshal revision-config: %v", err)
 	}
 
 	err = st.Get("config", &config)
-	if err == state.ErrNoState {
+	if errors.Is(err, state.ErrNoState) {
 		config = make(map[string]*json.RawMessage)
 	} else if err != nil {
 		return fmt.Errorf("internal error: cannot unmarshal configuration: %v", err)
@@ -248,7 +251,7 @@ func RestoreRevisionConfig(st *state.State, snapName string, rev snap.Revision)
 func DiscardRevisionConfig(st *state.State, snapName string, rev snap.Revision) error {
 	var revisionConfig map[string]map[string]*json.RawMessage // snap => revision => configuration
 	err := st.Get("revision-config", &revisionConfig)
-	if err == state.ErrNoState {
+	if errors.Is(err, state.ErrNoState) {
 		return nil
 	} else if err != nil {
 		return fmt.Errorf("internal error: cannot unmarshal revision-config: %v", err)
@@ -271,7 +274,7 @@ func DeleteSnapConfig(st *state.State, snapName string) error {
 	var config map[string]map[string]*json.RawMessage // snap => key => value
 
 	err := st.Get("config", &config)
-	if err == state.ErrNoState {
+	if errors.Is(err, state.ErrNoState) {
 		return nil
 	} else if err != nil {
 		return fmt.Errorf("internal error: cannot unmarshal configuration: %v", err)
@@ -283,28 +286,39 @@ func DeleteSnapConfig(st *state.State, snapName string) error {
 	return nil
 }
 
-// Conf is an interface describing both state and transaction.
-type Conf interface {
-	Get(snapName, key string, result interface{}) error
+// ConfSetter is an interface for setting of config values.
+type ConfSetter interface {
 	Set(snapName, key string, value interface{}) error
-	Changes() []string
-	State() *state.State
 }
 
-// GetFeatureFlag returns the value of a given feature flag.
-func GetFeatureFlag(tr Conf, feature features.SnapdFeature) (bool, error) {
-	var isEnabled interface{}
-	snapName, confName := feature.ConfigOption()
-	if err := tr.Get(snapName, confName, &isEnabled); err != nil && !IsNoOption(err) {
-		return false, err
+// Patch sets values in cfg for the provided snap's configuration
+// based on patch.
+// patch keys can be dotted as the key argument to Set.
+// The patch is applied according to the order of its keys sorted by depth,
+// with top keys sorted first.
+func Patch(cfg ConfSetter, snapName string, patch map[string]interface{}) error {
+	patchKeys := sortPatchKeysByDepth(patch)
+	for _, key := range patchKeys {
+		if err := cfg.Set(snapName, key, patch[key]); err != nil {
+			return err
+		}
 	}
-	switch isEnabled {
-	case true, "true":
-		return true, nil
-	case false, "false":
-		return false, nil
-	case nil, "":
-		return feature.IsEnabledWhenUnset(), nil
+	return nil
+}
+
+func sortPatchKeysByDepth(patch map[string]interface{}) []string {
+	if len(patch) == 0 {
+		return nil
 	}
-	return false, fmt.Errorf("%s can only be set to 'true' or 'false', got %q", feature, isEnabled)
+	depths := make(map[string]int, len(patch))
+	keys := make([]string, 0, len(patch))
+	for k := range patch {
+		depths[k] = strings.Count(k, ".")
+		keys = append(keys, k)
+	}
+
+	sort.Slice(keys, func(i, j int) bool {
+		return depths[keys[i]] < depths[keys[j]]
+	})
+	return keys
 }
