@@ -1,9 +1,8 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 //go:build !nomanagers
-// +build !nomanagers
 
 /*
- * Copyright (C) 2020 Canonical Ltd
+ * Copyright (C) 2020-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -26,6 +25,7 @@ import (
 	"strings"
 
 	"github.com/snapcore/snapd/overlord/configstate/config"
+	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/release"
 	"github.com/snapcore/snapd/sysconfig"
 )
@@ -41,7 +41,9 @@ func init() {
 	addWithStateHandler(nil, setCloudInfoWhenSeeding, nil)
 
 	// proxy.{http,https,ftp}
-	addWithStateHandler(validateProxyStore, handleProxyConfiguration, coreOnly)
+	addWithStateHandler(nil, handleProxyConfiguration, coreOnly)
+	// proxy.store
+	addWithStateHandler(validateProxyStore, handleProxyStore, nil)
 
 	// resilience.vitality-hint
 	addWithStateHandler(validateVitalitySettings, handleVitalityConfiguration, nil)
@@ -60,25 +62,67 @@ func init() {
 	addWithStateHandler(validateAutomaticSnapshotsExpiration, nil, validateOnly)
 
 	// netplan.*
-	addWithStateHandler(validateNetplanSettings, handleNetplanConfiguration, &flags{coreOnlyConfig: true})
+	addWithStateHandler(validateNetplanSettings, handleNetplanConfiguration, coreOnly)
+
+	// kernel.{,dangerous-}cmdline-append
+	addWithStateHandler(validateCmdlineAppend, handleCmdlineAppend, &flags{modeenvOnlyConfig: true})
+
+	// debug.snapd.log
+	addWithStateHandler(validateDebugSnapdLogSetting, handleDebugSnapdLogConfiguration, nil)
+
+	// debug.systemd.log-level
+	addWithStateHandler(validateDebugSystemdLogLevelSetting, handleDebugSystemdLogLevelConfiguration, coreOnly)
+
+	// experimental.apparmor-prompting
+	addWithStateHandler(nil, doExperimentalApparmorPromptingDaemonRestart, nil)
+}
+
+// RunTransaction is an interface describing how to access
+// the system configuration state and transaction.
+type RunTransaction interface {
+	Get(snapName, key string, result interface{}) error
+	GetMaybe(snapName, key string, result interface{}) error
+	GetPristine(snapName, key string, result interface{}) error
+	Task() *state.Task
+	Set(snapName, key string, value interface{}) error
+	Changes() []string
+	State() *state.State
+	Commit()
+}
+
+// runTransactionImpl holds a transaction with a task that is in charge of
+// appliying a change to the configuration. It is used in the context of
+// configcore.
+type runTransactionImpl struct {
+	*config.Transaction
+	task *state.Task
+}
+
+func (rt *runTransactionImpl) Task() *state.Task {
+	return rt.task
+}
+
+func NewRunTransaction(tr *config.Transaction, tk *state.Task) RunTransaction {
+	runTransaction := &runTransactionImpl{Transaction: tr, task: tk}
+	return runTransaction
 }
 
 type withStateHandler struct {
-	validateFunc func(config.Conf) error
-	handleFunc   func(config.Conf, *fsOnlyContext) error
+	validateFunc func(RunTransaction) error
+	handleFunc   func(RunTransaction, *fsOnlyContext) error
 	configFlags  flags
 }
 
-func (h *withStateHandler) validate(cfg config.ConfGetter) error {
-	conf := cfg.(config.Conf)
+func (h *withStateHandler) validate(cfg ConfGetter) error {
+	conf := cfg.(RunTransaction)
 	if h.validateFunc != nil {
 		return h.validateFunc(conf)
 	}
 	return nil
 }
 
-func (h *withStateHandler) handle(dev sysconfig.Device, cfg config.ConfGetter, opts *fsOnlyContext) error {
-	conf := cfg.(config.Conf)
+func (h *withStateHandler) handle(dev sysconfig.Device, cfg ConfGetter, opts *fsOnlyContext) error {
+	conf := cfg.(RunTransaction)
 	if h.handleFunc != nil {
 		return h.handleFunc(conf, opts)
 	}
@@ -95,7 +139,7 @@ func (h *withStateHandler) flags() flags {
 
 // addWithStateHandler registers functions to validate and handle a subset of
 // system config options requiring to access and manipulate state.
-func addWithStateHandler(validate func(config.Conf) error, handle func(config.Conf, *fsOnlyContext) error, flags *flags) {
+func addWithStateHandler(validate func(RunTransaction) error, handle func(RunTransaction, *fsOnlyContext) error, flags *flags) {
 	if handle == nil && (flags == nil || !flags.validatedOnlyStateConfig) {
 		panic("cannot have nil handle with addWithStateHandler if validatedOnlyStateConfig flag is not set")
 	}
@@ -109,11 +153,11 @@ func addWithStateHandler(validate func(config.Conf) error, handle func(config.Co
 	handlers = append(handlers, h)
 }
 
-func Run(dev sysconfig.Device, cfg config.Conf) error {
+func Run(dev sysconfig.Device, cfg RunTransaction) error {
 	return applyHandlers(dev, cfg, handlers)
 }
 
-func applyHandlers(dev sysconfig.Device, cfg config.Conf, handlers []configHandler) error {
+func applyHandlers(dev sysconfig.Device, cfg RunTransaction, handlers []configHandler) error {
 	// check if the changes
 	for _, k := range cfg.Changes() {
 		switch {
@@ -140,6 +184,9 @@ func applyHandlers(dev sysconfig.Device, cfg config.Conf, handlers []configHandl
 		if h.flags().coreOnlyConfig && dev.Classic() {
 			continue
 		}
+		if h.flags().modeenvOnlyConfig && !dev.HasModeenv() {
+			continue
+		}
 		if err := h.handle(dev, cfg, nil); err != nil {
 			return err
 		}
@@ -147,7 +194,7 @@ func applyHandlers(dev sysconfig.Device, cfg config.Conf, handlers []configHandl
 	return nil
 }
 
-func Early(dev sysconfig.Device, cfg config.Conf, values map[string]interface{}) error {
+func Early(dev sysconfig.Device, cfg RunTransaction, values map[string]interface{}) error {
 	early, relevant := applyFilters(func(f flags) filterFunc {
 		return f.earlyConfigFilter
 	}, values)
