@@ -27,18 +27,21 @@ import (
 
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/asserts/assertstest"
+	"github.com/snapcore/snapd/confdb"
 	"github.com/snapcore/snapd/dirs"
+	"github.com/snapcore/snapd/features"
+	"github.com/snapcore/snapd/i18n"
 	"github.com/snapcore/snapd/interfaces"
 	"github.com/snapcore/snapd/interfaces/ifacetest"
 	"github.com/snapcore/snapd/overlord/assertstate"
 	"github.com/snapcore/snapd/overlord/assertstate/assertstatetest"
+	"github.com/snapcore/snapd/overlord/confdbstate"
 	"github.com/snapcore/snapd/overlord/configstate"
 	"github.com/snapcore/snapd/overlord/configstate/config"
 	"github.com/snapcore/snapd/overlord/hookstate"
 	"github.com/snapcore/snapd/overlord/hookstate/ctlcmd"
 	"github.com/snapcore/snapd/overlord/hookstate/hooktest"
 	"github.com/snapcore/snapd/overlord/ifacestate/ifacerepo"
-	"github.com/snapcore/snapd/overlord/registrystate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/testutil"
@@ -447,7 +450,7 @@ func (s *getAttrSuite) TestSlotHookTests(c *C) {
 	}
 }
 
-type registrySuite struct {
+type confdbSuite struct {
 	testutil.BaseTest
 
 	state     *state.State
@@ -458,9 +461,9 @@ type registrySuite struct {
 	mockHandler *hooktest.MockHandler
 }
 
-var _ = Suite(&registrySuite{})
+var _ = Suite(&confdbSuite{})
 
-func (s *registrySuite) SetUpTest(c *C) {
+func (s *confdbSuite) SetUpTest(c *C) {
 	s.BaseTest.SetUpTest(c)
 	dirs.SetRootDir(c.MkDir())
 	s.AddCleanup(func() {
@@ -533,14 +536,14 @@ func (s *registrySuite) SetUpTest(c *C) {
   }
 }`)
 
-	as, err := s.signingDB.Sign(asserts.RegistryType, headers, body, "")
+	as, err := s.signingDB.Sign(asserts.ConfdbType, headers, body, "")
 	c.Assert(err, IsNil)
 	c.Assert(assertstate.Add(s.state, as), IsNil)
 
 	repo := interfaces.NewRepository()
 	ifacerepo.Replace(s.state, repo)
 
-	regIface := &ifacetest.TestInterface{InterfaceName: "registry"}
+	regIface := &ifacetest.TestInterface{InterfaceName: "confdb"}
 	err = repo.AddInterface(regIface)
 	c.Assert(err, IsNil)
 
@@ -549,15 +552,19 @@ type: app
 version: 1
 plugs:
   read-wifi:
-    interface: registry
+    interface: confdb
     account: %[1]s
     view: network/read-wifi
     role: observer
   write-wifi:
-    interface: registry
+    interface: confdb
     account: %[1]s
     view: network/write-wifi
-    role: manager
+    role: custodian
+  other:
+    interface: confdb
+    account: %[1]s
+    view: other/other
 `, s.devAccID)
 	info := mockInstalledSnap(c, s.state, snapYaml, "")
 
@@ -570,8 +577,8 @@ plugs:
 version: 1.0
 type: os
 slots:
- registry-slot:
-  interface: registry
+ confdb-slot:
+  interface: confdb
 `
 	info = mockInstalledSnap(c, s.state, coreYaml, "")
 
@@ -583,19 +590,33 @@ slots:
 
 	ref := &interfaces.ConnRef{
 		PlugRef: interfaces.PlugRef{Snap: "test-snap", Name: "read-wifi"},
-		SlotRef: interfaces.SlotRef{Snap: "core", Name: "registry-slot"},
+		SlotRef: interfaces.SlotRef{Snap: "core", Name: "confdb-slot"},
 	}
 	_, err = repo.Connect(ref, nil, nil, nil, nil, nil)
 	c.Assert(err, IsNil)
+
+	s.setConfdbFlag(true, c)
 }
 
-func (s *registrySuite) TestRegistryGetSingleView(c *C) {
-	s.state.Lock()
-	err := registrystate.SetViaView(s.state, s.devAccID, "network", "write-wifi", map[string]interface{}{
-		"ssid": "my-ssid",
-	})
-	s.state.Unlock()
+func (s *confdbSuite) setConfdbFlag(val bool, c *C) {
+	tr := config.NewTransaction(s.state)
+	_, confOption := features.Confdbs.ConfigOption()
+	err := tr.Set("core", confOption, val)
 	c.Assert(err, IsNil)
+	tr.Commit()
+}
+
+func (s *confdbSuite) TestConfdbGetSingleView(c *C) {
+	restore := ctlcmd.MockConfdbstateNewTransaction(func(st *state.State, account string, confdbName string) (*confdbstate.Transaction, error) {
+		c.Assert(account, Equals, s.devAccID)
+		c.Assert(confdbName, Equals, "network")
+
+		tx, _ := confdbstate.NewTransaction(st, account, confdbName)
+		c.Assert(tx.Set("wifi.ssid", "my-ssid"), IsNil)
+
+		return tx, nil
+	})
+	defer restore()
 
 	stdout, stderr, err := ctlcmd.Run(s.mockContext, []string{"get", "--view", ":read-wifi", "ssid"}, 0)
 	c.Assert(err, IsNil)
@@ -603,14 +624,18 @@ func (s *registrySuite) TestRegistryGetSingleView(c *C) {
 	c.Check(stderr, IsNil)
 }
 
-func (s *registrySuite) TestRegistryGetManyViews(c *C) {
-	s.state.Lock()
-	err := registrystate.SetViaView(s.state, s.devAccID, "network", "write-wifi", map[string]interface{}{
-		"ssid":     "my-ssid",
-		"password": "secret",
+func (s *confdbSuite) TestConfdbGetManyViews(c *C) {
+	restore := ctlcmd.MockConfdbstateNewTransaction(func(st *state.State, account string, confdbName string) (*confdbstate.Transaction, error) {
+		c.Assert(account, Equals, s.devAccID)
+		c.Assert(confdbName, Equals, "network")
+
+		tx, _ := confdbstate.NewTransaction(st, account, confdbName)
+		c.Assert(tx.Set("wifi.ssid", "my-ssid"), IsNil)
+		c.Assert(tx.Set("wifi.psk", "secret"), IsNil)
+
+		return tx, nil
 	})
-	s.state.Unlock()
-	c.Assert(err, IsNil)
+	defer restore()
 
 	stdout, stderr, err := ctlcmd.Run(s.mockContext, []string{"get", "--view", ":read-wifi", "ssid", "password"}, 0)
 	c.Assert(err, IsNil)
@@ -622,14 +647,18 @@ func (s *registrySuite) TestRegistryGetManyViews(c *C) {
 	c.Check(stderr, IsNil)
 }
 
-func (s *registrySuite) TestRegistryGetNoRequest(c *C) {
-	s.state.Lock()
-	err := registrystate.SetViaView(s.state, s.devAccID, "network", "write-wifi", map[string]interface{}{
-		"ssid":     "my-ssid",
-		"password": "secret",
+func (s *confdbSuite) TestConfdbGetNoRequest(c *C) {
+	restore := ctlcmd.MockConfdbstateNewTransaction(func(st *state.State, account string, confdbName string) (*confdbstate.Transaction, error) {
+		c.Assert(account, Equals, s.devAccID)
+		c.Assert(confdbName, Equals, "network")
+
+		tx, _ := confdbstate.NewTransaction(st, account, confdbName)
+		c.Assert(tx.Set("wifi.ssid", "my-ssid"), IsNil)
+		c.Assert(tx.Set("wifi.psk", "secret"), IsNil)
+
+		return tx, nil
 	})
-	s.state.Unlock()
-	c.Assert(err, IsNil)
+	defer restore()
 
 	stdout, stderr, err := ctlcmd.Run(s.mockContext, []string{"get", "--view", ":read-wifi"}, 0)
 	c.Assert(err, IsNil)
@@ -641,58 +670,7 @@ func (s *registrySuite) TestRegistryGetNoRequest(c *C) {
 	c.Check(stderr, IsNil)
 }
 
-func (s *registrySuite) TestRegistryGetHappensTransactionally(c *C) {
-	s.state.Lock()
-	err := registrystate.SetViaView(s.state, s.devAccID, "network", "write-wifi", map[string]interface{}{
-		"ssid": "my-ssid",
-	})
-	s.state.Unlock()
-	c.Assert(err, IsNil)
-
-	// registry transaction is created when snapctl runs for the first time
-	stdout, stderr, err := ctlcmd.Run(s.mockContext, []string{"get", "--view", ":read-wifi"}, 0)
-	c.Assert(err, IsNil)
-	c.Check(string(stdout), Equals, `{
-	"ssid": "my-ssid"
-}
-`)
-	c.Check(stderr, IsNil)
-
-	s.state.Lock()
-	err = registrystate.SetViaView(s.state, s.devAccID, "network", "write-wifi", map[string]interface{}{
-		"ssid": "other-ssid",
-	})
-	s.state.Unlock()
-	c.Assert(err, IsNil)
-
-	// the new write wasn't reflected because it didn't run in the same transaction
-	stdout, stderr, err = ctlcmd.Run(s.mockContext, []string{"get", "--view", ":read-wifi"}, 0)
-	c.Assert(err, IsNil)
-	c.Check(string(stdout), Equals, `{
-	"ssid": "my-ssid"
-}
-`)
-	c.Check(stderr, IsNil)
-
-	// make a new context so we get a new transaction
-	s.state.Lock()
-	task := s.state.NewTask("test-task", "my test task")
-	setup := &hookstate.HookSetup{Snap: "test-snap", Revision: snap.R(1), Hook: "test-hook"}
-	s.mockContext, err = hookstate.NewContext(task, s.state, setup, s.mockHandler, "")
-	s.state.Unlock()
-	c.Assert(err, IsNil)
-
-	// now we get the new data
-	stdout, stderr, err = ctlcmd.Run(s.mockContext, []string{"get", "--view", ":read-wifi"}, 0)
-	c.Assert(err, IsNil)
-	c.Check(string(stdout), Equals, `{
-	"ssid": "other-ssid"
-}
-`)
-	c.Check(stderr, IsNil)
-}
-
-func (s *registrySuite) TestRegistryGetInvalid(c *C) {
+func (s *confdbSuite) TestConfdbGetInvalid(c *C) {
 	type testcase struct {
 		args []string
 		err  string
@@ -709,7 +687,7 @@ func (s *registrySuite) TestRegistryGetInvalid(c *C) {
 		},
 		{
 			args: []string{":non-existent"},
-			err:  `cannot get registry: cannot find plug :non-existent for snap "test-snap"`,
+			err:  `cannot find plug :non-existent for snap "test-snap"`,
 		},
 	}
 
@@ -721,7 +699,7 @@ func (s *registrySuite) TestRegistryGetInvalid(c *C) {
 	}
 }
 
-func (s *registrySuite) TestRegistryGetAndSetNonRegistryPlug(c *C) {
+func (s *confdbSuite) TestConfdbGetAndSetNonConfdbPlug(c *C) {
 	dirs.SetRootDir(c.MkDir())
 	s.AddCleanup(func() {
 		dirs.SetRootDir("/")
@@ -772,17 +750,17 @@ slots:
 	s.state.Unlock()
 
 	stdout, stderr, err := ctlcmd.Run(s.mockContext, []string{"get", "--view", ":my-plug"}, 0)
-	c.Assert(err, ErrorMatches, "cannot get registry: cannot use --view with non-registry plug :my-plug")
+	c.Assert(err, ErrorMatches, "cannot use --view with non-confdb plug :my-plug")
 	c.Check(stdout, IsNil)
 	c.Check(stderr, IsNil)
 
 	stdout, stderr, err = ctlcmd.Run(s.mockContext, []string{"set", "--view", ":my-plug", "ssid=my-ssid"}, 0)
-	c.Assert(err, ErrorMatches, "cannot set registry: cannot use --view with non-registry plug :my-plug")
+	c.Assert(err, ErrorMatches, "cannot use --view with non-confdb plug :my-plug")
 	c.Check(stdout, IsNil)
 	c.Check(stderr, IsNil)
 }
 
-func (s *registrySuite) TestRegistryGetAndSetAssertionNotFound(c *C) {
+func (s *confdbSuite) TestConfdbGetAndSetAssertionNotFound(c *C) {
 	storeSigning := assertstest.NewStoreStack("can0nical", nil)
 	db, err := asserts.OpenDatabase(&asserts.DatabaseConfig{
 		Backstore: asserts.NewMemoryBackstore(),
@@ -796,18 +774,17 @@ func (s *registrySuite) TestRegistryGetAndSetAssertionNotFound(c *C) {
 	s.state.Unlock()
 
 	stdout, stderr, err := ctlcmd.Run(s.mockContext, []string{"get", "--view", ":read-wifi"}, 0)
-	c.Assert(err, ErrorMatches, fmt.Sprintf("cannot get registry: registry assertion %s/network not found", s.devAccID))
+	c.Assert(err, ErrorMatches, fmt.Sprintf("cannot find confdb %s/network: assertion not found", s.devAccID))
 	c.Check(stdout, IsNil)
 	c.Check(stderr, IsNil)
 
 	stdout, stderr, err = ctlcmd.Run(s.mockContext, []string{"set", "--view", ":write-wifi", "ssid=my-ssid"}, 0)
-	c.Assert(err, ErrorMatches, fmt.Sprintf("cannot set registry: registry assertion %s/network not found", s.devAccID))
+	c.Assert(err, ErrorMatches, fmt.Sprintf("cannot find confdb %s/network: assertion not found", s.devAccID))
 	c.Check(stdout, IsNil)
 	c.Check(stderr, IsNil)
-
 }
 
-func (s *registrySuite) TestRegistryGetAndSetViewNotFound(c *C) {
+func (s *confdbSuite) TestConfdbGetAndSetViewNotFound(c *C) {
 	headers := map[string]interface{}{
 		"authority-id": s.devAccID,
 		"account-id":   s.devAccID,
@@ -831,19 +808,105 @@ func (s *registrySuite) TestRegistryGetAndSetViewNotFound(c *C) {
   }
 }`)
 
-	as, err := s.signingDB.Sign(asserts.RegistryType, headers, body, "")
+	as, err := s.signingDB.Sign(asserts.ConfdbType, headers, body, "")
 	c.Assert(err, IsNil)
 	s.state.Lock()
 	c.Assert(assertstate.Add(s.state, as), IsNil)
 	s.state.Unlock()
 
 	stdout, stderr, err := ctlcmd.Run(s.mockContext, []string{"get", "--view", ":read-wifi"}, 0)
-	c.Assert(err, ErrorMatches, fmt.Sprintf("cannot get registry: view \"read-wifi\" not found in registry %s/network", s.devAccID))
+	c.Assert(err, ErrorMatches, fmt.Sprintf("cannot find view \"read-wifi\" in confdb %s/network", s.devAccID))
 	c.Check(stdout, IsNil)
 	c.Check(stderr, IsNil)
 
 	stdout, stderr, err = ctlcmd.Run(s.mockContext, []string{"set", "--view", ":write-wifi", "ssid=my-ssid"}, 0)
-	c.Assert(err, ErrorMatches, fmt.Sprintf("cannot set registry: view \"write-wifi\" not found in registry %s/network", s.devAccID))
+	c.Assert(err, ErrorMatches, fmt.Sprintf("cannot find view \"write-wifi\" in confdb %s/network", s.devAccID))
 	c.Check(stdout, IsNil)
 	c.Check(stderr, IsNil)
 }
+
+func (s *confdbSuite) TestConfdbGetPristine(c *C) {
+	restore := ctlcmd.MockConfdbstateGetStoredTransaction(func(*state.Task) (*confdbstate.Transaction, func(), error) {
+		tx, _ := confdbstate.NewTransaction(s.state, s.devAccID, "network")
+		c.Assert(tx.Set("wifi.ssid", "foo"), IsNil)
+		c.Assert(tx.Commit(s.state, confdb.NewJSONSchema()), IsNil)
+
+		c.Assert(tx.Set("wifi.ssid", "bar"), IsNil)
+		return tx, func() {}, nil
+	})
+	defer restore()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+	task := s.state.NewTask("run-hook", "")
+	setup := &hookstate.HookSetup{Snap: "test-snap", Hook: "save-view-plug"}
+	ctx, err := hookstate.NewContext(task, s.state, setup, s.mockHandler, "")
+	c.Assert(err, IsNil)
+
+	s.state.Unlock()
+	defer s.state.Lock()
+
+	stdout, stderr, err := ctlcmd.Run(ctx, []string{"get", "--view", "--pristine", ":read-wifi", "ssid"}, 0)
+	c.Assert(err, IsNil)
+	c.Check(string(stdout), Equals, "foo\n")
+	c.Check(stderr, IsNil)
+
+	stdout, stderr, err = ctlcmd.Run(ctx, []string{"get", "--view", ":read-wifi", "ssid"}, 0)
+	c.Assert(err, IsNil)
+	c.Check(string(stdout), Equals, "bar\n")
+	c.Check(stderr, IsNil)
+}
+
+func (s *confdbSuite) TestConfdbGetDifferentViewThanOngoingTx(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	tx, err := confdbstate.NewTransaction(s.state, s.devAccID, "network")
+	c.Assert(err, IsNil)
+
+	err = tx.Set("wifi.ssid", "foo")
+	c.Assert(err, IsNil)
+
+	task := s.state.NewTask("run-hook", "")
+	setup := &hookstate.HookSetup{Snap: "test-snap", Hook: "save-view-plug"}
+	ctx, err := hookstate.NewContext(task, s.state, setup, s.mockHandler, "")
+	c.Assert(err, IsNil)
+
+	// set ongoing tx related to the network confdb
+	task.Set("confdb-transaction", tx)
+
+	s.state.Unlock()
+	defer s.state.Lock()
+
+	restore := ctlcmd.MockConfdbstateGetView(func(st *state.State, account, confdbName, viewName string) (*confdb.View, error) {
+		reg, err := confdb.New(s.devAccID, "other", map[string]interface{}{
+			"other": map[string]interface{}{
+				"rules": []interface{}{
+					map[string]interface{}{"request": "ssid", "storage": "ssid"},
+				},
+			},
+		}, confdb.NewJSONSchema())
+		c.Assert(err, IsNil)
+		return reg.View("other"), nil
+	})
+	defer restore()
+
+	stdout, stderr, err := ctlcmd.Run(ctx, []string{"get", "--view", ":other", "ssid"}, 0)
+	// error is for no stored value, meaning we read the right confdb
+	c.Assert(err, ErrorMatches, `.*: no view data`)
+	c.Check(stdout, IsNil)
+	c.Check(stderr, IsNil)
+}
+
+func (s *confdbSuite) TestConfdbExperimentalFlag(c *C) {
+	s.state.Lock()
+	s.setConfdbFlag(false, c)
+	s.state.Unlock()
+
+	for _, cmd := range []string{"get", "set", "unset"} {
+		stdout, stderr, err := ctlcmd.Run(s.mockContext, []string{cmd, "--view", ":read-wifi"}, 0)
+		c.Assert(err, ErrorMatches, i18n.G(`"confdbs" feature flag is disabled: set 'experimental.confdbs' to true`))
+		c.Check(stdout, IsNil)
+		c.Check(stderr, IsNil)
+	}
+}
