@@ -24,11 +24,12 @@ import (
 
 	. "gopkg.in/check.v1"
 
+	"github.com/snapcore/snapd/confdb"
+	"github.com/snapcore/snapd/overlord/confdbstate"
 	"github.com/snapcore/snapd/overlord/configstate/config"
 	"github.com/snapcore/snapd/overlord/hookstate"
 	"github.com/snapcore/snapd/overlord/hookstate/ctlcmd"
 	"github.com/snapcore/snapd/overlord/hookstate/hooktest"
-	"github.com/snapcore/snapd/overlord/registrystate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/snap"
 )
@@ -162,57 +163,37 @@ func (s *unsetSuite) TestCommandWithoutContext(c *C) {
 	c.Check(err, ErrorMatches, `cannot invoke snapctl operation commands \(here "unset"\) from outside of a snap`)
 }
 
-func (s *registrySuite) TestRegistryUnsetManyViews(c *C) {
+func (s *confdbSuite) TestConfdbUnsetManyViews(c *C) {
 	s.state.Lock()
-	err := registrystate.SetViaView(s.state, s.devAccID, "network", "write-wifi", map[string]interface{}{"ssid": "my-ssid", "password": "my-secret"})
+	tx, err := confdbstate.NewTransaction(s.state, s.devAccID, "network")
 	s.state.Unlock()
 	c.Assert(err, IsNil)
 
+	err = tx.Set("wifi.ssid", "foo")
+	c.Assert(err, IsNil)
+
+	err = tx.Set("wifi.psk", "bar")
+	c.Assert(err, IsNil)
+
+	ctlcmd.MockConfdbstateGetTransaction(func(*hookstate.Context, *state.State, *confdb.View) (*confdbstate.Transaction, confdbstate.CommitTxFunc, error) {
+		return tx, nil, nil
+	})
+
 	stdout, stderr, err := ctlcmd.Run(s.mockContext, []string{"unset", "--view", ":write-wifi", "ssid", "password"}, 0)
 	c.Assert(err, IsNil)
 	c.Check(stdout, IsNil)
 	c.Check(stderr, IsNil)
-	s.mockContext.Lock()
-	c.Assert(s.mockContext.Done(), IsNil)
-	s.mockContext.Unlock()
+
+	_, err = tx.Get("wifi.ssid")
+	c.Assert(err, ErrorMatches, `no value was found under path "wifi.ssid"`)
 
 	s.state.Lock()
-	_, err = registrystate.GetViaView(s.state, s.devAccID, "network", "write-wifi", []string{"ssid", "password"})
+	_, err = confdbstate.Get(s.state, s.devAccID, "network", "write-wifi", []string{"ssid", "password"})
 	s.state.Unlock()
-	c.Assert(err, ErrorMatches, `cannot get "ssid", "password" .*: matching rules don't map to any values`)
+	c.Assert(err, ErrorMatches, `cannot get "ssid", "password" .*: no view data`)
 }
 
-func (s *registrySuite) TestRegistryUnsetHappensTransactionally(c *C) {
-	s.state.Lock()
-	err := registrystate.SetViaView(s.state, s.devAccID, "network", "write-wifi", map[string]interface{}{"ssid": "my-ssid"})
-	s.state.Unlock()
-	c.Assert(err, IsNil)
-
-	stdout, stderr, err := ctlcmd.Run(s.mockContext, []string{"unset", "--view", ":write-wifi", "ssid"}, 0)
-	c.Assert(err, IsNil)
-	c.Check(stdout, IsNil)
-	c.Check(stderr, IsNil)
-
-	s.state.Lock()
-	val, err := registrystate.GetViaView(s.state, s.devAccID, "network", "read-wifi", []string{"ssid"})
-	s.state.Unlock()
-	c.Assert(err, IsNil)
-	c.Assert(val, DeepEquals, map[string]interface{}{
-		"ssid": "my-ssid",
-	})
-
-	// commit transaction
-	s.mockContext.Lock()
-	c.Assert(s.mockContext.Done(), IsNil)
-	s.mockContext.Unlock()
-
-	s.state.Lock()
-	_, err = registrystate.GetViaView(s.state, s.devAccID, "network", "read-wifi", []string{"ssid"})
-	s.state.Unlock()
-	c.Assert(err, ErrorMatches, `cannot get "ssid" .*: matching rules don't map to any values`)
-}
-
-func (s *registrySuite) TestRegistryUnsetInvalid(c *C) {
+func (s *confdbSuite) TestConfdbUnsetInvalid(c *C) {
 	type testcase struct {
 		args []string
 		err  string
@@ -221,15 +202,15 @@ func (s *registrySuite) TestRegistryUnsetInvalid(c *C) {
 	tcs := []testcase{
 		{
 			args: []string{"snap:plug"},
-			err:  `cannot unset registry: plug must conform to format ":<plug-name>": snap:plug`,
+			err:  `cannot unset confdb: plug must conform to format ":<plug-name>": snap:plug`,
 		},
 		{
 			args: []string{":"},
-			err:  `cannot unset registry: plug name was not provided`,
+			err:  `cannot unset confdb: plug name was not provided`,
 		},
 		{
 			args: []string{":plug"},
-			err:  `cannot unset registry: no paths provided to unset`,
+			err:  `cannot unset confdb: no paths provided to unset`,
 		},
 	}
 
