@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2016-2017 Canonical Ltd
+ * Copyright (C) 2016-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -28,6 +28,7 @@ import (
 	"github.com/snapcore/snapd/interfaces/backends"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/overlord/hookstate"
+	"github.com/snapcore/snapd/overlord/ifacestate/apparmorprompting"
 	"github.com/snapcore/snapd/overlord/ifacestate/ifacerepo"
 	"github.com/snapcore/snapd/overlord/ifacestate/udevmonitor"
 	"github.com/snapcore/snapd/overlord/snapstate"
@@ -63,6 +64,11 @@ type InterfaceManager struct {
 	extraInterfaces []interfaces.Interface
 	extraBackends   []interfaces.SecurityBackend
 
+	// AppArmor Prompting
+	useAppArmorPrompting        bool
+	interfacesRequestsManagerMu sync.Mutex
+	interfacesRequestsManager   *apparmorprompting.InterfacesRequestsManager
+
 	preseed bool
 }
 
@@ -132,6 +138,28 @@ func Manager(s *state.State, hookManager *hookstate.HookManager, runner *state.T
 	return m, nil
 }
 
+// AppArmorPromptingRunning returns true if prompting is running.
+// This method may only be called after StartUp has been called on the manager.
+func (m *InterfaceManager) AppArmorPromptingRunning() bool {
+	return m.useAppArmorPrompting
+}
+
+// Allow m.UseAppArmorPrompting to be mocked in tests
+var assessAppArmorPrompting = func(m *InterfaceManager) bool {
+	return m.assesAppArmorPrompting()
+}
+
+// InterfacesRequestsManager returns the interfaces requests manager associated
+// with the receiver. This method may only be called after StartUp has been
+// called, and will return nil if AppArmor prompting is not running.
+func (m *InterfaceManager) InterfacesRequestsManager() apparmorprompting.Manager {
+	irm := m.interfacesRequestsManager
+	if irm == nil {
+		return nil
+	}
+	return irm
+}
+
 // StartUp implements StateStarterUp.Startup.
 func (m *InterfaceManager) StartUp() error {
 	s := m.state
@@ -140,7 +168,13 @@ func (m *InterfaceManager) StartUp() error {
 	s.Lock()
 	defer s.Unlock()
 
-	snaps, err := snapsWithSecurityProfiles(m.state)
+	// Check whether AppArmor prompting is supported and enabled. It is fine to
+	// do this once, as toggling the feature imposes a restart of snapd.
+	if assessAppArmorPrompting(m) {
+		m.useAppArmorPrompting = true
+	}
+
+	appSets, err := snapsWithSecurityProfiles(m.state)
 	if err != nil {
 		return err
 	}
@@ -149,7 +183,7 @@ func (m *InterfaceManager) StartUp() error {
 	// duration of this process always add implicit slots to snapd and not to
 	// any other type: os snap and use a mapper to use names core-snapd-system
 	// on state, in memory and in API responses, respectively.
-	m.selectInterfaceMapper(snaps)
+	m.selectInterfaceMapper(appSets)
 
 	if err := m.addInterfaces(m.extraInterfaces); err != nil {
 		return err
@@ -157,7 +191,7 @@ func (m *InterfaceManager) StartUp() error {
 	if err := m.addBackends(m.extraBackends); err != nil {
 		return err
 	}
-	if err := m.addSnaps(snaps); err != nil {
+	if err := m.addAppSets(appSets); err != nil {
 		return err
 	}
 	if err := m.renameCorePlugConnection(); err != nil {
@@ -169,7 +203,43 @@ func (m *InterfaceManager) StartUp() error {
 	if _, err := m.reloadConnections(""); err != nil {
 		return err
 	}
-	if profilesNeedRegeneration() {
+
+	if m.useAppArmorPrompting {
+		// Check if there is at least one snap on the system which has a
+		// connection using the "snap-interfaces-requests-control" plug
+		// with a "handler-service" attribute declared.
+		present, err := interfacesRequestsControlHandlerServicePresent(m)
+		if err != nil {
+			// Internal error, should not occur
+			logger.Noticef("failed to check the presence of a interfaces-requests-control handler service: %v", err)
+		} else if !present {
+			m.state.AddWarning(`"apparmor-prompting" feature flag enabled but no prompting client is present; requests will be auto-denied until a prompting client is installed`, nil)
+		}
+
+		func() {
+			// Must not hold state lock while starting interfaces requests
+			// manager, so that notices can be recorded if needed.
+			m.state.Unlock()
+			defer m.state.Lock()
+			if err := m.initInterfacesRequestsManager(); err != nil {
+				logger.Noticef("failed to start interfaces requests manager: %v", err)
+				// Set m.useAppArmorPrompting to false so external callers
+				// don't try to access nil backends.
+				m.useAppArmorPrompting = false
+				// This is done before profilesNeedRegeneration, so profiles
+				// will only be regenerated if prompting is newly enabled and
+				// the backends were successfully created.
+
+				// Do not set "apparmor-prompting" flag to false, since the
+				// user intends for prompting to be enabled, but do record a
+				// warning so the user knows prompting is not current running.
+				m.state.Lock()
+				defer m.state.Unlock()
+				m.state.AddWarning(fmt.Sprintf("cannot start prompting backend: %v; prompting will be inactive until snapd is restarted", err), nil)
+			}
+		}()
+	}
+	if m.profilesNeedRegeneration() {
 		if err := m.regenerateAllSecurityProfiles(perfTimings); err != nil {
 			return err
 		}
@@ -224,9 +294,17 @@ func (m *InterfaceManager) Ensure() error {
 	return nil
 }
 
-// Stop implements StateStopper. It stops the udev monitor,
+// Stop implements StateStopper. It stops the udev monitor and prompting,
 // if running.
 func (m *InterfaceManager) Stop() {
+	m.stopUDevMon()
+	// The state lock is not held when calling any of the manager methods
+	// driven by StateEngine. Thus, it is okay for stopInterfacesRequestsManager
+	// to acquire the state lock in order to record notices, if needed.
+	m.stopInterfacesRequestsManager()
+}
+
+func (m *InterfaceManager) stopUDevMon() {
 	m.udevMonMu.Lock()
 	udevMon := m.udevMon
 	m.udevMonMu.Unlock()
@@ -241,6 +319,29 @@ func (m *InterfaceManager) Stop() {
 	m.udevMon = nil
 }
 
+// interfacesRequestsManagerStop calls stop on the given manager. The state lock
+// must not be held while this function is called, as the manager may need to
+// record notices while it is stopping.
+var interfacesRequestsManagerStop = func(interfacesRequestsManager *apparmorprompting.InterfacesRequestsManager) error {
+	return interfacesRequestsManager.Stop()
+}
+
+func (m *InterfaceManager) stopInterfacesRequestsManager() {
+	m.interfacesRequestsManagerMu.Lock()
+	defer m.interfacesRequestsManagerMu.Unlock()
+	// May as well hold the interfacesRequestsManager lock while stopping prompting, so that
+	// we don't try to use or overwrite this prompting instance while it is
+	// stopping.
+	interfacesRequestsManager := m.interfacesRequestsManager
+	m.interfacesRequestsManager = nil
+	if interfacesRequestsManager == nil {
+		return
+	}
+	if err := interfacesRequestsManagerStop(interfacesRequestsManager); err != nil {
+		logger.Noticef("Cannot stop prompting: %s", err)
+	}
+}
+
 // Repository returns the interface repository used internally by the manager.
 //
 // This method has two use-cases:
@@ -448,8 +549,9 @@ func (m *InterfaceManager) DisableUDevMonitor() {
 }
 
 var (
-	udevInitRetryTimeout = time.Minute * 5
-	createUDevMonitor    = udevmonitor.New
+	udevInitRetryTimeout            = time.Minute * 5
+	createUDevMonitor               = udevmonitor.New
+	createInterfacesRequestsManager = apparmorprompting.New
 )
 
 func (m *InterfaceManager) initUDevMonitor() error {
@@ -467,6 +569,39 @@ func (m *InterfaceManager) initUDevMonitor() error {
 	return nil
 }
 
+// interfacesRequestsControlHandlerServicePresent returns true if there is at
+// least one snap which has a "snap-interfaces-requests-control" connection
+// with an app declared by the "handler-service" attribute.
+//
+// The caller must ensure that the state lock is held.
+var interfacesRequestsControlHandlerServicePresent = func(m *InterfaceManager) (bool, error) {
+	handlers, err := InterfacesRequestsControlHandlerServices(m.state)
+	if err != nil {
+		return false, err
+	}
+	return len(handlers) > 0, nil
+}
+
+// initInterfacesRequestsManager initializes the prompting backends which make
+// up the interfaces requests manager.
+//
+// This function should only be called if prompting is supported and enabled,
+// and at least one installed snap has a "snap-interfaces-requests-control"
+// connection with the "handler-service" attribute declared.
+//
+// The state lock must not be held when this function is called, so that
+// notices can be recorded if necessary.
+func (m *InterfaceManager) initInterfacesRequestsManager() error {
+	m.interfacesRequestsManagerMu.Lock()
+	defer m.interfacesRequestsManagerMu.Unlock()
+	interfacesRequestsManager, err := createInterfacesRequestsManager(m.state)
+	if err != nil {
+		return err
+	}
+	m.interfacesRequestsManager = interfacesRequestsManager
+	return nil
+}
+
 var securityBackendsOverride []interfaces.SecurityBackend
 
 // allSecurityBackends returns a set of the available security backends or the mocked ones, ready to be initialized.
