@@ -29,7 +29,6 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"os"
 	"path"
 	"path/filepath"
@@ -63,10 +62,11 @@ var (
 	// Stop is used to ask Iter to stop iteration, without it being an error.
 	Stop = errors.New("stop iteration")
 
-	osOpen      = os.Open
-	dirNames    = (*os.File).Readdirnames
-	backendOpen = Open
-	timeNow     = time.Now
+	osOpen               = os.Open
+	dirNames             = (*os.File).Readdirnames
+	backendOpen          = Open
+	timeNow              = time.Now
+	snapReadSnapshotYaml = snap.ReadSnapshotYaml
 
 	usersForUsernames = usersForUsernamesImpl
 )
@@ -258,7 +258,7 @@ func isSnapshotFilename(filePath string) (ok bool, setID uint64) {
 }
 
 // EstimateSnapshotSize calculates estimated size of the snapshot.
-func EstimateSnapshotSize(si *snap.Info, usernames []string) (uint64, error) {
+func EstimateSnapshotSize(si *snap.Info, usernames []string, dirOpts *dirs.SnapDirOptions) (uint64, error) {
 	var total uint64
 	calculateSize := func(path string, finfo os.FileInfo, err error) error {
 		if finfo.Mode().IsRegular() {
@@ -284,15 +284,15 @@ func EstimateSnapshotSize(si *snap.Info, usernames []string) (uint64, error) {
 		}
 	}
 
-	users, err := usersForUsernames(usernames)
+	users, err := usersForUsernames(usernames, dirOpts)
 	if err != nil {
 		return 0, err
 	}
 	for _, usr := range users {
-		if err := visitDir(si.UserDataDir(usr.HomeDir)); err != nil {
+		if err := visitDir(si.UserDataDir(usr.HomeDir, dirOpts)); err != nil {
 			return 0, err
 		}
-		if err := visitDir(si.UserCommonDataDir(usr.HomeDir)); err != nil {
+		if err := visitDir(si.UserCommonDataDir(usr.HomeDir, dirOpts)); err != nil {
 			return 0, err
 		}
 	}
@@ -302,7 +302,7 @@ func EstimateSnapshotSize(si *snap.Info, usernames []string) (uint64, error) {
 }
 
 // Save a snapshot
-func Save(ctx context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string) (*client.Snapshot, error) {
+func Save(ctx context.Context, id uint64, si *snap.Info, cfg map[string]interface{}, usernames []string, dynSnapshotOpts *snap.SnapshotOptions, dirOpts *dirs.SnapDirOptions) (*client.Snapshot, error) {
 	if err := os.MkdirAll(dirs.SnapshotsDir, 0700); err != nil {
 		return nil, err
 	}
@@ -315,12 +315,26 @@ func Save(ctx context.Context, id uint64, si *snap.Info, cfg map[string]interfac
 		Version:  si.Version,
 		Epoch:    si.Epoch,
 		Time:     timeNow(),
+		// Pass only dynamic snapshot options here. Static options are tied to the snap version
+		// and should not be repeated in snapshot metadata on every save.
+		Options:  dynSnapshotOpts,
 		SHA3_384: make(map[string]string),
 		Size:     0,
 		Conf:     cfg,
 		// Note: Auto is no longer set in the Snapshot.
 	}
 
+	snapshotOptions, err := snapReadSnapshotYaml(si)
+	if err != nil {
+		return nil, err
+	}
+
+	if dynSnapshotOpts != nil {
+		if err := snapshotOptions.MergeDynamicExcludes(dynSnapshotOpts.Exclude); err != nil {
+			return nil, err
+		}
+	}
+
 	aw, err := osutil.NewAtomicFile(Filename(snapshot), 0600, 0, osutil.NoChown, osutil.NoChown)
 	if err != nil {
 		return nil, err
@@ -330,17 +344,21 @@ func Save(ctx context.Context, id uint64, si *snap.Info, cfg map[string]interfac
 
 	w := zip.NewWriter(aw)
 	defer w.Close() // note this does not close the file descriptor (that's done by hand on the atomic writer, above)
-	if err := addDirToZip(ctx, snapshot, w, "root", archiveName, si.DataDir()); err != nil {
+	savingUserData := false
+	baseDataDir := snap.BaseDataDir(si.InstanceName())
+	if err := addSnapDirToZip(ctx, snapshot, w, "root", archiveName, baseDataDir, savingUserData, snapshotOptions.Exclude); err != nil {
 		return nil, err
 	}
 
-	users, err := usersForUsernames(usernames)
+	users, err := usersForUsernames(usernames, dirOpts)
 	if err != nil {
 		return nil, err
 	}
 
+	savingUserData = true
 	for _, usr := range users {
-		if err := addDirToZip(ctx, snapshot, w, usr.Username, userArchiveName(usr), si.UserDataDir(usr.HomeDir)); err != nil {
+		snapDataDir := filepath.Dir(si.UserDataDir(usr.HomeDir, dirOpts))
+		if err := addSnapDirToZip(ctx, snapshot, w, usr.Username, userArchiveName(usr), snapDataDir, savingUserData, snapshotOptions.Exclude); err != nil {
 			return nil, err
 		}
 	}
@@ -378,65 +396,79 @@ func Save(ctx context.Context, id uint64, si *snap.Info, cfg map[string]interfac
 
 var isTesting = snapdenv.Testing()
 
-func addDirToZip(ctx context.Context, snapshot *client.Snapshot, w *zip.Writer, username string, entry, dir string) error {
-	parent, revdir := filepath.Split(dir)
-	exists, isDir, err := osutil.DirExists(parent)
+// addSnapDirToZip adds the 'common' and the 'rev' revisioned dir under 'snapDir'
+// to the snapshot. If one doesn't exist, it's ignored. If none exists, the
+// operation is skipped.
+func addSnapDirToZip(ctx context.Context, snapshot *client.Snapshot, w *zip.Writer, username, entry, snapDir string, savingUserData bool, excludePaths []string) error {
+	paths, err := pathsForSnapshot(snapDir, snapshot)
 	if err != nil {
 		return err
 	}
-	if exists && !isDir {
-		logger.Noticef("Not saving directories under %q in snapshot #%d of %q as it is not a directory.", parent, snapshot.SetID, snapshot.Snap)
+
+	if len(paths) == 0 {
 		return nil
 	}
-	if !exists {
-		logger.Debugf("Not saving directories under %q in snapshot #%d of %q as it is does not exist.", parent, snapshot.SetID, snapshot.Snap)
-		return nil
+
+	expandSnapDataDirs := func(varName string) string {
+		// Validation of the environment variables has already been performed.
+		// We just need to make sure that we consider the right variables
+		// according to whether we are saving user or system data.
+		switch {
+		case varName == "SNAP_COMMON" && !savingUserData:
+			fallthrough
+		case varName == "SNAP_USER_COMMON" && savingUserData:
+			return "common"
+		case varName == "SNAP_DATA" && !savingUserData:
+			fallthrough
+		case varName == "SNAP_USER_DATA" && savingUserData:
+			return snapshot.Revision.String()
+		}
+		// The variable specified does not match the current operating mode
+		// (for example, the variable is SNAP_COMMON but we are saving user
+		// data); in this case, we need to inform our caller that the returned
+		// string should be ignored and not added to the "tar" parameters. In
+		// order to do this, we return a "-" as a sentinel.
+		return "-"
 	}
+
+	var expExcludePaths []string
+	for _, excludePath := range excludePaths {
+		expandedPath := os.Expand(excludePath, expandSnapDataDirs)
+		// "-" is the sentinel returned by expandSnapDataDirs() if the
+		// exclusion path is not relevant for the type of data being considered
+		if expandedPath[0] == '-' {
+			continue
+		}
+		expExcludePaths = append(expExcludePaths, expandedPath)
+	}
+
+	return addToZip(ctx, snapshot, w, username, entry, paths, expExcludePaths)
+}
+
+// addToZip adds 'paths' to the snapshot. tar will change into the paths' parent
+// directory before creating the archive so that parent dirs are not added.
+func addToZip(ctx context.Context, snapshot *client.Snapshot, w *zip.Writer, username, entry string, paths []string, excludePaths []string) error {
+	archiveWriter, err := w.CreateHeader(&zip.FileHeader{Name: entry})
+	if err != nil {
+		return err
+	}
+
 	tarArgs := []string{
 		"--create",
 		"--sparse", "--gzip",
 		"--format", "gnu",
-		"--directory", parent,
+		"--anchored",
+		"--no-wildcards-match-slash",
 	}
 
-	noRev, noCommon := true, true
-
-	exists, isDir, err = osutil.DirExists(dir)
-	if err != nil {
-		return err
-	}
-	switch {
-	case exists && isDir:
-		tarArgs = append(tarArgs, revdir)
-		noRev = false
-	case exists && !isDir:
-		logger.Noticef("Not saving %q in snapshot #%d of %q as it is not a directory.", dir, snapshot.SetID, snapshot.Snap)
-	case !exists:
-		logger.Debugf("Not saving %q in snapshot #%d of %q as it is does not exist.", dir, snapshot.SetID, snapshot.Snap)
+	for _, path := range excludePaths {
+		tarArgs = append(tarArgs, fmt.Sprintf("--exclude=%s", path))
 	}
 
-	common := filepath.Join(parent, "common")
-	exists, isDir, err = osutil.DirExists(common)
-	if err != nil {
-		return err
-	}
-	switch {
-	case exists && isDir:
-		tarArgs = append(tarArgs, "common")
-		noCommon = false
-	case exists && !isDir:
-		logger.Noticef("Not saving %q in snapshot #%d of %q as it is not a directory.", common, snapshot.SetID, snapshot.Snap)
-	case !exists:
-		logger.Debugf("Not saving %q in snapshot #%d of %q as it is does not exist.", common, snapshot.SetID, snapshot.Snap)
-	}
-
-	if noCommon && noRev {
-		return nil
-	}
-
-	archiveWriter, err := w.CreateHeader(&zip.FileHeader{Name: entry})
-	if err != nil {
-		return err
+	// use --directory so that the directory is added without its parent dirs
+	for _, path := range paths {
+		parent, dir := filepath.Split(path)
+		tarArgs = append(tarArgs, "--directory", parent, dir)
 	}
 
 	var sz osutil.Sizer
@@ -444,16 +476,30 @@ func addDirToZip(ctx context.Context, snapshot *client.Snapshot, w *zip.Writer,
 
 	cmd := tarAsUser(username, tarArgs...)
 	cmd.Stdout = io.MultiWriter(archiveWriter, hasher, &sz)
-	matchCounter := &strutil.MatchCounter{N: 1}
+
+	// keep (at most) the last 5 non-empty lines of what 'tar' writes to stderr
+	// (those are the most likely contain the reason for fatal errors)
+	matchCounter := &strutil.MatchCounter{
+		N:     5,
+		LastN: true,
+	}
 	cmd.Stderr = matchCounter
+
 	if isTesting {
 		matchCounter.N = -1
 		cmd.Stderr = io.MultiWriter(os.Stderr, matchCounter)
 	}
+
 	if err := osutil.RunWithContext(ctx, cmd); err != nil {
 		matches, count := matchCounter.Matches()
 		if count > 0 {
-			return fmt.Errorf("cannot create archive: %s (and %d more)", matches[0], count-1)
+			note := ""
+			if count > 5 {
+				note = fmt.Sprintf(" (showing last 5 lines out of %d)", count)
+			}
+			// we have at most 5 matches here
+			errStr := strings.Join(matches, "\n")
+			return fmt.Errorf("cannot create archive%s:\n%s", note, errStr)
 		}
 		return fmt.Errorf("tar failed: %v", err)
 	}
@@ -464,6 +510,41 @@ func addDirToZip(ctx context.Context, snapshot *client.Snapshot, w *zip.Writer,
 	return nil
 }
 
+// pathsForSnapshot returns a list of absolute paths under 'snapDir' that should
+// be included in the snapshot (based on what directories exist).
+func pathsForSnapshot(snapDir string, snapshot *client.Snapshot) ([]string, error) {
+	dirExists := func(path string) (bool, error) {
+		exists, isDir, err := osutil.DirExists(path)
+		if err != nil {
+			return false, err
+		}
+
+		if exists && isDir {
+			return true, nil
+		}
+
+		if !exists {
+			logger.Debugf("Not saving %q in snapshot #%d of %q as it is does not exist.", path, snapshot.SetID, snapshot.Snap)
+		} else if !isDir {
+			logger.Noticef("Not saving %q in snapshot #%d of %q as it is not a directory.", path, snapshot.SetID, snapshot.Snap)
+		}
+
+		return false, nil
+	}
+
+	var snapshotPaths []string
+	for _, subDir := range []string{snapshot.Revision.String(), "common"} {
+		subPath := filepath.Join(snapDir, subDir)
+		if ok, err := dirExists(subPath); err != nil {
+			return nil, err
+		} else if ok {
+			snapshotPaths = append(snapshotPaths, subPath)
+		}
+	}
+
+	return snapshotPaths, nil
+}
+
 var ErrCannotCancel = errors.New("cannot cancel: import already finished")
 
 // multiError collects multiple errors that affected an operation.
@@ -485,7 +566,7 @@ func (me *multiError) Error() string {
 	return me.nestedError(0)
 }
 
-// helper to ensure formating of nested multiErrors works.
+// helper to ensure formatting of nested multiErrors works.
 func (me *multiError) nestedError(level int) string {
 	indent := strings.Repeat(" ", level)
 	buf := bytes.NewBufferString(fmt.Sprintf("%s:\n", me.header))
@@ -604,7 +685,7 @@ func (t *importTransaction) Commit() error {
 }
 
 func (t *importTransaction) lock() error {
-	return ioutil.WriteFile(t.lockPath, nil, 0644)
+	return os.WriteFile(t.lockPath, nil, 0644)
 }
 
 func (t *importTransaction) unlock() error {
@@ -613,14 +694,14 @@ func (t *importTransaction) unlock() error {
 
 var filepathGlob = filepath.Glob
 
-// CleanupAbandondedImports will clean any import that is in progress.
+// CleanupAbandonedImports will clean any import that is in progress.
 // This is meant to be called at startup of snapd before any real imports
 // happen. It is not safe to run this concurrently with any other snapshot
 // operation.
 //
 // The amount of snapshots cleaned is returned and an error if one or
 // more cleanups did not succeed.
-func CleanupAbandondedImports() (cleaned int, err error) {
+func CleanupAbandonedImports() (cleaned int, err error) {
 	inProgressSnapshots, err := filepathGlob(filepath.Join(dirs.SnapshotsDir, importingFnGlob))
 	if err != nil {
 		return 0, err
@@ -654,6 +735,10 @@ type ImportFlags struct {
 
 // Import a snapshot from the export file format
 func Import(ctx context.Context, id uint64, r io.Reader, flags *ImportFlags) (snapNames []string, err error) {
+	if err := os.MkdirAll(dirs.SnapshotsDir, 0700); err != nil {
+		return nil, err
+	}
+
 	errPrefix := fmt.Sprintf("cannot import snapshot %d", id)
 
 	tr := newImportTransaction(id)
@@ -765,6 +850,11 @@ func unpackVerifySnapshotImport(ctx context.Context, r io.Reader, realSetID uint
 			return nil, errors.New("unexpected directory in import file")
 		}
 
+		// files within the snapshot should never use parent elements
+		if strings.Contains(header.Name, "../") {
+			return nil, fmt.Errorf("invalid filename in import file")
+		}
+
 		if header.Name == "content.json" {
 			var ej contentJSON
 			dec := json.NewDecoder(tr)
@@ -908,7 +998,7 @@ func NewSnapshotExport(ctx context.Context, setID uint64) (se *SnapshotExport, e
 // so it should be called without any locks. The SnapshotExport
 // keeps the FDs open so even files moved/deleted will be found.
 func (se *SnapshotExport) Init() error {
-	// Export once into a dummy writer so that we can set the size
+	// Export once into a fake writer so that we can set the size
 	// of the export. This is then used to set the Content-Length
 	// in the response correctly.
 	//
@@ -918,7 +1008,7 @@ func (se *SnapshotExport) Init() error {
 	// but a known issue with this approach here.
 	var sz osutil.Sizer
 	if err := se.StreamTo(&sz); err != nil {
-		return fmt.Errorf("cannot calculcate the size for %v: %s", se.setID, err)
+		return fmt.Errorf("cannot calculate the size for %v: %s", se.setID, err)
 	}
 	se.size = sz.Size()
 	return nil
@@ -991,6 +1081,16 @@ func (se *SnapshotExport) StreamTo(w io.Writer) error {
 		files = append(files, path.Base(snapshotFile.Name()))
 	}
 
+	// SnapshotExporter has se.Close() set as a finalizer, thus when the object
+	// is no longer referenced, se.Close() (which closes all files) will be
+	// called automatically after/during a GC pass. We don't know if the caller
+	// retains a reference to the object (eg. for any outstanding calls to some
+	// of its functions), and the last explicit reference in the code above was
+	// kept for the purpose of accessing the snapshot files list, which is done
+	// before the final file is read, so we need to mark object as alive until
+	// after every file has been read.
+	runtime.KeepAlive(se)
+
 	// write the metadata last, then the client can use that to
 	// validate the archive is complete
 	meta := exportMetadata{
