@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2017-2022 Canonical Ltd
+ * Copyright (C) 2017-2024 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -21,6 +21,7 @@ package snapstate_test
 
 import (
 	"context"
+	"encoding/json"
 	"errors"
 	"fmt"
 	"os"
@@ -35,9 +36,11 @@ import (
 	"github.com/snapcore/snapd/asserts/snapasserts"
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/httputil"
+	"github.com/snapcore/snapd/interfaces"
 	"github.com/snapcore/snapd/logger"
 	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/overlord/configstate/config"
+	"github.com/snapcore/snapd/overlord/ifacestate/ifacerepo"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/snapstate/snapstatetest"
 	"github.com/snapcore/snapd/overlord/state"
@@ -73,7 +76,7 @@ func (r *autoRefreshStore) SnapAction(ctx context.Context, currentSnaps []*store
 	if assertQuery != nil {
 		panic("no assertion query support")
 	}
-	if !opts.IsAutoRefresh {
+	if !opts.Scheduled {
 		panic("AutoRefresh snap action did not set IsAutoRefresh flag")
 	}
 
@@ -122,11 +125,14 @@ func (s *autoRefreshTestSuite) SetUpTest(c *C) {
 	defer s.state.Unlock()
 	snapstate.ReplaceStore(s.state, s.store)
 
+	repo := interfaces.NewRepository()
+	ifacerepo.Replace(s.state, repo)
+
 	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
 		Active: true,
-		Sequence: []*snap.SideInfo{
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
 			{RealName: "some-snap", Revision: snap.R(5), SnapID: "some-snap-id"},
-		},
+		}),
 		Current:  snap.R(5),
 		SnapType: "app",
 		UserID:   1,
@@ -146,7 +152,7 @@ func (s *autoRefreshTestSuite) SetUpTest(c *C) {
 	s.AddCleanup(snapstatetest.MockDeviceModel(DefaultModel()))
 
 	restore := snapstate.MockEnforcedValidationSets(func(st *state.State, extraVss ...*asserts.ValidationSet) (*snapasserts.ValidationSets, error) {
-		return nil, nil
+		return snapasserts.NewValidationSets(), nil
 	})
 	s.AddCleanup(restore)
 }
@@ -167,11 +173,6 @@ func (s *autoRefreshTestSuite) TestLastRefresh(c *C) {
 }
 
 func (s *autoRefreshTestSuite) TestLastRefreshRefreshManaged(c *C) {
-	snapstate.CanManageRefreshes = func(st *state.State) bool {
-		return true
-	}
-	defer func() { snapstate.CanManageRefreshes = nil }()
-
 	logbuf, restore := logger.MockLogger()
 	defer restore()
 
@@ -214,11 +215,6 @@ func (s *autoRefreshTestSuite) TestLastRefreshRefreshManaged(c *C) {
 }
 
 func (s *autoRefreshTestSuite) TestRefreshManagedTimerWins(c *C) {
-	snapstate.CanManageRefreshes = func(st *state.State) bool {
-		return true
-	}
-	defer func() { snapstate.CanManageRefreshes = nil }()
-
 	s.state.Lock()
 	defer s.state.Unlock()
 
@@ -242,18 +238,7 @@ func (s *autoRefreshTestSuite) TestRefreshManagedTimerWins(c *C) {
 	c.Check(err, IsNil)
 }
 
-func (s *autoRefreshTestSuite) TestRefreshManagedDenied(c *C) {
-	canManageCalled := false
-	snapstate.CanManageRefreshes = func(st *state.State) bool {
-		canManageCalled = true
-		// always deny
-		return false
-	}
-	defer func() { snapstate.CanManageRefreshes = nil }()
-
-	logbuf, restore := logger.MockLogger()
-	defer restore()
-
+func (s *autoRefreshTestSuite) TestRefreshManagedIsRespected(c *C) {
 	s.state.Lock()
 	defer s.state.Unlock()
 
@@ -269,23 +254,11 @@ func (s *autoRefreshTestSuite) TestRefreshManagedDenied(c *C) {
 			err := af.Ensure()
 			s.state.Lock()
 			c.Check(err, IsNil)
-			c.Check(s.store.ops, DeepEquals, []string{"list-refresh"})
-
-			refreshScheduleStr, _, err := af.RefreshSchedule()
-			c.Check(refreshScheduleStr, Equals, snapstate.DefaultRefreshSchedule)
-			c.Check(err, IsNil)
-			c.Check(canManageCalled, Equals, true)
-			count := strings.Count(logbuf.String(),
-				": managed refresh schedule denied, no properly configured snapd-control\n")
-			c.Check(count, Equals, 1, Commentf("too many occurrences:\n%s", logbuf.String()))
-
-			canManageCalled = false
+			c.Check(s.store.ops, HasLen, 0)
 		}
 
 		// ensure clean config for the next run
 		s.state.Set("config", nil)
-		logbuf.Reset()
-		canManageCalled = false
 	}
 }
 
@@ -843,6 +816,54 @@ func (s *autoRefreshTestSuite) TestAtSeedPolicy(c *C) {
 	c.Check(t1.Equal(t2), Equals, true)
 }
 
+func (s *autoRefreshTestSuite) TestAtSeedRefreshHeld(c *C) {
+	// it is possible that refresh.hold has already been set to a valid time
+	// or "forever" even before snapd was started for the first time
+	r := release.MockOnClassic(true)
+	defer r()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	tr := config.NewTransaction(s.state)
+	c.Assert(tr.Set("core", "refresh.hold", "forever"), IsNil)
+	tr.Commit()
+
+	af := snapstate.NewAutoRefresh(s.state)
+	err := af.AtSeed()
+	c.Assert(err, IsNil)
+	c.Check(af.NextRefresh().IsZero(), Equals, true)
+
+	// now use a valid timestamp
+	tr = config.NewTransaction(s.state)
+	c.Assert(tr.Set("core", "refresh.hold", time.Now().UTC()), IsNil)
+	tr.Commit()
+
+	af = snapstate.NewAutoRefresh(s.state)
+	err = af.AtSeed()
+	c.Assert(err, IsNil)
+	c.Check(af.NextRefresh().IsZero(), Equals, true)
+}
+
+func (s *autoRefreshTestSuite) TestAtSeedInvalidHold(c *C) {
+	// it is possible that refresh.hold has already been set to forever even
+	// before snapd was started for the first time
+	r := release.MockOnClassic(true)
+	defer r()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	tr := config.NewTransaction(s.state)
+	c.Assert(tr.Set("core", "refresh.hold", "this-is-invalid-time"), IsNil)
+	tr.Commit()
+
+	af := snapstate.NewAutoRefresh(s.state)
+
+	err := af.AtSeed()
+	c.Assert(err, ErrorMatches, `parsing time "this-is-invalid-time" .*cannot parse.*`)
+}
+
 func (s *autoRefreshTestSuite) TestCanRefreshOnMetered(c *C) {
 	s.state.Lock()
 	defer s.state.Unlock()
@@ -936,7 +957,7 @@ func (s *autoRefreshTestSuite) TestInitialInhibitRefreshWithinInhibitWindow(c *C
 	s.state.Lock()
 	defer s.state.Unlock()
 
-	restore := snapstate.MockAsyncPendingRefreshNotification(func(ctx context.Context, client *userclient.Client, refreshInfo *userclient.PendingSnapRefreshInfo) {
+	restore := snapstate.MockAsyncPendingRefreshNotification(func(ctx context.Context, refreshInfo *userclient.PendingSnapRefreshInfo) {
 		c.Fatal("shouldn't trigger pending refresh notification unless it was an auto-refresh and we're overdue")
 	})
 	defer restore()
@@ -944,7 +965,7 @@ func (s *autoRefreshTestSuite) TestInitialInhibitRefreshWithinInhibitWindow(c *C
 	si := &snap.SideInfo{RealName: "pkg", Revision: snap.R(1)}
 	info := &snap.Info{SideInfo: *si}
 	snapst := &snapstate.SnapState{
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	}
 	snapsup := &snapstate.SnapSetup{Flags: snapstate.Flags{IsAutoRefresh: true}}
@@ -954,8 +975,9 @@ func (s *autoRefreshTestSuite) TestInitialInhibitRefreshWithinInhibitWindow(c *C
 	})
 	defer restore()
 
-	err := snapstate.InhibitRefresh(s.state, snapst, snapsup, info)
+	inhibitionTimeout, err := snapstate.InhibitRefresh(s.state, snapst, snapsup, info)
 	c.Assert(err, ErrorMatches, `snap "pkg" has running apps or hooks, pids: 123`)
+	c.Check(inhibitionTimeout, Equals, false)
 
 	var timedErr *snapstate.TimedBusySnapError
 	c.Assert(errors.As(err, &timedErr), Equals, true)
@@ -970,18 +992,18 @@ func (s *autoRefreshTestSuite) TestSubsequentInhibitRefreshWithinInhibitWindow(c
 	s.state.Lock()
 	defer s.state.Unlock()
 
-	restore := snapstate.MockAsyncPendingRefreshNotification(func(ctx context.Context, client *userclient.Client, refreshInfo *userclient.PendingSnapRefreshInfo) {
+	restore := snapstate.MockAsyncPendingRefreshNotification(func(ctx context.Context, refreshInfo *userclient.PendingSnapRefreshInfo) {
 		c.Fatal("shouldn't trigger pending refresh notification unless it was an auto-refresh and we're overdue")
 	})
 	defer restore()
 
 	instant := time.Now()
-	pastInstant := instant.Add(-snapstate.MaxInhibition / 2) // In the middle of the allowed window
+	pastInstant := instant.Add(-snapstate.MaxInhibitionDuration(s.state) / 2) // In the middle of the allowed window
 
 	si := &snap.SideInfo{RealName: "pkg", Revision: snap.R(1)}
 	info := &snap.Info{SideInfo: *si}
 	snapst := &snapstate.SnapState{
-		Sequence:             []*snap.SideInfo{si},
+		Sequence:             snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:              si.Revision,
 		RefreshInhibitedTime: &pastInstant,
 	}
@@ -992,8 +1014,9 @@ func (s *autoRefreshTestSuite) TestSubsequentInhibitRefreshWithinInhibitWindow(c
 	})
 	defer restore()
 
-	err := snapstate.InhibitRefresh(s.state, snapst, snapsup, info)
+	inhibitionTimeout, err := snapstate.InhibitRefresh(s.state, snapst, snapsup, info)
 	c.Assert(err, ErrorMatches, `snap "pkg" has running apps or hooks, pids: 123`)
+	c.Check(inhibitionTimeout, Equals, false)
 
 	var timedErr *snapstate.TimedBusySnapError
 	c.Assert(errors.As(err, &timedErr), Equals, true)
@@ -1011,7 +1034,7 @@ func (s *autoRefreshTestSuite) TestInhibitRefreshRefreshesWhenOverdue(c *C) {
 	defer s.state.Unlock()
 
 	notificationCount := 0
-	restore := snapstate.MockAsyncPendingRefreshNotification(func(ctx context.Context, client *userclient.Client, refreshInfo *userclient.PendingSnapRefreshInfo) {
+	restore := snapstate.MockAsyncPendingRefreshNotification(func(ctx context.Context, refreshInfo *userclient.PendingSnapRefreshInfo) {
 		notificationCount++
 		c.Check(refreshInfo.InstanceName, Equals, "pkg")
 		c.Check(refreshInfo.TimeRemaining, Equals, time.Duration(0))
@@ -1019,12 +1042,12 @@ func (s *autoRefreshTestSuite) TestInhibitRefreshRefreshesWhenOverdue(c *C) {
 	defer restore()
 
 	instant := time.Now()
-	pastInstant := instant.Add(-snapstate.MaxInhibition * 2)
+	pastInstant := instant.Add(-snapstate.MaxInhibitionDuration(s.state) * 2)
 
 	si := &snap.SideInfo{RealName: "pkg", Revision: snap.R(1)}
 	info := &snap.Info{SideInfo: *si}
 	snapst := &snapstate.SnapState{
-		Sequence:             []*snap.SideInfo{si},
+		Sequence:             snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:              si.Revision,
 		RefreshInhibitedTime: &pastInstant,
 	}
@@ -1035,8 +1058,9 @@ func (s *autoRefreshTestSuite) TestInhibitRefreshRefreshesWhenOverdue(c *C) {
 	})
 	defer restore()
 
-	err := snapstate.InhibitRefresh(s.state, snapst, snapsup, info)
+	inhibitionTimeout, err := snapstate.InhibitRefresh(s.state, snapst, snapsup, info)
 	c.Assert(err == nil, Equals, true)
+	c.Check(inhibitionTimeout, Equals, true)
 	c.Check(notificationCount, Equals, 1)
 }
 
@@ -1044,17 +1068,17 @@ func (s *autoRefreshTestSuite) TestInhibitNoNotificationOnManualRefresh(c *C) {
 	s.state.Lock()
 	defer s.state.Unlock()
 
-	restore := snapstate.MockAsyncPendingRefreshNotification(func(ctx context.Context, client *userclient.Client, refreshInfo *userclient.PendingSnapRefreshInfo) {
+	restore := snapstate.MockAsyncPendingRefreshNotification(func(ctx context.Context, refreshInfo *userclient.PendingSnapRefreshInfo) {
 		c.Fatal("shouldn't trigger pending refresh notification if refresh was manual")
 	})
 	defer restore()
 
-	pastInstant := time.Now().Add(-snapstate.MaxInhibition)
+	pastInstant := time.Now().Add(-snapstate.MaxInhibitionDuration(s.state))
 
 	si := &snap.SideInfo{RealName: "pkg", Revision: snap.R(1)}
 	info := &snap.Info{SideInfo: *si}
 	snapst := &snapstate.SnapState{
-		Sequence:             []*snap.SideInfo{si},
+		Sequence:             snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:              si.Revision,
 		RefreshInhibitedTime: &pastInstant,
 	}
@@ -1066,8 +1090,9 @@ func (s *autoRefreshTestSuite) TestInhibitNoNotificationOnManualRefresh(c *C) {
 	})
 	defer restore()
 
-	err := snapstate.InhibitRefresh(s.state, snapst, snapsup, info)
+	inhibitionTimeout, err := snapstate.InhibitRefresh(s.state, snapst, snapsup, info)
 	c.Assert(err, testutil.ErrorIs, &snapstate.BusySnapError{})
+	c.Check(inhibitionTimeout, Equals, false)
 }
 
 func (s *autoRefreshTestSuite) TestBlockedAutoRefreshCreatesPreDownloads(c *C) {
@@ -1151,7 +1176,7 @@ func (s *autoRefreshTestSuite) addRefreshableSnap(names ...string) {
 		si := &snap.SideInfo{RealName: name, SnapID: fmt.Sprintf("%s-id", name), Revision: snap.R(1)}
 		snapstate.Set(s.state, name, &snapstate.SnapState{
 			Active:   true,
-			Sequence: []*snap.SideInfo{si},
+			Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 			Current:  si.Revision,
 			SnapType: string(snap.TypeApp),
 		})
@@ -1166,69 +1191,457 @@ func (s *autoRefreshTestSuite) addRefreshableSnap(names ...string) {
 	}
 }
 
-func (s *autoRefreshTestSuite) TestPartiallyInhibitedAutoRefreshIsContinued(c *C) {
-	s.addRefreshableSnap("foo")
-
-	now := time.Now()
-	restore := snapstate.MockTimeNow(func() time.Time {
-		return now
-	})
-	defer restore()
-
-	af := snapstate.NewAutoRefresh(s.state)
-	s.state.Lock()
-	s.state.Cache("auto-refresh-continue-attempt", 1)
-	s.state.Unlock()
-
-	err := af.Ensure()
-	c.Check(err, IsNil)
-	c.Check(s.store.ops, DeepEquals, []string{"list-refresh"})
-
-	s.state.Lock()
-	defer s.state.Unlock()
-	var lastRefresh time.Time
-	err = s.state.Get("last-refresh", &lastRefresh)
-	c.Assert(err, IsNil)
-	c.Check(lastRefresh.Equal(now), Equals, true)
-	c.Check(s.state.Cached("auto-refresh-continue-attempt"), IsNil)
-
-	// check that the "IsContinuedAutoRefresh" flag is set for a
-	// continued auto-refresh
-	chgs := s.state.Changes()
-	c.Assert(chgs, HasLen, 1)
-	checkIsContinuedAutoRefresh(c, chgs[0].Tasks(), true)
-}
-
-func (s *autoRefreshTestSuite) TestContinueAutorefreshOnlyFirstOverridesDelay(c *C) {
-	// fail the auto-refresh
-	s.store.err = fmt.Errorf("random store error")
-
-	af := snapstate.NewAutoRefresh(s.state)
-	// run it once so we're in the retry delay
-	err := af.Ensure()
-	c.Assert(err, ErrorMatches, "random store error")
-
-	s.state.Lock()
-	s.state.Cache("auto-refresh-continue-attempt", 1)
-	s.state.Unlock()
-
-	err = af.Ensure()
-	c.Check(err, ErrorMatches, "random store error")
-	c.Check(s.store.ops, DeepEquals, []string{"list-refresh", "list-refresh"})
-	s.state.Lock()
-	c.Check(s.state.Cached("auto-refresh-continue-attempt"), Equals, 2)
-	s.state.Unlock()
-
-	err = af.Ensure()
-	c.Check(err, IsNil)
-	c.Check(s.store.ops, DeepEquals, []string{"list-refresh", "list-refresh"})
-	s.state.Lock()
-	c.Check(s.state.Cached("auto-refresh-continue-attempt"), Equals, 2)
-	s.state.Unlock()
-}
-
 func (s *autoRefreshTestSuite) TestTooSoonError(c *C) {
 	c.Check(snapstate.TooSoonError{}, testutil.ErrorIs, snapstate.TooSoonError{})
 	c.Check(snapstate.TooSoonError{}, Not(testutil.ErrorIs), errors.New(""))
 	c.Check(snapstate.TooSoonError{}.Error(), Equals, "cannot auto-refresh so soon")
 }
+
+func setStoreAccess(s *state.State, access interface{}) {
+	s.Lock()
+	defer s.Unlock()
+
+	tr := config.NewTransaction(s)
+	tr.Set("core", "store.access", access)
+	tr.Commit()
+}
+
+func (s *autoRefreshTestSuite) TestSnapStoreOffline(c *C) {
+	s.addRefreshableSnap("foo")
+
+	setStoreAccess(s.state, "offline")
+
+	af := snapstate.NewAutoRefresh(s.state)
+	err := af.Ensure()
+	c.Check(err, IsNil)
+
+	s.state.Lock()
+	c.Check(s.state.Changes(), HasLen, 0)
+	s.state.Unlock()
+
+	setStoreAccess(s.state, nil)
+
+	err = af.Ensure()
+	c.Check(err, IsNil)
+
+	c.Check(s.store.ops, DeepEquals, []string{"list-refresh"})
+	s.state.Lock()
+	c.Check(s.state.Changes(), HasLen, 1)
+	s.state.Unlock()
+}
+
+func (s *autoRefreshTestSuite) testMaybeAddRefreshInhibitNotice(c *C, markerInterfaceConnected bool, warningFallback bool) {
+	st := s.state
+	st.Lock()
+	defer st.Unlock()
+
+	var connCheckCalled int
+	restore := snapstate.MockHasActiveConnection(func(st *state.State, iface string) (bool, error) {
+		connCheckCalled++
+		c.Check(iface, Equals, "snap-refresh-observe")
+		return markerInterfaceConnected, nil
+	})
+	defer restore()
+
+	// let's add some random warnings
+	st.Warnf("this is a random warning 1")
+	st.Warnf("this is a random warning 2")
+
+	err := snapstate.MaybeAddRefreshInhibitNotice(st)
+	c.Assert(err, IsNil)
+	// empty set of inhibited snaps unchanged -> []
+	// no notice recorded
+	c.Assert(st.Notices(nil), HasLen, 0)
+	// no "refresh inhibition" warnings recorded
+	checkNoRefreshInhibitWarning(c, st)
+	// Verify list is empty
+	checkLastRecordedInhibitedSnaps(c, st, nil)
+
+	now := time.Now()
+	warningTime := now
+	// mock time to determine if recorded warning is recent
+	defer state.MockTime(warningTime)()
+	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "some-snap", SnapID: "some-snap-id", Revision: snap.R(1)},
+		}),
+		Current:              snap.R(1),
+		SnapType:             string(snap.TypeApp),
+		RefreshInhibitedTime: &now,
+	})
+	err = snapstate.MaybeAddRefreshInhibitNotice(st)
+	c.Assert(err, IsNil)
+	// set of inhibited snaps changed -> ["some-snap"]
+	// notice recorded
+	expectedOccurrances := 1
+	checkRefreshInhibitNotice(c, st, expectedOccurrances)
+	// check warnings fallback
+	if warningFallback {
+		checkRefreshInhibitWarning(c, st, []string{"some-snap"}, warningTime)
+	} else {
+		checkNoRefreshInhibitWarning(c, st)
+	}
+	// check that the set of last recorded inhibited snaps is persisted
+	checkLastRecordedInhibitedSnaps(c, st, []string{"some-snap"})
+
+	// mock time to determine if recorded warning is recent
+	warningTime = warningTime.Add(1 * time.Hour)
+	defer state.MockTime(warningTime)()
+	err = snapstate.MaybeAddRefreshInhibitNotice(st)
+	c.Assert(err, IsNil)
+	// set of inhibited snaps unchanged -> ["some-snap"]
+	// no new notice recorded
+	checkRefreshInhibitNotice(c, st, expectedOccurrances)
+	// check warnings fallback
+	if warningFallback {
+		checkRefreshInhibitWarning(c, st, []string{"some-snap"}, warningTime)
+	} else {
+		checkNoRefreshInhibitWarning(c, st)
+	}
+	checkLastRecordedInhibitedSnaps(c, st, []string{"some-snap"})
+
+	// mark "some-snap" as not inhibited
+	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "some-snap", SnapID: "some-snap-id", Revision: snap.R(1)},
+		}),
+		Current:              snap.R(1),
+		SnapType:             string(snap.TypeApp),
+		RefreshInhibitedTime: nil,
+	})
+	// mark "some-other-snap" as inhibited
+	snapstate.Set(s.state, "some-other-snap", &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "some-other-snap", SnapID: "some-other-snap-id", Revision: snap.R(1)},
+		}),
+		Current:              snap.R(1),
+		SnapType:             string(snap.TypeApp),
+		RefreshInhibitedTime: &now,
+	})
+	// mock time to determine if recorded warning is recent
+	warningTime = warningTime.Add(1 * time.Hour)
+	defer state.MockTime(warningTime)()
+	err = snapstate.MaybeAddRefreshInhibitNotice(st)
+	c.Assert(err, IsNil)
+	// set of inhibited snaps changed -> ["some-other-snap"]
+	// notice recorded
+	expectedOccurrances++
+	checkRefreshInhibitNotice(c, st, expectedOccurrances)
+	// check warnings fallback
+	if warningFallback {
+		checkRefreshInhibitWarning(c, st, []string{"some-other-snap"}, warningTime)
+	} else {
+		checkNoRefreshInhibitWarning(c, st)
+	}
+	checkLastRecordedInhibitedSnaps(c, st, []string{"some-other-snap"})
+
+	// mark "some-other-snap" as not inhibited
+	snapstate.Set(s.state, "some-other-snap", &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "some-other-snap", SnapID: "some-other-snap-id", Revision: snap.R(1)},
+		}),
+		Current:              snap.R(1),
+		SnapType:             string(snap.TypeApp),
+		RefreshInhibitedTime: nil,
+	})
+	// mock time to determine if recorded warning is recent
+	warningTime = warningTime.Add(1 * time.Hour)
+	defer state.MockTime(warningTime)()
+	err = snapstate.MaybeAddRefreshInhibitNotice(st)
+	c.Assert(err, IsNil)
+	// set of inhibited snaps changed -> []
+	// notice recorded
+	expectedOccurrances++
+	checkRefreshInhibitNotice(c, st, expectedOccurrances)
+	// no warning should be recorded and existing warning should be
+	// removed if inhibited snaps set is empty
+	checkNoRefreshInhibitWarning(c, st)
+	checkLastRecordedInhibitedSnaps(c, st, []string{})
+
+	// exercise multiple snaps inhibited
+	// mark "some-snap" as inhibited
+	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "some-snap", SnapID: "some-snap-id", Revision: snap.R(1)},
+		}),
+		Current:              snap.R(1),
+		SnapType:             string(snap.TypeApp),
+		RefreshInhibitedTime: &now,
+	})
+	// mark "some-other-snap" as inhibited
+	snapstate.Set(s.state, "some-other-snap", &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "some-other-snap", SnapID: "some-other-snap-id", Revision: snap.R(1)},
+		}),
+		Current:              snap.R(1),
+		SnapType:             string(snap.TypeApp),
+		RefreshInhibitedTime: &now,
+	})
+	// mock time to determine if recorded warning is recent
+	warningTime = warningTime.Add(1 * time.Hour)
+	defer state.MockTime(warningTime)()
+	err = snapstate.MaybeAddRefreshInhibitNotice(st)
+	c.Assert(err, IsNil)
+	// set of inhibited snaps changed -> ["some-snap", "some-other-snap"]
+	// notice recorded
+	expectedOccurrances++
+	checkRefreshInhibitNotice(c, st, expectedOccurrances)
+	// check warnings fallback
+	if warningFallback {
+		checkRefreshInhibitWarning(c, st, []string{"some-snap", "some-other-snap"}, warningTime)
+	} else {
+		checkNoRefreshInhibitWarning(c, st)
+	}
+	// check that the set of last recorded inhibited snaps is persisted
+	checkLastRecordedInhibitedSnaps(c, st, []string{"some-snap", "some-other-snap"})
+}
+
+func (s *autoRefreshTestSuite) TestMaybeAddRefreshInhibitNotice(c *C) {
+	s.enableRefreshAppAwarenessUX()
+	const markerInterfaceConnected = true
+	const warningFallback = false
+	s.testMaybeAddRefreshInhibitNotice(c, markerInterfaceConnected, warningFallback)
+}
+
+func (s *autoRefreshTestSuite) TestMaybeAddRefreshInhibitNoticeWarningFallbackError(c *C) {
+	st := s.state
+	st.Lock()
+	defer st.Unlock()
+
+	logbuf, restore := logger.MockLogger()
+	defer restore()
+
+	// mark "some-snap" as inhibited
+	now := time.Now()
+	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
+		Active: true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{
+			{RealName: "some-snap", SnapID: "some-snap-id", Revision: snap.R(1)},
+		}),
+		Current:              snap.R(1),
+		SnapType:             string(snap.TypeApp),
+		RefreshInhibitedTime: &now,
+	})
+
+	// Highly unlikely but just in case
+	tr := config.NewTransaction(s.state)
+	tr.Set("core", "experimental.refresh-app-awareness-ux", "trigger-error")
+	tr.Commit()
+
+	err := snapstate.MaybeAddRefreshInhibitNotice(st)
+	// warning fallback error is not propagated, only logged
+	c.Assert(err, IsNil)
+	// check error is logged
+	c.Check(logbuf.String(), testutil.Contains, `Cannot add refresh inhibition warning: refresh-app-awareness-ux can only be set to 'true' or 'false', got "trigger-error"`)
+	// notice recorded
+	checkRefreshInhibitNotice(c, st, 1)
+	// no warnings recorded due to error
+	checkNoRefreshInhibitWarning(c, st)
+
+	restore = snapstate.MockHasActiveConnection(func(st *state.State, iface string) (bool, error) {
+		return false, fmt.Errorf("boom")
+	})
+	defer restore()
+
+	st.Unlock()
+	s.enableRefreshAppAwarenessUX()
+	st.Lock()
+	err = snapstate.MaybeAddRefreshInhibitNotice(st)
+	// warning fallback error is not propagated, only logged
+	c.Assert(err, IsNil)
+	// check error is logged
+	c.Check(logbuf.String(), testutil.Contains, "Cannot add refresh inhibition warning: boom")
+	// set of inhibited snaps unchanged -> ["some-snap"]
+	// no new notice recorded
+	checkRefreshInhibitNotice(c, st, 1)
+	// no warnings recorded due to error
+	checkNoRefreshInhibitWarning(c, st)
+}
+
+func (s *autoRefreshTestSuite) TestMaybeAddRefreshInhibitNoticeWarningFallback(c *C) {
+	s.enableRefreshAppAwarenessUX()
+	const markerInterfaceConnected = false
+	const warningFallback = true
+	s.testMaybeAddRefreshInhibitNotice(c, markerInterfaceConnected, warningFallback)
+}
+
+func (s *autoRefreshTestSuite) TestMaybeAddRefreshInhibitNoticeWarningFallbackNoRAAUX(c *C) {
+	const markerInterfaceConnected = false
+	const warningFallback = false // because refresh-app-awareness-ux is disabled
+	s.testMaybeAddRefreshInhibitNotice(c, markerInterfaceConnected, warningFallback)
+}
+
+func (s *autoRefreshTestSuite) enableRefreshAppAwarenessUX() {
+	s.state.Lock()
+	tr := config.NewTransaction(s.state)
+	tr.Set("core", "experimental.refresh-app-awareness-ux", true)
+	tr.Commit()
+	s.state.Unlock()
+}
+
+func checkNoRefreshInhibitWarning(c *C, st *state.State) {
+	for _, warning := range st.AllWarnings() {
+		if strings.HasSuffix(warning.String(), "close running apps to continue refresh.") {
+			c.Error("inhibition warning found")
+			return
+		}
+	}
+}
+
+func checkRefreshInhibitWarning(c *C, st *state.State, snaps []string, warningTime time.Time) {
+	var inhibitionWarning *state.Warning
+	for _, warning := range st.AllWarnings() {
+		if !strings.HasSuffix(warning.String(), "close running apps to continue refresh.") {
+			continue
+		}
+		if inhibitionWarning != nil {
+			c.Errorf("found multiple inhibition warnings")
+			return
+		}
+		inhibitionWarning = warning
+	}
+
+	// There is always one warning
+	c.Assert(inhibitionWarning, NotNil)
+	w := warningToMap(c, inhibitionWarning)
+	c.Check(w["message"], Matches, "cannot refresh (.*) due running apps; close running apps to continue refresh.")
+	for _, snap := range snaps {
+		c.Check(w["message"], testutil.Contains, snap)
+	}
+	c.Check(w["repeat-after"], Equals, "24h0m0s")
+	if !warningTime.IsZero() {
+		c.Check(w["last-added"], Equals, warningTime.UTC().Format(time.RFC3339Nano))
+	}
+}
+
+// warningToMap converts a Warning to a map using a JSON marshal-unmarshal round trip.
+func warningToMap(c *C, warning *state.Warning) map[string]any {
+	buf, err := json.Marshal(warning)
+	c.Assert(err, IsNil)
+	var n map[string]any
+	err = json.Unmarshal(buf, &n)
+	c.Assert(err, IsNil)
+	return n
+}
+
+func (s *autoRefreshTestSuite) TestMaybeAsyncPendingRefreshNotification(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	var connCheckCalled int
+	restore := snapstate.MockHasActiveConnection(func(st *state.State, iface string) (bool, error) {
+		connCheckCalled++
+		c.Check(iface, Equals, "snap-refresh-observe")
+		// no snap has the marker interface connected
+		return false, nil
+	})
+	defer restore()
+
+	expectedInfo := &userclient.PendingSnapRefreshInfo{
+		InstanceName:  "pkg",
+		TimeRemaining: 10 * time.Second,
+	}
+	var notificationCalled int
+	restore = snapstate.MockAsyncPendingRefreshNotification(func(ctx context.Context, psri *userclient.PendingSnapRefreshInfo) {
+		notificationCalled++
+		c.Check(psri, Equals, expectedInfo)
+	})
+	defer restore()
+
+	tr := config.NewTransaction(s.state)
+	tr.Set("core", "experimental.refresh-app-awareness-ux", true)
+	tr.Commit()
+
+	snapstate.MaybeAsyncPendingRefreshNotification(context.TODO(), s.state, expectedInfo)
+	// no notification as refresh-appawareness-ux is enabled
+	// i.e. notices + warnings fallback is used instead
+	c.Check(connCheckCalled, Equals, 0)
+	c.Check(notificationCalled, Equals, 0)
+
+	tr.Set("core", "experimental.refresh-app-awareness-ux", false)
+	tr.Commit()
+
+	snapstate.MaybeAsyncPendingRefreshNotification(context.TODO(), s.state, expectedInfo)
+	// notification sent as refresh-appawareness-ux is now disabled
+	c.Check(connCheckCalled, Equals, 1)
+	c.Check(notificationCalled, Equals, 1)
+}
+
+func (s *autoRefreshTestSuite) TestMaybeAsyncPendingRefreshNotificationSkips(c *C) {
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	var connCheckCalled int
+	restore := snapstate.MockHasActiveConnection(func(st *state.State, iface string) (bool, error) {
+		connCheckCalled++
+		c.Check(iface, Equals, "snap-refresh-observe")
+		// marker interface found
+		return true, nil
+	})
+	defer restore()
+
+	restore = snapstate.MockAsyncPendingRefreshNotification(func(ctx context.Context, psri *userclient.PendingSnapRefreshInfo) {
+		c.Fatal("shouldn't trigger pending refresh notification because marker interface is connected")
+	})
+	defer restore()
+
+	snapstate.MaybeAsyncPendingRefreshNotification(context.TODO(), s.state, &userclient.PendingSnapRefreshInfo{})
+}
+
+func (s *autoRefreshTestSuite) TestAutoRefreshWithConfdbs(c *C) {
+	si := &snap.SideInfo{
+		RealName: "foo",
+		SnapID:   "foo-id",
+		Revision: snap.R(1),
+	}
+	s.state.Lock()
+	snapstate.Set(s.state, "foo", &snapstate.SnapState{
+		Active:   true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
+		Current:  si.Revision,
+		SnapType: string(snap.TypeApp),
+	})
+	s.state.Unlock()
+
+	info := &snap.Info{
+		SnapType:      snap.TypeApp,
+		Architectures: []string{"all"},
+		SideInfo: snap.SideInfo{
+			RealName: "foo",
+			Revision: snap.R(8),
+		},
+	}
+	info.Plugs = map[string]*snap.PlugInfo{
+		"my-plug": {
+			Snap:      info,
+			Name:      "my-plug",
+			Interface: "confdb",
+			Attrs: map[string]interface{}{
+				"account": "my-publisher",
+				"view":    "my-reg/my-view",
+			},
+		},
+	}
+	s.store.refreshable = append(s.store.refreshable, info)
+
+	af := snapstate.NewAutoRefresh(s.state)
+	err := af.Ensure()
+	c.Check(err, IsNil)
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	chgs := s.state.Changes()
+	c.Assert(chgs, HasLen, 1)
+	task := findLastTask(chgs[0], "validate-snap")
+
+	snapsup, err := snapstate.TaskSnapSetup(task)
+	c.Assert(err, IsNil)
+	c.Assert(snapsup.Confdbs, DeepEquals, []snapstate.ConfdbID{{Account: "my-publisher", Confdb: "my-reg"}})
+}
