@@ -20,9 +20,9 @@
 package snapstate_test
 
 import (
+	"context"
 	"errors"
 	"fmt"
-	"os/user"
 
 	. "gopkg.in/check.v1"
 
@@ -30,6 +30,7 @@ import (
 	"github.com/snapcore/snapd/asserts"
 	"github.com/snapcore/snapd/dirs"
 	"github.com/snapcore/snapd/osutil"
+	"github.com/snapcore/snapd/osutil/user"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/snapstate/snapstatetest"
 	"github.com/snapcore/snapd/overlord/state"
@@ -201,6 +202,8 @@ var assumesTests = []struct {
 	assumes: "[command-chain]",
 }, {
 	assumes: "[kernel-assets]",
+}, {
+	assumes: "[snap-uid-envvars]",
 },
 }
 
@@ -318,7 +321,7 @@ version: 1
 	snapstate.Set(st, "gadget", &snapstate.SnapState{
 		SnapType: "gadget",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -360,7 +363,7 @@ version: 1
 	snapstate.Set(st, "gadget", &snapstate.SnapState{
 		SnapType: "gadget",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -402,7 +405,7 @@ version: 1
 	snapstate.Set(st, "gadget", &snapstate.SnapState{
 		SnapType: "gadget",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -435,7 +438,7 @@ func (s *checkSnapSuite) setupKernelGadgetSnaps(st *state.State) {
 	snapstate.Set(st, "gadget", &snapstate.SnapState{
 		SnapType: string(snap.TypeGadget),
 		Active:   true,
-		Sequence: []*snap.SideInfo{gadgetInfo},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{gadgetInfo}),
 		Current:  gadgetInfo.Revision,
 	})
 
@@ -447,7 +450,7 @@ func (s *checkSnapSuite) setupKernelGadgetSnaps(st *state.State) {
 	snapstate.Set(st, "kernel", &snapstate.SnapState{
 		SnapType: string(snap.TypeKernel),
 		Active:   true,
-		Sequence: []*snap.SideInfo{kernelInfo},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{kernelInfo}),
 		Current:  kernelInfo.Revision,
 	})
 }
@@ -556,7 +559,7 @@ version: 1
 	snapstate.Set(st, "gadget", &snapstate.SnapState{
 		SnapType: "gadget",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -597,7 +600,7 @@ version: 1
 	snapstate.Set(st, "gadget", &snapstate.SnapState{
 		SnapType: "gadget",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -757,7 +760,7 @@ version: 1
 	snapstate.Set(st, "kernel", &snapstate.SnapState{
 		SnapType: "kernel",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -799,7 +802,7 @@ version: 1
 	snapstate.Set(st, "kernel", &snapstate.SnapState{
 		SnapType: "kernel",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -843,7 +846,7 @@ version: 1
 	snapstate.Set(st, "other-kernel", &snapstate.SnapState{
 		SnapType: "kernel",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -931,7 +934,7 @@ version: 1
 	snapstate.Set(st, "some-base", &snapstate.SnapState{
 		SnapType: "base",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -967,7 +970,7 @@ version: 1
 `, si)
 	snapstate.Set(st, "foo", &snapstate.SnapState{
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -1073,7 +1076,7 @@ version: 1
 	snapstate.Set(st, "core", &snapstate.SnapState{
 		SnapType: "os",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -1265,21 +1268,21 @@ func (s *checkSnapSuite) TestCheckSnapSystemUsernames(c *C) {
 		restore = release.MockOnClassic(test.classic)
 		defer restore()
 
-		var osutilEnsureUserGroupCalls int
+		var osutilEnsureSnapUserGroupCalls int
 		if test.noRangeUser {
-			restore = snapstate.MockOsutilEnsureUserGroup(func(name string, id uint32, extraUsers bool) error {
+			restore = snapstate.MockOsutilEnsureSnapUserGroup(func(name string, id uint32, extraUsers bool) error {
 				return fmt.Errorf(`cannot add user/group "%s", group exists and user does not`, name)
 			})
 		} else if test.noUser {
-			restore = snapstate.MockOsutilEnsureUserGroup(func(name string, id uint32, extraUsers bool) error {
+			restore = snapstate.MockOsutilEnsureSnapUserGroup(func(name string, id uint32, extraUsers bool) error {
 				if name == "snapd-range-524288-root" {
 					return nil
 				}
 				return fmt.Errorf(`cannot add user/group "%s", group exists and user does not`, name)
 			})
 		} else {
-			restore = snapstate.MockOsutilEnsureUserGroup(func(name string, id uint32, extraUsers bool) error {
-				osutilEnsureUserGroupCalls++
+			restore = snapstate.MockOsutilEnsureSnapUserGroup(func(name string, id uint32, extraUsers bool) error {
+				osutilEnsureSnapUserGroupCalls++
 				return nil
 			})
 		}
@@ -1299,11 +1302,11 @@ func (s *checkSnapSuite) TestCheckSnapSystemUsernames(c *C) {
 		err = snapstate.CheckSnap(s.st, "snap-path", "foo", nil, nil, snapstate.Flags{}, nil)
 		if test.error != "" {
 			c.Check(err, ErrorMatches, test.error)
-			c.Check(osutilEnsureUserGroupCalls, Equals, 0)
+			c.Check(osutilEnsureSnapUserGroupCalls, Equals, 0)
 		} else {
 			c.Assert(err, IsNil)
 			// one call for the range user, one for the system user
-			c.Check(osutilEnsureUserGroupCalls, Equals, 2)
+			c.Check(osutilEnsureSnapUserGroupCalls, Equals, 2)
 		}
 	}
 }
@@ -1408,7 +1411,7 @@ version: 1
 	snapstate.Set(st, "kernel", &snapstate.SnapState{
 		SnapType: "kernel",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -1459,7 +1462,7 @@ version: 1
 	snapstate.Set(st, "gadget", &snapstate.SnapState{
 		SnapType: "gadget",
 		Active:   true,
-		Sequence: []*snap.SideInfo{si},
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
 		Current:  si.Revision,
 	})
 
@@ -1493,3 +1496,187 @@ version: 2
 	st.Lock()
 	c.Check(err, IsNil)
 }
+
+func (s *checkSnapSuite) TestCheckConfigureHooksHappy(c *C) {
+	var openSnapFile = func(path string, si *snap.SideInfo) (*snap.Info, snap.Container, error) {
+		info := snaptest.MockInfo(c, "{name: snap-with-default-configure, version: 1.0}", si)
+		info.Hooks["default-configure"] = &snap.HookInfo{}
+		info.Hooks["configure"] = &snap.HookInfo{}
+		return info, emptyContainer(c), nil
+	}
+
+	restore := snapstate.MockOpenSnapFile(openSnapFile)
+	defer restore()
+
+	err := snapstate.CheckSnap(s.st, "snap-path", "snap-with-default-configure", nil, nil, snapstate.Flags{}, nil)
+	c.Check(err, IsNil)
+}
+
+func (s *checkSnapSuite) TestCheckConfigureHooksUnHappy(c *C) {
+	var openSnapFile = func(path string, si *snap.SideInfo) (*snap.Info, snap.Container, error) {
+		info := snaptest.MockInfo(c, "{name: snap-with-default-configure, version: 1.0}", si)
+		info.Hooks["default-configure"] = &snap.HookInfo{}
+		return info, emptyContainer(c), nil
+	}
+
+	restore := snapstate.MockOpenSnapFile(openSnapFile)
+	defer restore()
+
+	err := snapstate.CheckSnap(s.st, "snap-path", "snap-with-default-configure", nil, nil, snapstate.Flags{}, nil)
+	c.Check(err, ErrorMatches, `cannot specify "default-configure" hook without "configure" hook`)
+}
+
+const desktopFileIDsYamlTemplate = `
+name: %s
+version: 1.0
+plugs:
+  desktop:
+    desktop-file-ids: [org.example.Foo]
+`
+
+func (s *snapmgrTestSuite) TestCheckDesktopFileIDsConflicts(c *C) {
+	someSnap, err := snap.InfoFromSnapYaml([]byte(fmt.Sprintf(desktopFileIDsYamlTemplate, "some-snap")))
+	c.Assert(err, IsNil)
+	c.Assert(someSnap.Plugs["desktop"], NotNil)
+	otherSnap, err := snap.InfoFromSnapYaml([]byte(fmt.Sprintf(desktopFileIDsYamlTemplate, "other-snap")))
+	c.Assert(err, IsNil)
+	c.Assert(otherSnap.Plugs["desktop"], NotNil)
+
+	restore := snapstate.MockSnapReadInfo(func(name string, si *snap.SideInfo) (*snap.Info, error) {
+		switch name {
+		case "some-snap":
+			return someSnap, nil
+		case "other-snap":
+			return otherSnap, nil
+		default:
+			return s.fakeBackend.ReadInfo(name, si)
+		}
+	})
+	defer restore()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+	si := &snap.SideInfo{
+		RealName: "other-snap",
+		Revision: snap.R(-42),
+	}
+	snapstate.Set(s.state, "other-snap", &snapstate.SnapState{
+		Active:   true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
+		Current:  si.Revision,
+		SnapType: "app",
+	})
+
+	err = snapstate.CheckDesktopFileIDsConflicts(s.state, someSnap)
+	c.Assert(err, ErrorMatches, `snap "some-snap" requesting desktop-file-id "org.example.Foo" conflicts with snap "other-snap" use`)
+}
+
+func (s *snapmgrTestSuite) TestCheckDesktopFileIDsConflictsNoConflictWithSelf(c *C) {
+	someSnap, err := snap.InfoFromSnapYaml([]byte(fmt.Sprintf(desktopFileIDsYamlTemplate, "some-snap")))
+	c.Assert(err, IsNil)
+	c.Assert(someSnap.Plugs["desktop"], NotNil)
+
+	restore := snapstate.MockSnapReadInfo(func(name string, si *snap.SideInfo) (*snap.Info, error) {
+		switch name {
+		case "some-snap":
+			return someSnap, nil
+		default:
+			return s.fakeBackend.ReadInfo(name, si)
+		}
+	})
+	defer restore()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+	si := &snap.SideInfo{
+		RealName: "some-snap",
+		Revision: snap.R(-42),
+	}
+	snapstate.Set(s.state, "some-snap", &snapstate.SnapState{
+		Active:   true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
+		Current:  si.Revision,
+		SnapType: "app",
+	})
+
+	err = snapstate.CheckDesktopFileIDsConflicts(s.state, someSnap)
+	c.Assert(err, IsNil)
+}
+
+func (s *snapmgrTestSuite) TestInstallDesktopFileIDsConflicts(c *C) {
+	someSnap, err := snap.InfoFromSnapYaml([]byte(fmt.Sprintf(desktopFileIDsYamlTemplate, "some-snap")))
+	c.Assert(err, IsNil)
+	c.Assert(someSnap.Plugs["desktop"], NotNil)
+	otherSnap, err := snap.InfoFromSnapYaml([]byte(fmt.Sprintf(desktopFileIDsYamlTemplate, "other-snap")))
+	c.Assert(err, IsNil)
+	c.Assert(otherSnap.Plugs["desktop"], NotNil)
+
+	restore := snapstate.MockSnapReadInfo(func(name string, si *snap.SideInfo) (*snap.Info, error) {
+		switch name {
+		case "some-snap":
+			return someSnap, nil
+		case "other-snap":
+			return otherSnap, nil
+		default:
+			return s.fakeBackend.ReadInfo(name, si)
+		}
+	})
+	defer restore()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+	si := &snap.SideInfo{
+		RealName: "other-snap",
+		Revision: snap.R(-42),
+	}
+	snapstate.Set(s.state, "other-snap", &snapstate.SnapState{
+		Active:   true,
+		Sequence: snapstatetest.NewSequenceFromSnapSideInfos([]*snap.SideInfo{si}),
+		Current:  si.Revision,
+		SnapType: "app",
+	})
+
+	// Conflict should be detected in early checks
+	opts := &snapstate.RevisionOptions{Channel: "channel-for-desktop-file-ids"}
+	_, err = snapstate.Install(context.Background(), s.state, "some-snap", opts, s.user.ID, snapstate.Flags{})
+	c.Assert(err, ErrorMatches, `snap "some-snap" requesting desktop-file-id "org.example.Foo" conflicts with snap "other-snap" use`)
+}
+
+func (s *snapmgrTestSuite) TestInstallManyDesktopFileIDsConflicts(c *C) {
+	someSnap, err := snap.InfoFromSnapYaml([]byte(fmt.Sprintf(desktopFileIDsYamlTemplate, "some-snap")))
+	c.Assert(err, IsNil)
+	c.Assert(someSnap.Plugs["desktop"], NotNil)
+	otherSnap, err := snap.InfoFromSnapYaml([]byte(fmt.Sprintf(desktopFileIDsYamlTemplate, "other-snap")))
+	c.Assert(err, IsNil)
+	c.Assert(otherSnap.Plugs["desktop"], NotNil)
+
+	restore := snapstate.MockSnapReadInfo(func(name string, si *snap.SideInfo) (*snap.Info, error) {
+		switch name {
+		case "some-snap":
+			return someSnap, nil
+		case "other-snap":
+			return otherSnap, nil
+		default:
+			return s.fakeBackend.ReadInfo(name, si)
+		}
+	})
+	defer restore()
+
+	s.state.Lock()
+	defer s.state.Unlock()
+
+	snapNames := []string{"some-snap", "other-snap"}
+	// Don't use channel-for-desktop-file-ids to pass early checks and instead fail inside the install transaction
+	_, tss, err := snapstate.InstallMany(s.state, snapNames, nil, s.user.ID, nil)
+	c.Assert(err, IsNil)
+
+	chg := s.state.NewChange("install", "install two snaps")
+	for _, ts := range tss {
+		chg.AddAll(ts)
+	}
+
+	s.settle(c)
+
+	// The order of installation is indeterminant, but one will fail
+	c.Check(chg.Err(), ErrorMatches, `cannot perform the following tasks:\n- Make snap "(some|other)-snap" \(11\) available to the system \(snap "(some|other)-snap" requesting desktop-file-id "org.example.Foo" conflicts with snap "(some|other)-snap" use\)`)
+}
