@@ -191,15 +191,7 @@ func (m *SnapManager) doDownloadComponent(t *state.Task, tomb *tomb.Tomb) error
 		rate = autoRefreshRateLimited(st)
 	}
 
-	cpi := snap.MinimalComponentContainerPlaceInfo(
-		compsup.ComponentName(), compsup.CompSideInfo.Revision,
-		snapsup.InstanceName(),
-	)
-
-	// TODO:COMPS: to be consistent with snaps, this should be able to point
-	// somewhere else, based on a path that is in the compsup. this would be
-	// used for creating new recovery systems, like it is now for snaps
-	target := cpi.MountFile()
+	target := compsup.BlobPath(snapsup.InstanceName())
 
 	sto := Store(st, deviceCtx)
 	meter := NewTaskProgressAdapterUnlocked(t)
@@ -296,7 +288,7 @@ func (m *SnapManager) doMountComponent(t *state.Task, _ *tomb.Tomb) (err error)
 	var readInfoErr error
 	for i := 0; i < 10; i++ {
 		compMntDir := cpi.MountDir()
-		_, readInfoErr = readComponentInfo(compMntDir, nil, csi)
+		_, readInfoErr = readComponentInfoAt(compMntDir, nil, csi)
 		if readInfoErr == nil {
 			logger.Debugf("component %q (%v) available at %q",
 				csi.Component, compSetup.Revision(), compMntDir)
@@ -342,8 +334,15 @@ func (m *SnapManager) doMountComponent(t *state.Task, _ *tomb.Tomb) (err error)
 	return nil
 }
 
+// ReadComponentInfo reads the snap's component and returns a ComponentInfo.
+func ReadComponentInfo(snapInfo *snap.Info, csi *snap.ComponentSideInfo) (*snap.ComponentInfo, error) {
+	compName, compRev := csi.Component.ComponentName, csi.Revision
+	mountDir := snap.ComponentMountDir(compName, compRev, snapInfo.InstanceName())
+	return readComponentInfoAt(mountDir, snapInfo, csi)
+}
+
 // Maybe we will need flags as in readInfo
-var readComponentInfo = func(compMntDir string, snapInfo *snap.Info, csi *snap.ComponentSideInfo) (*snap.ComponentInfo, error) {
+var readComponentInfoAt = func(compMntDir string, snapInfo *snap.Info, csi *snap.ComponentSideInfo) (*snap.ComponentInfo, error) {
 	cont := snapdir.New(compMntDir)
 	return snap.ReadComponentInfoFromContainer(cont, snapInfo, csi)
 }
@@ -717,6 +716,42 @@ func (m *SnapManager) doPrepareKernelModulesComponents(t *state.Task, _ *tomb.To
 	// Inject fault during the kernel components setup
 	osutil.MaybeInjectFault("prepare-kernel-components")
 
+	var newInfo *snap.Info
+	// Set the default to false for compatibility with older snapd (case of
+	// joint refresh of snapd and kernel).
+	setNextBoot := false
+	if err := t.Get("set-next-boot", &setNextBoot); err != nil &&
+		!errors.Is(err, state.ErrNoState) {
+		return err
+	}
+	if setNextBoot {
+		// TODO we have to revert changes in bootloader config/modeenv if an
+		// error happens later in this method. This is not likely as possible
+		// errors after this would happen only due to internal errors or not
+		// being able to write to the filesystem, but still. There is also the
+		// question of what would happen if a restart happens when the boot
+		// configuration has been already written but DoneStatus in the state
+		// has not.
+		cand := sequence.NewRevisionSideState(snapsup.SideInfo, nil)
+		newInfo, err = readInfo(snapsup.InstanceName(), cand.Snap, 0)
+		if err != nil {
+			return err
+		}
+		deviceCtx, err := DeviceCtx(st, t, nil)
+		if err != nil {
+			return err
+		}
+		isUndo := false
+		rebootInfo, err := m.backend.MaybeSetNextBoot(newInfo, deviceCtx, isUndo)
+		if err != nil {
+			return err
+		}
+		if rebootInfo.RebootRequired {
+			return m.finishTaskWithMaybeRestart(t, state.DoneStatus,
+				restartPossibility{info: newInfo, RebootInfo: rebootInfo})
+		}
+	}
+
 	// Make sure we won't be rerun
 	t.SetStatus(state.DoneStatus)
 	return nil
