@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2017 Canonical Ltd
+ * Copyright (C) 2017-2023 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -22,6 +22,7 @@ package snapstate
 import (
 	"errors"
 	"fmt"
+	"reflect"
 	"time"
 
 	"github.com/snapcore/snapd/features"
@@ -30,8 +31,8 @@ import (
 	"github.com/snapcore/snapd/overlord/configstate/config"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/release"
-	"github.com/snapcore/snapd/snap"
 	"github.com/snapcore/snapd/store"
+	"github.com/snapcore/snapd/strutil"
 	"github.com/snapcore/snapd/timings"
 )
 
@@ -70,17 +71,22 @@ func (r *refreshHints) needsUpdate() (bool, error) {
 }
 
 func (r *refreshHints) refresh() error {
-	var refreshManaged bool
-	refreshManaged, _, _ = refreshScheduleManaged(r.state)
+	scheduleConf, _, _ := getRefreshScheduleConf(r.state)
+	refreshManaged := scheduleConf == "managed"
 
 	var err error
 	perfTimings := timings.New(map[string]string{"ensure": "refresh-hints"})
 	defer perfTimings.Save(r.state)
 
-	var updates []*snap.Info
-	var ignoreValidationByInstanceName map[string]bool
+	allSnaps, err := All(r.state)
+	if err != nil {
+		return err
+	}
+
+	var plan updatePlan
 	timings.Run(perfTimings, "refresh-candidates", "query store for refresh candidates", func(tm timings.Measurer) {
-		updates, _, ignoreValidationByInstanceName, err = refreshCandidates(auth.EnsureContextTODO(), r.state, nil, nil, &store.RefreshOptions{RefreshManaged: refreshManaged})
+		plan, err = storeUpdatePlan(auth.EnsureContextTODO(),
+			r.state, allSnaps, nil, nil, &store.RefreshOptions{RefreshManaged: refreshManaged}, Options{})
 	})
 	// TODO: we currently set last-refresh-hints even when there was an
 	// error. In the future we may retry with a backoff.
@@ -93,11 +99,15 @@ func (r *refreshHints) refresh() error {
 	if err != nil {
 		return err
 	}
-	hints, err := refreshHintsFromCandidates(r.state, updates, ignoreValidationByInstanceName, deviceCtx)
+
+	hints, err := refreshHintsFromUpdatePlan(r.state, plan, deviceCtx)
 	if err != nil {
 		return fmt.Errorf("internal error: cannot get refresh-candidates: %v", err)
 	}
-	r.state.Set("refresh-candidates", hints)
+
+	// update candidates in state dropping all entries which are not part of
+	// the new hints
+	updateRefreshCandidates(r.state, hints, nil)
 	return nil
 }
 
@@ -122,8 +132,13 @@ func (r *refreshHints) Ensure() error {
 	r.state.Lock()
 	defer r.state.Unlock()
 
+	online, err := isStoreOnline(r.state)
+	if err != nil || !online {
+		return err
+	}
+
 	// CanAutoRefresh is a hook that is set by the devicestate
-	// code to ensure that we only AutoRefersh if the device has
+	// code to ensure that we only AutoRefresh if the device has
 	// bootstraped itself enough. This is only nil when snapstate
 	// is used in isolation (like in tests).
 	if CanAutoRefresh == nil {
@@ -143,54 +158,55 @@ func (r *refreshHints) Ensure() error {
 	return r.refresh()
 }
 
-func refreshHintsFromCandidates(st *state.State, updates []*snap.Info, ignoreValidationByInstanceName map[string]bool, deviceCtx DeviceContext) (map[string]*refreshCandidate, error) {
-	if ValidateRefreshes != nil && len(updates) != 0 {
-		userID := 0
-		var err error
-		updates, err = ValidateRefreshes(st, updates, ignoreValidationByInstanceName, userID, deviceCtx)
+func refreshHintsFromUpdatePlan(st *state.State, plan updatePlan, deviceCtx DeviceContext) (map[string]*refreshCandidate, error) {
+	if ValidateRefreshes != nil && len(plan.targets) != 0 {
+		ignoreValidation := make(map[string]bool, len(plan.targets))
+		for _, t := range plan.targets {
+			if t.setup.IgnoreValidation {
+				ignoreValidation[t.info.InstanceName()] = true
+			}
+		}
+
+		const userID = 0
+
+		// if an error isn't returned here, then the returned list of snaps to
+		// refresh will match the input
+		_, err := ValidateRefreshes(st, plan.targetInfos(), ignoreValidation, userID, deviceCtx)
 		if err != nil {
 			return nil, err
 		}
 	}
 
-	hints := make(map[string]*refreshCandidate, len(updates))
-	for _, update := range updates {
+	hints := make(map[string]*refreshCandidate, len(plan.targets))
+	for _, t := range plan.targets {
+		info := t.info
 		var snapst SnapState
-		if err := Get(st, update.InstanceName(), &snapst); err != nil {
+		if err := Get(st, info.InstanceName(), &snapst); err != nil {
 			return nil, err
 		}
 
+		// we don't need to handle potential channel switches here, since those
+		// shouldn't happen during a auto-refresh
+		if snapst.IsInstalled() && !info.Revision.Unset() && snapst.Current == info.Revision {
+			continue
+		}
+
 		flags := snapst.Flags
 		flags.IsAutoRefresh = true
-		flags, err := earlyChecks(st, &snapst, update, flags)
+		snapsup, compsups, err := t.setups(st, Options{
+			DeviceCtx: deviceCtx,
+			Flags:     flags,
+		})
 		if err != nil {
-			logger.Debugf("update hint for %q is not applicable: %v", update.InstanceName(), err)
+			logger.Debugf("update hint for %q is not applicable: %v", info.InstanceName(), err)
 			continue
 		}
 
-		providerContentAttrs := defaultProviderContentAttrs(st, update)
-		snapsup := &refreshCandidate{
-			SnapSetup: SnapSetup{
-				Base:               update.Base,
-				Prereq:             getKeys(providerContentAttrs),
-				PrereqContentAttrs: providerContentAttrs,
-				Channel:            snapst.TrackingChannel,
-				CohortKey:          snapst.CohortKey,
-				// UserID not set
-				Flags:        flags.ForSnapSetup(),
-				DownloadInfo: &update.DownloadInfo,
-				SideInfo:     &update.SideInfo,
-				Type:         update.Type(),
-				PlugsOnly:    len(update.Slots) == 0,
-				InstanceKey:  update.InstanceKey,
-				auxStoreInfo: auxStoreInfo{
-					Website: update.Website,
-					Media:   update.Media,
-				},
-			},
-			Version: update.Version,
+		hints[info.InstanceName()] = &refreshCandidate{
+			SnapSetup:  snapsup,
+			Components: compsups,
+			Monitored:  IsSnapMonitored(st, info.InstanceName()),
 		}
-		hints[update.InstanceName()] = snapsup
 	}
 	return hints, nil
 }
@@ -204,7 +220,7 @@ func pruneRefreshCandidates(st *state.State, snaps ...string) error {
 		return err
 	}
 	// Remove refresh-candidates from state if gate-auto-refresh-hook feature is
-	// not enabled. This acts as a workaround for the case where a snapd from
+	// not enabled and it is not a map. This acts as a workaround for the case where a snapd from
 	// edge was used and created refresh-candidates in the old format (an array)
 	// with the feature enabled, but the feature was then disabled so the new
 	// map format will never make it into the state.
@@ -212,8 +228,24 @@ func pruneRefreshCandidates(st *state.State, snaps ...string) error {
 	// refresh-candidates in the correct format expected here.
 	// See https://forum.snapcraft.io/t/cannot-r-emove-snap-json-cannot-unmarshal-array-into-go-value-of-type-map-string-snapstate-refreshcandidate/27276
 	if !gateAutoRefreshHook {
-		st.Set("refresh-candidates", nil)
-		return nil
+		var rc interface{}
+		err = st.Get("refresh-candidates", &rc)
+		if err != nil {
+			if errors.Is(err, state.ErrNoState) {
+				// nothing to do
+				return nil
+			}
+		}
+		v := reflect.ValueOf(rc)
+		if !v.IsValid() {
+			// nothing to do
+			return nil
+		}
+		if v.Kind() != reflect.Map {
+			// just remove
+			st.Set("refresh-candidates", nil)
+			return nil
+		}
 	}
 
 	var candidates map[string]*refreshCandidate
@@ -228,8 +260,67 @@ func pruneRefreshCandidates(st *state.State, snaps ...string) error {
 
 	for _, snapName := range snaps {
 		delete(candidates, snapName)
+		abortMonitoring(st, snapName)
+	}
+
+	if len(candidates) == 0 {
+		st.Set("refresh-candidates", nil)
+	} else {
+		st.Set("refresh-candidates", candidates)
 	}
 
-	st.Set("refresh-candidates", candidates)
+	return nil
+}
+
+// updateRefreshCandidates updates the current set of refresh candidates stored
+// in the state. When the list of canDropOldNames is empty, existing entries
+// which aren't part of the update are dropped. When the list is non empty, only
+// those entries mentioned in the list are dropped, other existing entries are
+// preserved. Whenever an existing entry is to be replaced, the caller must have
+// provided a hint which preserves the hint's state outside of snap-setup.
+func updateRefreshCandidates(st *state.State, hints map[string]*refreshCandidate, canDropOldNames []string) error {
+	var oldHints map[string]*refreshCandidate
+	if err := st.Get("refresh-candidates", &oldHints); err != nil {
+		if !errors.Is(err, &state.NoStateError{}) {
+			return err
+		}
+	}
+
+	if len(oldHints) == 0 {
+		st.Set("refresh-candidates", hints)
+		return nil
+	}
+
+	dropAllOld := len(canDropOldNames) == 0
+
+	var deleted []string
+
+	// selectively process existing entries
+	for oldHintName := range oldHints {
+		if newHint, hasUpdate := hints[oldHintName]; hasUpdate {
+			// XXX we rely on the new hint preserving the state
+			oldHints[oldHintName] = newHint
+		} else {
+			if dropAllOld || strutil.ListContains(canDropOldNames, oldHintName) {
+				// we have no new hint for this snap
+				deleted = append(deleted, oldHintName)
+				delete(oldHints, oldHintName)
+			}
+		}
+	}
+	// now add all new entries
+	for newHintName, newHint := range hints {
+		// preserved entries have already been processed
+		if _, processed := oldHints[newHintName]; !processed {
+			oldHints[newHintName] = newHint
+		}
+	}
+
+	// stop monitoring candidates which were deleted
+	for _, dropped := range deleted {
+		abortMonitoring(st, dropped)
+	}
+
+	st.Set("refresh-candidates", oldHints)
 	return nil
 }
