@@ -244,6 +244,18 @@ func (s *seed20) availableAssertedContainers() (map[string]bool, map[string]map[
 	return assertedNames, assertedComps
 }
 
+// Copy copies the system seed to the given seed directory.
+//
+// Note that while most files are copied directly, there are some exceptions
+// that are modified to contain only the information needed for the snaps and
+// components that are actually copied. These include assertions/snaps,
+// options.yaml, and aux-info.json.
+//
+// Note that the grubenv file is not copied at all, as it would not really be
+// possible within the confines of this interface to rewrite this file to
+// contain valid filepaths. This file should be regenerated by the caller with
+// boot.MakeRecoverySystemBootable.
+//
 // Copy implements the Copier interface.
 func (s *seed20) Copy(seedDir string, opts CopyOptions, tm timings.Measurer) (err error) {
 	srcSystemDir, err := filepath.Abs(s.systemDir)
@@ -290,8 +302,11 @@ func (s *seed20) Copy(seedDir string, opts CopyOptions, tm timings.Measurer) (er
 		}
 	}
 
-	// these files are the files we can safely copy without an modifications
-	for _, name := range []string{"assertions/model-etc", "grubenv", "model"} {
+	// these files are the files we can safely copy without an modifications. we
+	// don't copy the grubenv file that might be present, since it might contain
+	// paths that won't be valid in newly copied seed. callers should make the
+	// system bootable if that is needed.
+	for _, name := range []string{"assertions/model-etc", "model"} {
 		if err := osutil.CopyFile(
 			filepath.Join(srcSystemDir, name),
 			filepath.Join(destSystemDir, name),
@@ -1486,52 +1501,78 @@ func (s *seed20) ModeSnaps(mode string) ([]*Snap, error) {
 
 	snapsWithMode := make([]*Snap, 0)
 	for _, sn := range s.snaps[s.essentialSnapsNum:] {
-		// since we're handing out pointers here, we need to make a copy of the
-		// snap
-		copied := *sn
-
-		ms, ok := s.modelSnaps[sn.SnapName()]
-		if !ok {
-			// snaps not in the model will be considered as run mode, and so
-			// will all of its components
-			if mode == "run" {
-				// make a copy of the slice so that the caller can't mess with our
-				// internal state
-				copied.Components = append([]Component(nil), sn.Components...)
-				snapsWithMode = append(snapsWithMode, &copied)
-			}
+		copied := s.modeSnap(sn, mode)
+		if copied == nil {
 			continue
 		}
 
-		if !snapModesInclude(ms.Modes, mode) {
-			continue
-		}
-
-		// we'll rebuild the slice of components with only the components that
-		// are applicable to the requested mode
-		copied.Components = nil
-
-		for _, comp := range sn.Components {
-			modelComp, ok := ms.Components[comp.CompSideInfo.Component.ComponentName]
-			if !ok {
-				// components not in the model will be considered as run mode
-				if mode == "run" {
-					copied.Components = append(copied.Components, comp)
-				}
-				continue
-			}
-
-			if snapModesInclude(modelComp.Modes, mode) {
-				copied.Components = append(copied.Components, comp)
-			}
-		}
-
-		snapsWithMode = append(snapsWithMode, &copied)
+		snapsWithMode = append(snapsWithMode, copied)
 	}
 
 	return snapsWithMode, nil
 }
 
+func (s *seed20) modeSnap(sn *Snap, mode string) *Snap {
+	copied := *sn
+	ms, ok := s.modelSnaps[sn.SnapName()]
+	if !ok {
+		// snaps not in the model will be considered as run mode, and so
+		// will all of its components
+		if mode == "run" {
+			// make a copy of the slice so that the caller can't mess with our
+			// internal state
+			copied.Components = append([]Component(nil), copied.Components...)
+			return &copied
+		}
+		return nil
+	}
+
+	if !snapModesInclude(ms.Modes, mode) {
+		return nil
+	}
+
+	// we'll rebuild the slice of components with only the components that
+	// are applicable to the requested mode
+	copied.Components = nil
+
+	for _, comp := range sn.Components {
+		modelComp, ok := ms.Components[comp.CompSideInfo.Component.ComponentName]
+		if !ok {
+			// components not in the model will be considered as run mode
+			if mode == "run" {
+				copied.Components = append(copied.Components, comp)
+			}
+			continue
+		}
+
+		if snapModesInclude(modelComp.Modes, mode) {
+			copied.Components = append(copied.Components, comp)
+		}
+	}
+
+	return &copied
+}
+
+func (s *seed20) ModeSnap(snapName, mode string) (*Snap, error) {
+	if s.mode != AllModes && mode != s.mode {
+		return nil, fmt.Errorf("metadata was loaded only for snaps for mode %s not %s", s.mode, mode)
+	}
+
+	for _, sn := range s.snaps {
+		if sn.SnapName() != snapName {
+			continue
+		}
+
+		copied := s.modeSnap(sn, mode)
+		if copied != nil {
+			return copied, nil
+		}
+		return nil, fmt.Errorf("snap %s is not available for %q mode", snapName, mode)
+	}
+
+	return nil, fmt.Errorf("while looking for mode snap: snap %s not found", snapName)
+}
+
 func (s *seed20) NumSnaps() int {
 	return len(s.snaps)
 }
