@@ -1965,21 +1965,27 @@ func (s *infoSuite) TestSortByType(c *C) {
 		{SuggestedName: "app1", SnapType: "app"},
 		{SuggestedName: "os1", SnapType: "os"},
 		{SuggestedName: "base1", SnapType: "base"},
+		{SuggestedName: "internal1", SnapType: "internal-boot-base"},
 		{SuggestedName: "gadget1", SnapType: "gadget"},
 		{SuggestedName: "kernel1", SnapType: "kernel"},
 		{SuggestedName: "app2", SnapType: "app"},
 		{SuggestedName: "os2", SnapType: "os"},
+		{SuggestedName: "internal2", SnapType: "internal-boot-base"},
 		{SuggestedName: "snapd", SnapType: "snapd"},
 		{SuggestedName: "base2", SnapType: "base"},
 		{SuggestedName: "gadget2", SnapType: "gadget"},
 		{SuggestedName: "kernel2", SnapType: "kernel"},
 	}
-	sort.Stable(snap.ByType(infos))
+	sort.SliceStable(infos, func(i, j int) bool {
+		return infos[i].Type().SortsBefore(infos[j].Type())
+	})
 
 	c.Check(infos, DeepEquals, []*snap.Info{
 		{SuggestedName: "snapd", SnapType: "snapd"},
 		{SuggestedName: "os1", SnapType: "os"},
 		{SuggestedName: "os2", SnapType: "os"},
+		{SuggestedName: "internal1", SnapType: "internal-boot-base"},
+		{SuggestedName: "internal2", SnapType: "internal-boot-base"},
 		{SuggestedName: "kernel1", SnapType: "kernel"},
 		{SuggestedName: "kernel2", SnapType: "kernel"},
 		{SuggestedName: "base1", SnapType: "base"},
@@ -1999,7 +2005,9 @@ func (s *infoSuite) TestSortByTypeAgain(c *C) {
 	snapd.SideInfo = snap.SideInfo{RealName: "snapd"}
 
 	byType := func(snaps ...*snap.Info) []*snap.Info {
-		sort.Stable(snap.ByType(snaps))
+		sort.SliceStable(snaps, func(i, j int) bool {
+			return snaps[i].Type().SortsBefore(snaps[j].Type())
+		})
 		return snaps
 	}
 
@@ -2275,7 +2283,7 @@ version: 1.0`, nil)
 	c.Assert(err, IsNil)
 
 	_, _, err = snap.SnapdAssertionMaxFormatsFromSnapFile(snapf)
-	c.Check(err, ErrorMatches, `cannot extract assertion max formats information, snaps of type app do not carry snapd`)
+	c.Check(err, ErrorMatches, `cannot extract snapd information, snaps of type app do not carry snapd information`)
 }
 
 func (s *infoSuite) TestAppsForPlug(c *C) {
@@ -2502,11 +2510,11 @@ apps:
 	})
 }
 
-func (s *infoSuite) TestRegistryPlugAttrs(c *C) {
+func (s *infoSuite) TestConfdbPlugAttrs(c *C) {
 	plug := &snap.PlugInfo{
 		Snap:      &snap.Info{SuggestedName: "test-snap"},
 		Name:      "test-plug",
-		Interface: "registry",
+		Interface: "confdb",
 		Attrs: map[string]interface{}{
 			"account": "foo",
 			"view":    "bar/baz",
@@ -2514,14 +2522,14 @@ func (s *infoSuite) TestRegistryPlugAttrs(c *C) {
 		},
 	}
 
-	account, registry, view, err := snap.RegistryPlugAttrs(plug)
+	account, confdb, view, err := snap.ConfdbPlugAttrs(plug)
 	c.Assert(err, IsNil)
 	c.Assert(account, Equals, "foo")
-	c.Assert(registry, Equals, "bar")
+	c.Assert(confdb, Equals, "bar")
 	c.Assert(view, Equals, "baz")
 }
 
-func (s *infoSuite) TestRegistryPlugAttrsInvalid(c *C) {
+func (s *infoSuite) TestConfdbPlugAttrsInvalid(c *C) {
 	type testcase struct {
 		iface   string
 		account string
@@ -2532,14 +2540,14 @@ func (s *infoSuite) TestRegistryPlugAttrsInvalid(c *C) {
 	tcs := []testcase{
 		{
 			iface: "other-thing",
-			err:   "must be registry plug: other-thing",
+			err:   "must be confdb plug: other-thing",
 		},
 
 		{
-			iface:   "registry",
+			iface:   "confdb",
 			account: "my-acc",
 			view:    "reg",
-			err:     "\"view\" must conform to <registry>/<view>: reg",
+			err:     "\"view\" must conform to <confdb>/<view>: reg",
 		},
 	}
 
@@ -2554,7 +2562,7 @@ func (s *infoSuite) TestRegistryPlugAttrsInvalid(c *C) {
 			},
 		}
 
-		_, _, _, err := snap.RegistryPlugAttrs(plug)
+		_, _, _, err := snap.ConfdbPlugAttrs(plug)
 		c.Assert(err, ErrorMatches, tc.err)
 	}
 }
