@@ -30,6 +30,7 @@ import (
 	"github.com/juju/ratelimit"
 	"gopkg.in/retry.v1"
 
+	"github.com/snapcore/snapd/httputil"
 	"github.com/snapcore/snapd/overlord/auth"
 	"github.com/snapcore/snapd/progress"
 	"github.com/snapcore/snapd/snap"
@@ -41,6 +42,8 @@ var (
 	ApiURL        = apiURL
 	Download      = download
 
+	DownloadIconImpl = downloadIcon
+
 	ApplyDelta = applyDelta
 
 	AuthLocation      = authLocation
@@ -115,6 +118,8 @@ func IsTransferSpeedError(err error) (ok bool, speed float64) {
 }
 
 func (w *TransferSpeedMonitoringWriter) MeasuredWindowsCount() int {
+	w.mu.Lock()
+	defer w.mu.Unlock()
 	return w.measuredWindows
 }
 
@@ -146,6 +151,14 @@ func MockDownload(f func(ctx context.Context, name, sha3_384, downloadURL string
 	}
 }
 
+func MockMaxIconFilesize(maxSize int64) (restore func()) {
+	return testutil.Mock(&maxIconFilesize, maxSize)
+}
+
+func MockDownloadIcon(f func(ctx context.Context, name, downloadURL string, w ReadWriteSeekTruncater) error) (restore func()) {
+	return testutil.Mock(&downloadIcon, f)
+}
+
 func MockDoDownloadReq(f func(ctx context.Context, storeURL *url.URL, cdnHeader string, resume int64, s *Store, user *auth.UserState) (*http.Response, error)) (restore func()) {
 	orig := doDownloadReq
 	doDownloadReq = f
@@ -170,6 +183,14 @@ func (sto *Store) MockCacher(obs downloadCache) (restore func()) {
 	}
 }
 
+func MockHttputilNewHTTPClient(f func(opts *httputil.ClientOptions) *http.Client) (restore func()) {
+	old := httputilNewHTTPClient
+	httputilNewHTTPClient = f
+	return func() {
+		httputilNewHTTPClient = old
+	}
+}
+
 func (sto *Store) SetDeltaFormat(dfmt string) {
 	sto.deltaFormat = dfmt
 }
