@@ -32,7 +32,6 @@ import (
 	"os/exec"
 	"path/filepath"
 	"strconv"
-	"strings"
 	"sync"
 	"time"
 
@@ -162,10 +161,6 @@ func (s *Store) cdnHeader() (string, error) {
 		return "none", nil
 	}
 
-	if s.dauthCtx == nil {
-		return "", nil
-	}
-
 	// set Snap-CDN from cloud instance information
 	// if available
 
@@ -173,25 +168,7 @@ func (s *Store) cdnHeader() (string, error) {
 	// where we first to send this header and if the
 	// operation fails that way to even get the connection
 	// then we retry without sending this?
-
-	cloudInfo, err := s.dauthCtx.CloudInfo()
-	if err != nil {
-		return "", err
-	}
-
-	if cloudInfo != nil {
-		cdnParams := []string{fmt.Sprintf("cloud-name=%q", cloudInfo.Name)}
-		if cloudInfo.Region != "" {
-			cdnParams = append(cdnParams, fmt.Sprintf("region=%q", cloudInfo.Region))
-		}
-		if cloudInfo.AvailabilityZone != "" {
-			cdnParams = append(cdnParams, fmt.Sprintf("availability-zone=%q", cloudInfo.AvailabilityZone))
-		}
-
-		return strings.Join(cdnParams, " "), nil
-	}
-
-	return "", nil
+	return s.buildLocationString()
 }
 
 type HashError struct {
@@ -206,7 +183,7 @@ func (e HashError) Error() string {
 
 type DownloadOptions struct {
 	RateLimit           int64
-	IsAutoRefresh       bool
+	Scheduled           bool
 	LeavePartialOnError bool
 }
 
@@ -215,11 +192,18 @@ type DownloadOptions struct {
 // The file is saved in temporary storage, and should be removed
 // after use to prevent the disk from running out of space.
 func (s *Store) Download(ctx context.Context, name string, targetPath string, downloadInfo *snap.DownloadInfo, pbar progress.Meter, user *auth.UserState, dlOpts *DownloadOptions) error {
+	// most other store network operations use s.endpointURL, which returns an
+	// error if the store is offline. this doesn't, so we need to explicitly
+	// check.
+	if err := s.checkStoreOnline(); err != nil {
+		return err
+	}
+
 	if err := os.MkdirAll(filepath.Dir(targetPath), 0755); err != nil {
 		return err
 	}
 
-	if err := s.cacher.Get(downloadInfo.Sha3_384, targetPath); err == nil {
+	if s.cacher.Get(downloadInfo.Sha3_384, targetPath) {
 		logger.Debugf("Cache hit for SHA3_384 …%.5s.", downloadInfo.Sha3_384)
 		return nil
 	}
@@ -230,10 +214,17 @@ func (s *Store) Download(ctx context.Context, name string, targetPath string, do
 		if len(downloadInfo.Deltas) == 1 {
 			err := s.downloadAndApplyDelta(name, targetPath, downloadInfo, pbar, user, dlOpts)
 			if err == nil {
-				return nil
+				// try to place the file in the cacher
+				if err = s.cacher.Put(downloadInfo.Sha3_384, targetPath); err == nil {
+					// file is in the cache now
+					return nil
+				} else {
+					logger.Noticef("Cannot place rebuilt blob for %s in cache: %v", name, err)
+				}
+			} else {
+				// We revert to normal downloads if there is any error.
+				logger.Noticef("Cannot download or apply deltas for %s: %v", name, err)
 			}
-			// We revert to normal downloads if there is any error.
-			logger.Noticef("Cannot download or apply deltas for %s: %v", name, err)
 		}
 	}
 
@@ -317,6 +308,42 @@ func (s *Store) Download(ctx context.Context, name string, targetPath string, do
 	return s.cacher.Put(downloadInfo.Sha3_384, targetPath)
 }
 
+// DownloadIcon downloads the icon for the snap from the given download URL to
+// the given target path. Snap icons are small (<256kB) files served from an
+// ordinary unauthenticated file server, so this does not require store
+// authentication or user state, nor a progress bar. They are also not revision-
+// specific, and do not use a download cache.
+func DownloadIcon(ctx context.Context, name string, targetPath string, downloadURL string) error {
+	if err := os.MkdirAll(filepath.Dir(targetPath), 0o755); err != nil {
+		return err
+	}
+
+	// Download icon to a temporary file location, since there may be an active
+	// snap icon hard-linked to targetPath. If download fails, don't want to
+	// corrupt the existing icon file contents. Instead, commit the new file
+	// once the download succeeds, thus leaving any previous icon linked to the
+	// original unchanged icon contents, until a future task re-links to the
+	// new contents.
+	aw, err := osutil.NewAtomicFile(targetPath, 0o644, 0, osutil.NoChown, osutil.NoChown)
+	if err != nil {
+		return fmt.Errorf("cannot create file for snap icon for snap %s: %v", name, err)
+	}
+	// on success, Cancel becomes a no-op
+	defer aw.Cancel()
+
+	logger.Debugf("Starting download of %q to %q.", downloadURL, targetPath)
+
+	if err = downloadIcon(ctx, name, downloadURL, aw); err != nil {
+		logger.Debugf("download of %q failed: %#v", downloadURL, err)
+		return err
+	}
+
+	if err = aw.Commit(); err != nil {
+		return fmt.Errorf("cannot commit snap icon file for snap %s: %v", name, err)
+	}
+	return nil
+}
+
 func downloadReqOpts(storeURL *url.URL, cdnHeader string, opts *DownloadOptions) *requestOptions {
 	reqOptions := requestOptions{
 		Method:       "GET",
@@ -328,7 +355,7 @@ func downloadReqOpts(storeURL *url.URL, cdnHeader string, opts *DownloadOptions)
 	if cdnHeader != "" {
 		reqOptions.ExtraHeaders["Snap-CDN"] = cdnHeader
 	}
-	if opts != nil && opts.IsAutoRefresh {
+	if opts != nil && opts.Scheduled {
 		reqOptions.ExtraHeaders["Snap-Refresh-Reason"] = "scheduled"
 	}
 
@@ -494,7 +521,18 @@ func downloadImpl(ctx context.Context, name, sha3_384, downloadURL string, user
 			return fmt.Errorf("the download has been cancelled: %s", downloadCtx.Err())
 		}
 		var resp *http.Response
-		cli := s.newHTTPClient(nil)
+		cli := s.newHTTPClient(nil) // XXX: there's no timeout defined for this client, and the context is context.TODO(), so it won't be cancelled
+		oldCheckRedirect := cli.CheckRedirect
+		if oldCheckRedirect == nil {
+			panic("internal error: the httputil.NewHTTPClient-produced http.Client must have CheckRedirect defined")
+		}
+		cli.CheckRedirect = func(req *http.Request, via []*http.Request) error {
+			// remove user/device auth headers from being sent in "CDN" redirects
+			// see also: https://bugs.launchpad.net/snapd/+bug/2027993
+			// TODO: do we need to remove other identifying headers?
+			dropAuthorization(req, &AuthorizeOptions{deviceAuth: true, apiLevel: reqOptions.APILevel})
+			return oldCheckRedirect(req, via)
+		}
 		resp, finalErr = s.doRequest(downloadCtx, cli, reqOptions, user)
 		if cancelled(downloadCtx) {
 			return fmt.Errorf("the download has been cancelled: %s", downloadCtx.Err())
@@ -598,8 +636,101 @@ func downloadImpl(ctx context.Context, name, sha3_384, downloadURL string, user
 	return finalErr
 }
 
+type ReadWriteSeekTruncater interface {
+	io.Reader
+	io.Writer
+	io.Seeker
+
+	Truncate(size int64) error
+}
+
+var (
+	maxIconFilesize int64 = 300000
+	downloadIcon          = downloadIconImpl
+)
+
+// downloadIconImpl writes an http.Request which does not require authentication
+// or a progress.Meter.
+func downloadIconImpl(ctx context.Context, name, downloadURL string, w ReadWriteSeekTruncater) error {
+	iconURL, err := url.Parse(downloadURL)
+	if err != nil {
+		return err
+	}
+
+	startTime := time.Now()
+	errRetry := errors.New("retry")
+	for attempt := retry.Start(downloadRetryStrategy, nil); attempt.Next(); {
+		httputil.MaybeLogRetryAttempt(iconURL.String(), attempt, startTime)
+
+		err = func() error {
+			clientOpts := &httputil.ClientOptions{} // XXX: should this have a timeout?
+			cli := httputil.NewHTTPClient(clientOpts)
+			var resp *http.Response
+			resp, err = doIconRequest(ctx, cli, iconURL)
+			if err != nil {
+				if httputil.ShouldRetryAttempt(attempt, err) {
+					return errRetry
+				}
+				return err
+			}
+
+			defer resp.Body.Close()
+
+			if httputil.ShouldRetryHttpResponse(attempt, resp) {
+				return errRetry
+			}
+
+			if resp.StatusCode != 200 {
+				return &DownloadError{Code: resp.StatusCode, URL: resp.Request.URL}
+			}
+
+			if resp.ContentLength == 0 || resp.ContentLength > maxIconFilesize {
+				return fmt.Errorf("unsupported Content-Length for %s (must be nonzero and <%dB): %d", downloadURL, maxIconFilesize, resp.ContentLength)
+			}
+			logger.Debugf("download size for %s: %d", downloadURL, resp.ContentLength)
+
+			_, err = io.CopyN(w, resp.Body, resp.ContentLength)
+
+			if err != nil {
+				if httputil.ShouldRetryAttempt(attempt, err) {
+					if _, err := w.Seek(0, io.SeekStart); err != nil {
+						return err
+					}
+					if err := w.Truncate(0); err != nil {
+						return err
+					}
+					return errRetry
+				}
+				return err
+			}
+
+			return nil
+		}()
+
+		if err == errRetry {
+			continue
+		}
+
+		if err == nil {
+			logger.Debugf("icon download succeeded for %s", name)
+		}
+
+		return err
+	}
+	// The loop will return an error directly if retries are exhausted and an
+	// error occurs, so we never actually break out of the loop and get here.
+	return errors.New("icon download retries exhausted")
+}
+
 // DownloadStream will copy the snap from the request to the io.Reader
 func (s *Store) DownloadStream(ctx context.Context, name string, downloadInfo *snap.DownloadInfo, resume int64, user *auth.UserState) (io.ReadCloser, int, error) {
+	// most other store network operations use s.endpointURL, which returns an
+	// error if the store is offline. this doesn't, so we need to explicitly
+	// check.
+	if err := s.checkStoreOnline(); err != nil {
+		return nil, 0, err
+	}
+
 	// XXX: coverage of this is rather poor
 	if path := s.cacher.GetPath(downloadInfo.Sha3_384); path != "" {
 		logger.Debugf("Cache hit for SHA3_384 …%.5s.", downloadInfo.Sha3_384)
