@@ -20,6 +20,7 @@
 package store
 
 import (
+	"bufio"
 	"context"
 	"encoding/base64"
 	"encoding/json"
@@ -29,6 +30,7 @@ import (
 	"net/http"
 	"net/url"
 	"os"
+	"path"
 	"path/filepath"
 	"regexp"
 	"strconv"
@@ -58,6 +60,11 @@ func hexify(in string) string {
 	return fmt.Sprintf("%x", bs)
 }
 
+type snapCachedInfo struct {
+	info *snap.Info
+	err  error
+}
+
 // Store is our snappy software store implementation
 type Store struct {
 	url       string
@@ -68,6 +75,10 @@ type Store struct {
 	fallback       *store.Store
 
 	srv *http.Server
+
+	channelRepository *ChannelRepository
+
+	snapsCache map[string]snapCachedInfo
 }
 
 // NewStore creates a new store server serving snaps from the given top directory and assertions from topDir/asserts. If assertFallback is true missing assertions are looked up in the main online store.
@@ -90,6 +101,10 @@ func NewStore(topDir, addr string, assertFallback bool) *Store {
 			Addr:    addr,
 			Handler: mux,
 		},
+		channelRepository: &ChannelRepository{
+			rootDir: filepath.Join(topDir, "channels"),
+		},
+		snapsCache: make(map[string]snapCachedInfo),
 	}
 
 	mux.HandleFunc("/", rootEndpoint)
@@ -97,6 +112,10 @@ func NewStore(topDir, addr string, assertFallback bool) *Store {
 	mux.HandleFunc("/api/v1/snaps/details/", store.detailsEndpoint)
 	mux.HandleFunc("/api/v1/snaps/metadata", store.bulkEndpoint)
 	mux.Handle("/download/", http.StripPrefix("/download/", http.FileServer(http.Dir(topDir))))
+
+	mux.HandleFunc("/api/v1/snaps/auth/nonces", store.nonceEndpoint)
+	mux.HandleFunc("/api/v1/snaps/auth/sessions", store.sessionEndpoint)
+
 	// v2
 	mux.HandleFunc("/v2/assertions/", store.assertionsEndpoint)
 	mux.HandleFunc("/v2/snaps/refresh", store.snapActionEndpoint)
@@ -111,6 +130,14 @@ func (s *Store) URL() string {
 	return s.url
 }
 
+func (s *Store) RealURL(req *http.Request) string {
+	if req.Host == "" {
+		return s.url
+	} else {
+		return fmt.Sprintf("http://%s", req.Host)
+	}
+}
+
 func (s *Store) SnapsDir() string {
 	return s.blobDir
 }
@@ -179,34 +206,34 @@ type essentialInfo struct {
 	Base        string
 }
 
-var errInfo = errors.New("cannot get info")
-
-func snapEssentialInfo(w http.ResponseWriter, fn, snapID string, bs asserts.Backstore) (*essentialInfo, error) {
-	f, err := snapfile.Open(fn)
-	if err != nil {
-		http.Error(w, fmt.Sprintf("cannot read: %v: %v", fn, err), 400)
-		return nil, errInfo
-	}
-
+func (s *Store) snapEssentialInfo(fn, snapID string, bs asserts.Backstore) (*essentialInfo, error) {
 	restoreSanitize := snap.MockSanitizePlugsSlots(func(snapInfo *snap.Info) {})
 	defer restoreSanitize()
 
-	info, err := snap.ReadInfoFromSnapFile(f, nil)
-	if err != nil {
-		http.Error(w, fmt.Sprintf("cannot get info for: %v: %v", fn, err), 400)
-		return nil, errInfo
+	cached, isCached := s.snapsCache[fn]
+	if !isCached {
+		f, err := snapfile.Open(fn)
+		if err != nil {
+			return nil, fmt.Errorf("cannot read: %v: %v", fn, err)
+		}
+
+		info, err := snap.ReadInfoFromSnapFile(f, nil)
+		cached.info = info
+		cached.err = err
+		s.snapsCache[fn] = cached
+	}
+	if cached.err != nil {
+		return nil, fmt.Errorf("cannot get info for: %v: %v", fn, cached.err)
 	}
 
 	snapDigest, size, err := asserts.SnapFileSHA3_384(fn)
 	if err != nil {
-		http.Error(w, fmt.Sprintf("cannot get digest for: %v: %v", fn, err), 400)
-		return nil, errInfo
+		return nil, fmt.Errorf("cannot get digest for: %v: %v", fn, err)
 	}
 
 	snapRev, devAcct, err := findSnapRevision(snapDigest, bs)
 	if err != nil && !errors.Is(err, &asserts.NotFoundError{}) {
-		http.Error(w, fmt.Sprintf("cannot get info for: %v: %v", fn, err), 400)
-		return nil, errInfo
+		return nil, fmt.Errorf("cannot get info for: %v: %v", fn, err)
 	}
 
 	var devel, develID string
@@ -220,24 +247,97 @@ func snapEssentialInfo(w http.ResponseWriter, fn, snapID string, bs asserts.Back
 		// XXX: fallback until we are always assertion based
 		develID = defaultDeveloperID
 		devel = defaultDeveloper
-		revision = makeRevision(info)
+		revision = makeRevision(cached.info)
 	}
 
 	return &essentialInfo{
-		Name:        info.SnapName(),
+		Name:        cached.info.SnapName(),
 		SnapID:      snapID,
 		DeveloperID: develID,
 		DevelName:   devel,
 		Revision:    revision,
-		Version:     info.Version,
+		Version:     cached.info.Version,
 		Digest:      snapDigest,
 		Size:        size,
-		Confinement: string(info.Confinement),
-		Type:        string(info.Type()),
-		Base:        info.Base,
+		Confinement: string(cached.info.Confinement),
+		Type:        string(cached.info.Type()),
+		Base:        cached.info.Base,
 	}, nil
 }
 
+func addComponentBlobToRevisionSet(snaps map[string]*revisionSet, snapIDs map[string]string, fn string, bs asserts.Backstore) error {
+	f, err := snapfile.Open(fn)
+	if err != nil {
+		return fmt.Errorf("cannot read: %v: %v", fn, err)
+	}
+
+	info, err := snap.ReadComponentInfoFromContainer(f, nil, nil)
+	if err != nil {
+		return fmt.Errorf("cannot get info for: %v: %v", fn, err)
+	}
+
+	compName := info.Component.ComponentName
+	snapName := info.Component.SnapName
+
+	digest, _, err := asserts.SnapFileSHA3_384(fn)
+	if err != nil {
+		return fmt.Errorf("cannot get digest for: %v: %v", fn, err)
+	}
+
+	set, ok := snaps[snapName]
+	if !ok {
+		return fmt.Errorf("cannot find snap %q for component: %q", snapName, info.Component)
+	}
+
+	snapID, ok := snapIDs[snapName]
+	if !ok {
+		return fmt.Errorf("cannot find snap id for snap %q", snapName)
+	}
+
+	pk, err := asserts.PrimaryKeyFromHeaders(asserts.SnapResourceRevisionType, map[string]string{
+		"snap-id":           snapID,
+		"resource-name":     compName,
+		"resource-sha3-384": digest,
+	})
+	if err != nil {
+		return err
+	}
+
+	a, err := bs.Get(asserts.SnapResourceRevisionType, pk, asserts.SnapResourceRevisionType.MaxSupportedFormat())
+	if err != nil {
+		return err
+	}
+	compRev := snap.R(a.(*asserts.SnapResourceRevision).ResourceRevision())
+
+	for snapRev := range set.revisions {
+		pk, err := asserts.PrimaryKeyFromHeaders(asserts.SnapResourcePairType, map[string]string{
+			"resource-name":     compName,
+			"snap-id":           snapID,
+			"resource-revision": compRev.String(),
+			"snap-revision":     snapRev.String(),
+		})
+		if err != nil {
+			return err
+		}
+
+		_, err = bs.Get(asserts.SnapResourcePairType, pk, asserts.SnapResourcePairType.MaxSupportedFormat())
+		if err != nil {
+			// no pair assertion for this snap revision, so this one isn't
+			// associated with this snap revision
+			if errors.Is(err, &asserts.NotFoundError{}) {
+				continue
+			}
+			return err
+		}
+
+		if err := set.addComponent(compName, compRev, fn, snapRev); err != nil {
+			return err
+		}
+	}
+
+	return nil
+}
+
 type detailsReplyJSON struct {
 	Architectures   []string `json:"architecture"`
 	SnapID          string   `json:"snap_id"`
@@ -348,23 +448,23 @@ func (s *Store) detailsEndpoint(w http.ResponseWriter, req *http.Request) {
 		http.Error(w, fmt.Sprintf("internal error collecting assertions: %v", err), 500)
 		return
 	}
-	snaps, err := s.collectSnaps()
+	snaps, err := s.collectSnaps(bs)
 	if err != nil {
 		http.Error(w, fmt.Sprintf("internal error collecting snaps: %v", err), 500)
 		return
 	}
 
-	fn, ok := snaps[pkg]
+	set, ok := snaps[pkg]
 	if !ok {
 		http.NotFound(w, req)
 		return
 	}
 
-	essInfo, err := snapEssentialInfo(w, fn, "", bs)
-	if essInfo == nil {
-		if err != errInfo {
-			panic(err)
-		}
+	sn := set.getLatest()
+
+	essInfo, err := s.snapEssentialInfo(sn.path, "", bs)
+	if err != nil {
+		http.Error(w, err.Error(), 400)
 		return
 	}
 
@@ -374,8 +474,8 @@ func (s *Store) detailsEndpoint(w http.ResponseWriter, req *http.Request) {
 		PackageName:     essInfo.Name,
 		Developer:       essInfo.DevelName,
 		DeveloperID:     essInfo.DeveloperID,
-		AnonDownloadURL: fmt.Sprintf("%s/download/%s", s.URL(), filepath.Base(fn)),
-		DownloadURL:     fmt.Sprintf("%s/download/%s", s.URL(), filepath.Base(fn)),
+		AnonDownloadURL: fmt.Sprintf("%s/download/%s", s.RealURL(req), filepath.Base(sn.path)),
+		DownloadURL:     fmt.Sprintf("%s/download/%s", s.RealURL(req), filepath.Base(sn.path)),
 		Version:         essInfo.Version,
 		Revision:        essInfo.Revision,
 		DownloadDigest:  hexify(essInfo.Digest),
@@ -394,28 +494,116 @@ func (s *Store) detailsEndpoint(w http.ResponseWriter, req *http.Request) {
 	w.Write(out)
 }
 
-func (s *Store) collectSnaps() (map[string]string, error) {
+type revisionSet struct {
+	latest    snap.Revision
+	revisions map[snap.Revision]availableSnap
+}
+
+type availableSnap struct {
+	path       string
+	components map[string]availableComponent
+}
+
+type availableComponent struct {
+	path     string
+	revision snap.Revision
+}
+
+func (rs *revisionSet) get(rev snap.Revision) (availableSnap, bool) {
+	if rev.Unset() {
+		rev = rs.latest
+	}
+
+	sn, ok := rs.revisions[rev]
+	return sn, ok
+}
+
+func (rs *revisionSet) getLatest() availableSnap {
+	sn, ok := rs.revisions[rs.latest]
+	if !ok {
+		panic("internal error: revision set should always contain latest revision")
+	}
+
+	return sn
+}
+
+func (rs *revisionSet) add(rev snap.Revision, path string) {
+	if rs.revisions == nil {
+		rs.revisions = make(map[snap.Revision]availableSnap)
+	}
+
+	if rs.latest.N < rev.N {
+		rs.latest = rev
+	}
+	rs.revisions[rev] = availableSnap{path: path, components: make(map[string]availableComponent)}
+}
+
+func (rs *revisionSet) addComponent(name string, compRev snap.Revision, path string, snapRev snap.Revision) error {
+	sn, ok := rs.revisions[snapRev]
+	if !ok {
+		return fmt.Errorf("cannot find snap revision %q", snapRev)
+	}
+
+	sn.components[name] = availableComponent{path: path, revision: compRev}
+
+	return nil
+}
+
+func (s *Store) collectSnaps(bs asserts.Backstore) (map[string]*revisionSet, error) {
 	snapFns, err := filepath.Glob(filepath.Join(s.blobDir, "*.snap"))
 	if err != nil {
 		return nil, err
 	}
 
-	snaps := map[string]string{}
-
 	restoreSanitize := snap.MockSanitizePlugsSlots(func(snapInfo *snap.Info) {})
 	defer restoreSanitize()
 
+	snaps := make(map[string]*revisionSet)
+	snapNamesToID := make(map[string]string, len(snapFns))
 	for _, fn := range snapFns {
-		f, err := snapfile.Open(fn)
+		// if the snap is asserted, then the returned info will contain the ID
+		// taken from the database
+		const snapID = ""
+		info, err := s.snapEssentialInfo(fn, snapID, bs)
 		if err != nil {
 			return nil, err
 		}
-		info, err := snap.ReadInfoFromSnapFile(f, nil)
+
+		if _, ok := snaps[info.Name]; !ok {
+			snaps[info.Name] = &revisionSet{}
+		}
+
+		snaps[info.Name].add(snap.R(info.Revision), fn)
+
+		channels, err := s.channelRepository.findSnapChannels(info.Digest)
 		if err != nil {
 			return nil, err
 		}
-		snaps[info.SnapName()] = fn
-		logger.Debugf("found snap %q at %v", info.SnapName(), fn)
+
+		for _, channel := range channels {
+			compositeName := fmt.Sprintf("%s|%s", info.Name, channel)
+			if _, ok := snaps[compositeName]; !ok {
+				snaps[compositeName] = &revisionSet{}
+			}
+			snaps[compositeName].add(snap.R(info.Revision), fn)
+		}
+
+		if info.SnapID != "" {
+			snapNamesToID[info.Name] = info.SnapID
+		}
+
+		logger.Debugf("found snap %q (revision %d) at %v", info.Name, info.Revision, fn)
+	}
+
+	compFns, err := filepath.Glob(filepath.Join(s.blobDir, "*.comp"))
+	if err != nil {
+		return nil, err
+	}
+
+	for _, fn := range compFns {
+		if err := addComponentBlobToRevisionSet(snaps, snapNamesToID, fn, bs); err != nil {
+			return nil, err
+		}
 	}
 
 	return snaps, err
@@ -484,7 +672,7 @@ func (s *Store) bulkEndpoint(w http.ResponseWriter, req *http.Request) {
 		return
 	}
 
-	snaps, err := s.collectSnaps()
+	snaps, err := s.collectSnaps(bs)
 	if err != nil {
 		http.Error(w, fmt.Sprintf("internal error collecting snaps: %v", err), 500)
 		return
@@ -498,31 +686,34 @@ func (s *Store) bulkEndpoint(w http.ResponseWriter, req *http.Request) {
 			return
 		}
 
-		if fn, ok := snaps[name]; ok {
-			essInfo, err := snapEssentialInfo(w, fn, pkg.SnapID, bs)
-			if essInfo == nil {
-				if err != errInfo {
-					panic(err)
-				}
-				return
-			}
-
-			replyData.Payload.Packages = append(replyData.Payload.Packages, detailsReplyJSON{
-				Architectures:   []string{"all"},
-				SnapID:          essInfo.SnapID,
-				PackageName:     essInfo.Name,
-				Developer:       essInfo.DevelName,
-				DeveloperID:     essInfo.DeveloperID,
-				DownloadURL:     fmt.Sprintf("%s/download/%s", s.URL(), filepath.Base(fn)),
-				AnonDownloadURL: fmt.Sprintf("%s/download/%s", s.URL(), filepath.Base(fn)),
-				Version:         essInfo.Version,
-				Revision:        essInfo.Revision,
-				DownloadDigest:  hexify(essInfo.Digest),
-				Confinement:     essInfo.Confinement,
-				Type:            essInfo.Type,
-				Base:            essInfo.Base,
-			})
+		set, ok := snaps[name]
+		if !ok {
+			continue
 		}
+
+		sn := set.getLatest()
+
+		essInfo, err := s.snapEssentialInfo(sn.path, pkg.SnapID, bs)
+		if err != nil {
+			http.Error(w, err.Error(), 400)
+			return
+		}
+
+		replyData.Payload.Packages = append(replyData.Payload.Packages, detailsReplyJSON{
+			Architectures:   []string{"all"},
+			SnapID:          essInfo.SnapID,
+			PackageName:     essInfo.Name,
+			Developer:       essInfo.DevelName,
+			DeveloperID:     essInfo.DeveloperID,
+			DownloadURL:     fmt.Sprintf("%s/download/%s", s.RealURL(req), filepath.Base(sn.path)),
+			AnonDownloadURL: fmt.Sprintf("%s/download/%s", s.RealURL(req), filepath.Base(sn.path)),
+			Version:         essInfo.Version,
+			Revision:        essInfo.Revision,
+			DownloadDigest:  hexify(essInfo.Digest),
+			Confinement:     essInfo.Confinement,
+			Type:            essInfo.Type,
+			Base:            essInfo.Base,
+		})
 	}
 
 	// use indent because this is a development tool, output
@@ -575,8 +766,9 @@ func (s *Store) collectAssertions() (asserts.Backstore, error) {
 }
 
 type currentSnap struct {
-	SnapID      string `json:"snap-id"`
-	InstanceKey string `json:"instance-key"`
+	SnapID          string `json:"snap-id"`
+	InstanceKey     string `json:"instance-key"`
+	TrackingChannel string `json:"tracking-channel"`
 }
 
 type snapAction struct {
@@ -585,6 +777,7 @@ type snapAction struct {
 	SnapID      string `json:"snap-id"`
 	Name        string `json:"name"`
 	Revision    int    `json:"revision,omitempty"`
+	Channel     string `json:"channel,omitempty"`
 }
 
 type snapActionRequest struct {
@@ -614,15 +807,26 @@ type detailsResultV2 struct {
 		ID       string `json:"id"`
 		Username string `json:"username"`
 	} `json:"publisher"`
-	Download struct {
-		URL      string `json:"url"`
-		Sha3_384 string `json:"sha3-384"`
-		Size     uint64 `json:"size"`
-	} `json:"download"`
-	Version     string `json:"version"`
-	Revision    int    `json:"revision"`
-	Confinement string `json:"confinement"`
-	Type        string `json:"type"`
+	Download    downloadInfo         `json:"download"`
+	Version     string               `json:"version"`
+	Revision    int                  `json:"revision"`
+	Confinement string               `json:"confinement"`
+	Type        string               `json:"type"`
+	Resources   []snapResourceResult `json:"resources,omitempty"`
+}
+
+type downloadInfo struct {
+	URL      string `json:"url"`
+	Sha3_384 string `json:"sha3-384"`
+	Size     uint64 `json:"size"`
+}
+
+type snapResourceResult struct {
+	Download downloadInfo `json:"download"`
+	Type     string       `json:"type"`
+	Name     string       `json:"name"`
+	Revision int          `json:"revision"`
+	Version  string       `json:"version"`
 }
 
 func (s *Store) snapActionEndpoint(w http.ResponseWriter, req *http.Request) {
@@ -653,7 +857,7 @@ func (s *Store) snapActionEndpoint(w http.ResponseWriter, req *http.Request) {
 		return
 	}
 
-	snaps, err := s.collectSnaps()
+	snaps, err := s.collectSnaps(bs)
 	if err != nil {
 		http.Error(w, fmt.Sprintf("internal error collecting snaps: %v", err), 500)
 		return
@@ -667,6 +871,7 @@ func (s *Store) snapActionEndpoint(w http.ResponseWriter, req *http.Request) {
 				Action:      "refresh",
 				SnapID:      s.SnapID,
 				InstanceKey: s.InstanceKey,
+				Channel:     s.TrackingChannel,
 			}
 		}
 	}
@@ -680,43 +885,107 @@ func (s *Store) snapActionEndpoint(w http.ResponseWriter, req *http.Request) {
 		}
 
 		if name == "" {
+			if isAutoRefreshRequest(req) {
+				// Skip auto-refresh for snap installed on the system not added to the fakestore
+				continue
+			}
 			http.Error(w, fmt.Sprintf("unknown snap-id: %q", snapID), 400)
 			return
 		}
 
-		if fn, ok := snaps[name]; ok {
-			essInfo, err := snapEssentialInfo(w, fn, snapID, bs)
-			if essInfo == nil {
-				if err != errInfo {
-					panic(err)
-				}
+		var set *revisionSet
+		var foundSnap bool
+		if a.Channel != "" {
+			set, foundSnap = snaps[fmt.Sprintf("%s|%s", name, a.Channel)]
+		}
+		if !foundSnap {
+			// FIXME: It is possible that many tests do
+			// not use channels correctly. So we have to
+			// fallback to searching for snaps by just
+			// name, without channel. Maybe we should
+			// remove that, and fix all the tests instead.
+			set, foundSnap = snaps[name]
+		}
+
+		if !foundSnap {
+			continue
+		}
+
+		sn, ok := set.get(snap.R(a.Revision))
+
+		if !ok {
+			// TODO: this should send back some error?
+			continue
+		}
+
+		essInfo, err := s.snapEssentialInfo(sn.path, snapID, bs)
+		if err != nil {
+			http.Error(w, err.Error(), 400)
+			return
+		}
+
+		resources := make([]snapResourceResult, 0, len(sn.components))
+		for compName, comp := range sn.components {
+			f, err := snapfile.Open(path.Join(comp.path))
+			if err != nil {
+				http.Error(w, fmt.Sprintf("cannot read: %v: %v", compName, err), 400)
 				return
 			}
 
-			res := &snapActionResult{
-				Result:      a.Action,
-				InstanceKey: a.InstanceKey,
-				SnapID:      essInfo.SnapID,
-				Name:        essInfo.Name,
-				Snap: detailsResultV2{
-					Architectures: []string{"all"},
-					SnapID:        essInfo.SnapID,
-					Name:          essInfo.Name,
-					Version:       essInfo.Version,
-					Revision:      essInfo.Revision,
-					Confinement:   essInfo.Confinement,
-					Type:          essInfo.Type,
-					Base:          essInfo.Base,
-				},
+			digest, size, err := asserts.SnapFileSHA3_384(comp.path)
+			if err != nil {
+				http.Error(w, fmt.Sprintf("cannot get digest for: %v: %v", compName, err), 400)
+				return
 			}
-			logger.Debugf("requested snap %q revision %d", essInfo.Name, a.Revision)
-			res.Snap.Publisher.ID = essInfo.DeveloperID
-			res.Snap.Publisher.Username = essInfo.DevelName
-			res.Snap.Download.URL = fmt.Sprintf("%s/download/%s", s.URL(), filepath.Base(fn))
-			res.Snap.Download.Sha3_384 = hexify(essInfo.Digest)
-			res.Snap.Download.Size = essInfo.Size
-			replyData.Results = append(replyData.Results, res)
+
+			compInfo, err := snap.ReadComponentInfoFromContainer(f, nil, nil)
+			if err != nil {
+				http.Error(w, fmt.Sprintf("cannot get info for: %v: %v", compName, err), 400)
+				return
+			}
+
+			resources = append(resources, snapResourceResult{
+				Name:     compName,
+				Revision: comp.revision.N,
+				Type:     fmt.Sprintf("component/%s", compInfo.Type),
+				Version:  compInfo.Version(essInfo.Version),
+				Download: downloadInfo{
+					URL:      fmt.Sprintf("%s/download/%s", s.RealURL(req), filepath.Base(comp.path)),
+					Sha3_384: hexify(digest),
+					Size:     size,
+				},
+			})
 		}
+
+		details := detailsResultV2{
+			Architectures: []string{"all"},
+			SnapID:        essInfo.SnapID,
+			Name:          essInfo.Name,
+			Version:       essInfo.Version,
+			Revision:      essInfo.Revision,
+			Confinement:   essInfo.Confinement,
+			Type:          essInfo.Type,
+			Base:          essInfo.Base,
+		}
+		if len(resources) > 0 {
+			details.Resources = resources
+		}
+
+		res := &snapActionResult{
+			Result:      a.Action,
+			InstanceKey: a.InstanceKey,
+			SnapID:      essInfo.SnapID,
+			Name:        essInfo.Name,
+			Snap:        details,
+		}
+
+		logger.Debugf("requested snap %q revision %d", essInfo.Name, a.Revision)
+		res.Snap.Publisher.ID = essInfo.DeveloperID
+		res.Snap.Publisher.Username = essInfo.DevelName
+		res.Snap.Download.URL = fmt.Sprintf("%s/download/%s", s.RealURL(req), filepath.Base(sn.path))
+		res.Snap.Download.Sha3_384 = hexify(essInfo.Digest)
+		res.Snap.Download.Size = essInfo.Size
+		replyData.Results = append(replyData.Results, res)
 	}
 
 	// use indent because this is a development tool, output
@@ -729,6 +998,10 @@ func (s *Store) snapActionEndpoint(w http.ResponseWriter, req *http.Request) {
 	w.Write(out)
 }
 
+func isAutoRefreshRequest(req *http.Request) bool {
+	return req.Header.Get("Snap-Refresh-Reason") == "scheduled"
+}
+
 func (s *Store) retrieveAssertion(bs asserts.Backstore, assertType *asserts.AssertionType, primaryKey []string) (asserts.Assertion, error) {
 	a, err := bs.Get(assertType, primaryKey, assertType.MaxSupportedFormat())
 	if errors.Is(err, &asserts.NotFoundError{}) && s.assertFallback {
@@ -869,3 +1142,38 @@ func findSnapRevision(snapDigest string, bs asserts.Backstore) (*asserts.SnapRev
 
 	return snapRev, devAcct, nil
 }
+
+func (s *Store) nonceEndpoint(w http.ResponseWriter, req *http.Request) {
+	w.Header().Set("Content-Type", "application/json")
+	w.WriteHeader(200)
+	w.Write([]byte(`{"nonce": "blah"}`))
+}
+
+func (s *Store) sessionEndpoint(w http.ResponseWriter, req *http.Request) {
+	w.Header().Set("Content-Type", "application/json")
+	w.WriteHeader(200)
+	w.Write([]byte(`{"macaroon": "blahblah"}`))
+}
+
+type ChannelRepository struct {
+	rootDir string
+}
+
+func (cr *ChannelRepository) findSnapChannels(snapDigest string) ([]string, error) {
+	dataPath := filepath.Join(cr.rootDir, snapDigest)
+	f, err := os.Open(dataPath)
+	if err != nil {
+		if errors.Is(err, os.ErrNotExist) {
+			return nil, nil
+		}
+		return nil, err
+	}
+	defer f.Close()
+
+	sc := bufio.NewScanner(f)
+	var lines []string
+	for sc.Scan() {
+		lines = append(lines, sc.Text())
+	}
+	return lines, nil
+}
